<?php
/**
 * @package   ats
 * @copyright Copyright (c)2011-2022 Nicholas K. Dionysopoulos / Akeeba Ltd
 * @license   GNU General Public License version 3, or later
 */

defined('_JEXEC') or die();

use Akeeba\TicketSystem\Admin\Model\Tickets;
use FOF40\Container\Container;
use FOF40\Model\DataModel\Collection;
use FOF40\Timer\Timer;
use Joomla\CMS\Language\Text;
use Joomla\CMS\Log\Log;
use Joomla\CMS\Plugin\CMSPlugin;

/**
 * Automatically closes stale tickets
 *
 * @since 3.2.0
 */
class plgAtsAutoclose extends CMSPlugin
{
	const atsCommandName = 'autoclose';

	/**
	 * @inheritDoc
	 */
	public function __construct(&$subject, $config = [])
	{
		parent::__construct($subject, $config);

		if (!defined('FOF40_INCLUDED') && !@include_once(JPATH_LIBRARIES . '/fof40/include.php'))
		{
			throw new RuntimeException('This extension requires FOF 4.', 500);
		}

		$this->loadLanguage();
	}

	/**
	 * Returns information about the CRON task provided by this plugin
	 *
	 * @return  array
	 *
	 * @since   3.2.0
	 */
	public function onAtsCronTaskInfo()
	{
		$container = Container::getInstance('com_ats');

		return [
			'command'     => self::atsCommandName,
			'label'       => 'PLG_ATS_' . $this->_name . '_TASK_LABEL',
			'description' => 'PLG_ATS_' . $this->_name . '_TASK_DESC',
		];
	}

	/**
	 * Handles the ATS CRON tasks.
	 *
	 * @param   string  $command  The command name ATS CRON was asked to execute.
	 * @param   array   $options  Any options passed to this CRON job
	 *
	 * @return  bool|null  True on success, null if it's not the expected command name.
	 * @since   3.2.0
	 */
	public function onAtsCronTask($command, array $options = [])
	{
		// Make sure we are calling the correct command
		if ($command != self::atsCommandName)
		{
			return null;
		}

		$timeLimit  = array_key_exists('time_limit', $options) ? $options['time_limit'] : 86400;
		$timer      = new Timer($timeLimit);
		$container  = Container::getInstance('com_ats');
		$max_age    = $this->params->get('maxinactiveticket', 30);
		$silent_age = $this->params->get('silentticketcloseage', 60);

		Log::add(sprintf("Closing tickets inactive over %s days", $max_age), Log::DEBUG, 'ats.cron');
		Log::add(sprintf("Silent close for tickets older than %s days", $silent_age), Log::DEBUG, 'ats.cron');

		// Load the language files
		$jlang = $container->platform->getLanguage();
		$jlang->load('com_ats', JPATH_ADMINISTRATOR);
		$jlang->load('com_ats.override', JPATH_ADMINISTRATOR);

		// Loop while we have tickets and we have not hit the time limit
		while ($timer->getTimeLeft())
		{
			$noSilentLeft  = false;
			$noVerboseLeft = false;

			if ($silent_age >= 1)
			{
				Log::add("Looking for up to 100 tickets to close silently", Log::DEBUG, 'ats.cron');

				$until   = time() - $silent_age * 3600 * 24;
				$tickets = $this->getTicketsUntil($container, $until);

				Log::add(sprintf("Found %d tickets to close silently", $tickets->count()), Log::DEBUG, 'ats.cron');

				if ($tickets->count() == 0)
				{
					$noSilentLeft = true;
				}

				/** @var Tickets $ticket */
				foreach ($tickets as $ticket)
				{
					if ($timer->getTimeLeft() <= 0)
					{
						Log::add(sprintf('I have reached the time limit of %u second(s). Stopping for now.', $timeLimit), Log::DEBUG, 'ats.cron');

						return true;
					}

					Log::add(sprintf("\tSilently closing #%s %s", $ticket->ats_ticket_id, $ticket->title), Log::DEBUG, 'ats.cron');

					$ticket->status = 'C';

					Tickets::$noPermissionsCheck = 1;
					$ticket->save();
					Tickets::$noPermissionsCheck = 0;
				}
			}
			else
			{
				$noSilentLeft = true;
			}

			if ($timer->getTimeLeft() <= 0)
			{
				Log::add(sprintf('I have reached the time limit of %u second(s). Stopping for now.', $timeLimit), Log::DEBUG, 'ats.cron');

				return true;
			}

			// Do I have to close tickets verbosely?
			if (($max_age < $silent_age) && ($max_age > 0))
			{
				Log::add("Looking for up to 100 tickets to close verbosely", Log::DEBUG, 'ats.cron');

				$until   = time() - $max_age * 3600 * 24;
				$tickets = $this->getTicketsUntil($container, $until);

				Log::add(sprintf("Found %d tickets to close verbosely", $tickets->count()), Log::DEBUG, 'ats.cron');

				if ($tickets->count() == 0)
				{
					$noVerboseLeft = true;
				}

				$jNow = $container->platform->getDate();

				/** @var Tickets $ticket */
				foreach ($tickets as $ticket)
				{
					if ($timer->getTimeLeft() <= 0)
					{
						Log::add(sprintf('I have reached the time limit of %u second(s). Stopping for now.', $timeLimit), Log::DEBUG, 'ats.cron');

						return true;
					}

					Log::add(sprintf("\tVerbosely closing #%s %s", $ticket->ats_ticket_id, $ticket->title), Log::DEBUG, 'ats.cron');

					$post = [
						'ats_ticket_id' => $ticket->ats_ticket_id,
						'content'       => Text::_('COM_ATS_CLI_CLOSEPOSTMSG'),
						'created_on'    => $jNow->toSql(),
						'created_by'    => -1,
						'enabled'       => 1,
					];

					$ticket->getNew('posts')->save($post);

					$ticket->status              = 'C';
					Tickets::$noPermissionsCheck = 1;
					$ticket->save();
					Tickets::$noPermissionsCheck = 0;
				}
			}
			else
			{
				$noVerboseLeft = true;
			}

			if ($noSilentLeft && $noVerboseLeft)
			{
				return true;
			}
		}

		Log::add("Done auto-closing inactive tickets.", Log::DEBUG, 'ats.cron');

		return true;
	}

	/**
	 * Get a Tickets collection for up to 100 ATS tickets modified up to and including the $until date string.
	 *
	 * @param   Container  $container
	 * @param   string     $until
	 *
	 * @return  Collection
	 */
	private function getTicketsUntil(Container $container, $until)
	{
		$jUntil = $container->platform->getDate($until);

		/** @var Tickets $ticketModel */
		$ticketModel = $container->factory->model('Tickets')->tmpInstance();

		$tickets = $ticketModel
			->modified_on([
				'method'   => 'search',
				'operator' => '<=',
				'value'    => $jUntil->toSql(),
			])
			->status_array('O,P')
			->get(true, 0, 100);

		return $tickets;
	}

}