<?php
/**
 * @package   ats
 * @copyright Copyright (c)2011-2022 Nicholas K. Dionysopoulos / Akeeba Ltd
 * @license   GNU General Public License version 3, or later
 */

namespace Akeeba\TicketSystem\Admin\Controller;

defined('_JEXEC') or die;

use Akeeba\TicketSystem\Admin\Model\SchedulingInfo as SchedulingInfoModel;
use Exception;
use FOF40\Container\Container;
use FOF40\Controller\Controller;
use FOF40\Controller\Mixin\PredefinedTaskList;
use Joomla\CMS\Language\Text;
use RuntimeException;

class SchedulingInfo extends Controller
{
	use PredefinedTaskList;

	/** @inheritDoc */
	public function __construct(Container $container, array $config = [])
	{
		parent::__construct($container, $config);

		$this->predefinedTaskList = ['default', 'applySecretWord', 'fixPlugin'];
	}

	/**
	 * Applies the randomly generated Secret Word in the session
	 *
	 * @return  void
	 *
	 * @throws  Exception
	 *
	 * @since   3.2.0
	 */
	public function applySecretWord()
	{
		$this->csrfProtection();

		/** @var SchedulingInfoModel $model */
		$model = $this->getModel();
		$model->applySecretWord();

		$this->setRedirect('index.php?option=com_ats&view=SchedulingInfo');
	}

	/**
	 * Publish a plugin as public and set its Enable CRON option (if it exists) to 1
	 *
	 * @return  void
	 *
	 * @since   3.2.0
	 */
	public function fixPlugin()
	{
		$plugin = $this->input->get('plugin', '');

		/** @var SchedulingInfoModel $model */
		$model = $this->getModel();
		$model->publishPlugin($plugin);

		$this->setRedirect('index.php?option=com_ats&view=SchedulingInfo');
	}

	/**
	 * Apply custom access control for all tasks.
	 *
	 * @param   string  $task  The task being executed
	 *
	 * @return  void
	 *
	 * @since   3.2.0
	 */
	protected function onBeforeExecute(&$task)
	{
		if (!$this->container->platform->authorise('core.manage', 'com_ats'))
		{
			throw new RuntimeException(Text::_('JERROR_ALERTNOAUTHOR'), 403);
		}
	}

	/**
	 * Executed before the main display task.
	 *
	 * @return  void
	 *
	 * @throws  Exception
	 *
	 * @since   3.2.0
	 */
	protected function onBeforeDefault()
	{
		if ($this->input->getMethod() == 'POST')
		{
			$this->csrfProtection();

			$commandName = $this->input->post->getCmd('commandName', 'commandName');
			$commandName = empty($commandName) ? 'commandName' : $commandName;

			$this->container->platform->setSessionVar('commandName', $commandName, 'ats.cron');
		}

		$this->conditionallyGenerateNewSecretWord();
	}

	/**
	 * Generate a new, random Secret Word if one is not set and we have not already stored one in the session.
	 *
	 * @return  void
	 *
	 * @since   3.2.0
	 */
	private function conditionallyGenerateNewSecretWord()
	{
		/** @var SchedulingInfoModel $model */
		$model = $this->getModel();

		// Make sure we have a secret word
		if ($model->hasSecretWord())
		{
			return;
		}

		// If we don't have a secret word generate a new, random one but only once per session.
		$newSecretWord = $this->container->platform->getSessionVar('newSecretWord', '', 'ats.cron');

		if (empty($newSecretWord))
		{
			$model->createNewSecretWord();
		}
	}
}