<?php
/**
 * @package   ats
 * @copyright Copyright (c)2011-2022 Nicholas K. Dionysopoulos / Akeeba Ltd
 * @license   GNU General Public License version 3, or later
 */

namespace Akeeba\TicketSystem\Admin\CustomField;

defined('_JEXEC') or die;

use Akeeba\TicketSystem\Admin\Model\CustomFields;
use FOF40\Container\Container;
use stdClass;

/**
 * Abstract base class for custom per-user and per-subscription fields
 *
 * @since 2.6.0
 */
abstract class Base
{
	/** The field type, set by the constructor */
	protected $fieldType = null;

	protected $db = null;

	/**
	 * Create a custom field object instance
	 *
	 * @param   array  $config  Custom configuration parameters
	 */
	public function __construct(array $config = [])
	{
		$container = Container::getInstance('com_ats');
		$this->db  = $container->db;

		// Set up the field type
		if (!isset($config['field_type']))
		{
			if (empty($this->fieldType))
			{
				$parts                = $container->inflector->explode(get_called_class());
				$type                 = strtolower(array_pop($parts));
				$config['field_type'] = $type;
			}
			else
			{
				$config['field_type'] = $this->fieldType;
			}
		}
		$this->fieldType = $config['field_type'];
	}

	/**
	 * Creates a custom field array which will be used by the renderer to be
	 * shown on the front- or back-end page, for per-user fields
	 *
	 * @param   CustomFields  $item   A custom field definition
	 * @param   array         $cache  The values cache
	 *
	 * @return  array  Field definition
	 */
	public function getField($item, $cache)
	{
		return null;
	}

	/**
	 * Validate a per-user field
	 *
	 * @param   CustomFields  $item    The custom field to validate
	 * @param   array         $custom  The custom fields' values array
	 *
	 * @return  int|null  1 if the field is valid, 0 otherwise, null if not supported
	 */
	public function validate($item, &$custom)
	{
		return null;
	}

	/**
	 * Render the field's value for display in the ticket
	 *
	 * @param   CustomFields  $item   The custom field to render
	 * @param   mixed         $value  The custom field's current raw value
	 *
	 * @return  string  The HTML to display for this field
	 *
	 * @since   3.0.0
	 */
	public function getDisplay(CustomFields $item, $value)
	{
		return $value;
	}

	/**
	 * Returns the current value of a field
	 *
	 * @param   CustomFields  $item        A custom field definition
	 * @param   array         $cache       The values cache
	 * @param   boolean       $isCheckbox  Is it a checkbox? If true, the default value will be parsed as a
	 *                                     human-readable boolean
	 *
	 * @return  mixed
	 *
	 * @since   3.2.0
	 */
	protected function getCurrentValue(CustomFields $item, array $cache, $isCheckbox = false)
	{
		if (array_key_exists($item->slug, $cache['params']))
		{
			return $cache['params'][$item->slug];
		}

		$default = $item->default;

		if ($isCheckbox)
		{
			$default = in_array(!is_string($default) ? 'NO' : strtoupper($default), [
				'YES', 'TRUE', 'ON', '1', 'ENABLED', 'CHECKED', 'SELECTED',
			]) ? 1 : 0;
		}

		return $default;
	}
}
