<?php
/**
 * @package   ats
 * @copyright Copyright (c)2011-2022 Nicholas K. Dionysopoulos / Akeeba Ltd
 * @license   GNU General Public License version 3, or later
 */

namespace Akeeba\TicketSystem\Admin\CustomField;

defined('_JEXEC') or die();

use Akeeba\TicketSystem\Admin\Helper\Format;
use Akeeba\TicketSystem\Admin\Model\CustomFields;
use Joomla\CMS\HTML\HTMLHelper;
use Joomla\CMS\Language\Text as JText;

/**
 * A date input field
 *
 * @author Nicholas K. Dionysopoulos
 */
class Date extends Base
{
	public function getField($item, $cache)
	{
		// Get the current value
		$current = $this->getCurrentValue($item, $cache);

		// Is this a required field?
		$required = $item->allow_empty ? '' : '* ';

		// Set up field's HTML content
		$escapeSlug = htmlentities($item->slug);
		$escapeCurrent = htmlentities($current);

		if (version_compare(JVERSION, '3.999.999', 'lt'))
		{
			$html = HTMLHelper::_('calendar', $current, 'params[' . $item->slug . ']', $item->slug);
		}
		else
		{
			$html = <<< HTML
<input
        type="date"
        pattern="[0-9]{4}-[0-9]{2}-[0-9]{2}"
        name="params[{$escapeSlug}]"
        id="$escapeSlug"
        value="{$escapeCurrent}"
>

HTML;

		}



		// Setup the field
		$field = [
			'id'          => $item->slug,
			'label'       => $required . JText::_($item->title),
			'elementHTML' => $html,
			'isValid'     => $required ? (!empty($current) && ($current != '0000-00-00 00:00:00') && ($current != $this->db->getNullDate())) : true,
		];

		if ($item->invalid_label)
		{
			$field['invalidLabel'] = JText::_($item->invalid_label);
		}
		if ($item->valid_label)
		{
			$field['validLabel'] = JText::_($item->valid_label);
		}

		return $field;
	}

	/**
	 * Validate a date field
	 *
	 * @param   CustomFields  $item    The custom field to validate
	 * @param   array         $custom  The custom fields' values array
	 *
	 * @return  int  1 if the field is valid, 0 otherwise
	 */
	public function validate($item, &$custom)
	{
		if (!is_array($custom) || !array_key_exists($item->slug, $custom))
		{
			$custom[$item->slug] = '';
		}

		$value = $custom[$item->slug];

		// Do I allow empty values and it's actually empty? If so return true immediately
		if ($item->allow_empty)
		{
			if (empty($value) || ($value == '0000-00-00 00:00:00') || ($value == $this->db->getNullDate()))
			{
				return 1;
			}
		}

		// Special case: we do not want empty values and we have a MySQL-style "empty" value. Return false immediately
		if (!$item->allow_empty && (($value == '0000-00-00 00:00:00') || ($value == $this->db->getNullDate())))
		{
			return 0;
		}

		// In any other case, let's use the regex to validate the date
		$regex = '/^\d{1,4}(\/|-)\d{1,2}(\/|-)\d{2,4}[[:space:]]{0,}(\d{1,2}:\d{1,2}(:\d{1,2}){0,1}){0,1}$/';
		$valid = preg_match($regex, $value);

		return $valid ? 1 : 0;
	}

	public function getDisplay(CustomFields $item, $value)
	{
		if (!$value)
		{
			return '';
		}

		return Format::date($value);
	}
}
