<?php
/**
 * @package   ats
 * @copyright Copyright (c)2011-2022 Nicholas K. Dionysopoulos / Akeeba Ltd
 * @license   GNU General Public License version 3, or later
 */

namespace Akeeba\TicketSystem\Admin\Helper;

defined('_JEXEC') or die;

use FOF40\Container\Container;
use \Joomla\CMS\User\User as JUser;

/**
 * A helper class to make it easier working with add-on notification email addresses.
 *
 * Add-on notification email addresses are managed at the end-user level through the plugins/user/ats plugin.
 *
 * Add-on emails and their associated names are a repeatable field in the user profile. The JSON data returned by Joomla
 * is stored in the #__user_profiles table under the ats.notification_emails key.
 *
 * This helper class can return a list of these emails per user, as well as perform a reverse query (find a user who
 * owns this particular email address). Moreover it provides validation tools which prevent users defining add-on
 * email addresses already used by other Joomla! users, either as their main email or as an add-on email. The latter
 * is used by the user plugin to auto-filter the list.
 */
abstract class AddonEmails
{
	/**
	 * Get a list of add-on emails for the given user ID. The emails are returned as an associative array of email =>
	 * full name.
	 *
	 * @param   int  $user_id  The numeric ID of the user
	 *
	 * @return  array
	 *
	 * @since   2.4.0
	 */
	public static function getAddonEmails($user_id)
	{
		$container = Container::getInstance('com_ats');
		$db        = $container->db;

		$query = $db->getQuery(true)
			->select('profile_value')
			->from($db->qn('#__user_profiles'))
			->where($db->qn('profile_key') . ' = ' . $db->q('ats.notification_emails', false))
			->where($db->qn('user_id') . ' = ' . $db->q($user_id));

		$json = $db->setQuery($query)->loadResult();

		if (empty($json))
		{
			return [];
		}

		return self::decodeAddonEmailJSON($json);
	}

	/**
	 * Get a list of add-on emails for the given user object. The emails are returned as an associative array of
	 * email => full name.
	 *
	 * @param   JUser  $user  The user object
	 *
	 * @return  array
	 *
	 * @since   2.4.0
	 */
	public static function getAddonEmailsByUser(JUser $user)
	{
		return self::getAddonEmails($user->id);
	}

	/**
	 * Find the user ID given an add-on email.
	 *
	 * This method will first look for a user whose main email address matches $email. If none is found, it will look
	 * into the user profiles, doing a text match against the JSON data. The matched user records will then be loaded
	 * and processed properly. The user ID of the first record to legitimately include $email in the add-on email list
	 * will be returned.
	 *
	 * @param   string  $email  The email to search
	 *
	 * @return  int  The user ID or 0 if no match is found
	 *
	 * @since   2.4.0
	 */
	public static function findUserIdByAddonEmail($email)
	{
		$container   = Container::getInstance('com_ats');
		$db          = $container->db;
		$email_lower = strtolower($email);
		$email_json  = json_encode($email_lower);

		// First look for an exact match in #__users
		$query = $db->getQuery(true)
			->select('id')
			->from('#__users')
			->where($db->qn('email') . ' = ' . $db->q($email), 'OR')
			->where($db->qn('email') . ' = ' . $db->q($email_lower), 'OR');
		$id = $db->setQuery($query)->loadResult();

		if (!empty($id))
		{
			return $id;
		}

		// Then look for add-on emails in the user profiles
		$query = $db->getQuery(true)
			->select('user_id')
			->from($db->qn('#__user_profiles'))
			->where($db->qn('profile_key') . ' = ' . $db->q('ats.notification_emails', false))
			->where($db->qn('profile_value') . ' LIKE ' . $db->q('%' . $email_json . '%'));

		$userIDs = $db->setQuery($query)->loadColumn(0);

		// No match? Return 0.
		if (empty($userIDs))
		{
			return 0;
		}

		// Finally make sure we really did find the exact add-on email in a user profile record, or return 0
		foreach ($userIDs as $uid)
		{
			$emails = self::getAddonEmails($uid);

			if (array_key_exists($email, $emails))
			{
				return $uid;
			}
		}

		return 0;
	}

	/**
	 * Similar to findUserIdByAddonEmail but returns the JUser record instead.
	 *
	 * @param   string  $email  The email to search
	 *
	 * @return  JUser
	 *
	 * @since   2.4.0
	 */
	public static function findUserByAddonEmail($email)
	{
		$container = Container::getInstance('com_ats');

		return Permissions::getUser(self::findUserIdByAddonEmail($email));
	}

	/**
	 * Filters an associative list of email => full name. If the email exists as the main or add-on email of another
	 * user it will be removed from the list of emails. It's crucial to provide the correct $user_id value. Any records
	 * found to exist as add-on emails for a user _other_ than $user_id will be filtered out.
	 *
	 * @param   array  $emailAssoc  The list of emails to filter
	 * @param   int    $user_id     The user record these emails will be added to.
	 *
	 * @return  array  The filtered array
	 *
	 * @since   2.4.0
	 */
	public static function filterAddonEmails($emailAssoc, $user_id)
	{
		if (empty($emailAssoc))
		{
			return [];
		}

		$ret = [];

		foreach ($emailAssoc as $email => $name)
		{
			$id = self::findUserIdByAddonEmail($email);

			if ($id && ($id != $user_id))
			{
				continue;
			}

			$ret[$email] = $name;
		}

		return $ret;
	}

	/**
	 * Decodes the clumsy Joomla! repeatable field JSON format into an email => real name array.
	 *
	 * Joomla! data example:
	 *
	 * {"email":["client1@local.web","client2@local.web","client3@local.web"],"name":["Client One","Client Two","Client Three"]}
	 *
	 * Resulting array example:
	 * [
	 *  'client1@local.web' => 'Client One',
	 *  'client2@local.web' => 'Client Two',
	 *  'client3@local.web' => 'Client Three',
	 * ]
	 *
	 * @param   string  $json
	 *
	 * @return  array
	 *
	 * @since   2.4.0
	 */
	public static function decodeAddonEmailJSON($json)
	{
		$assoc = @json_decode($json, true);

		if (empty($assoc))
		{
			return [];
		}

		if (!isset($assoc['email']) || !isset($assoc['name']))
		{
			return [];
		}

		$ret = [];
		$i   = 0;

		foreach ($assoc['email'] as $email)
		{
			$ret[$email] = $assoc['name'][$i];
			$i++;
		}

		return $ret;
	}

	/**
	 * The inverse of decodeAddonEmailJSON. It takes the associative array of email => full name and encodes it back
	 * into the clumsy Joomla! repeatable field JSON format.
	 *
	 * @param   array  $emailAssoc
	 *
	 * @return  string  JSON encoded string
	 *
	 * @since   2.4.0
	 */
	public static function encodeAddonEmailJSON($emailAssoc)
	{
		$ret = [
			'email' => [],
			'name' => [],
		];

		if (!empty($emailAssoc))
		{
			foreach ($emailAssoc as $email => $name)
			{
				$ret['email'][] = $email;
				$ret['name'][] = $name;
			}
		}

		return json_encode($ret);
	}
}
