<?php
/**
 * @package   ats
 * @copyright Copyright (c)2011-2022 Nicholas K. Dionysopoulos / Akeeba Ltd
 * @license   GNU General Public License version 3, or later
 */

namespace Akeeba\TicketSystem\Admin\Helper;

defined('_JEXEC') or die;

use FOF40\Container\Container;
use FOF40\View\Compiler\Blade as BladeCompiler;

/**
 * A helper class to extend the Blade compiler with custom elements used in ATS
 *
 * @since       3.0.0
 */
abstract class Blade
{
	/**
	 * Extend the Blade compiler with custom Blade templates.
	 *
	 * Remember that createMatcher generates a regex which matches the argument *including* parentheses. That's why you
	 * see $2 (the matched argument list) WITHOUT surrounding parentheses.
	 *
	 * @param   Container  $container  The FOF container where the custom Blade templates will be applied to.
	 *
	 * @return  void
	 *
	 * @since   3.0.0
	 */
	public static function extend(Container $container)
	{
		self::addAtsTicketStatusSelect($container);
		self::addAtsCategorySelect($container);
		self::addAtsFormatDate($container);
		self::addAtsFormatDate2($container);
		self::addAtsPrioritySelect($container);
		self::addParseBBCode($container);
	}

	/**
	 * Add `@atsTicketStatusSelect`. Shows a dropdown to select the ticket status.
	 *
	 * @param   Container  $container
	 *
	 * @return  void
	 *
	 * @since   3.0.0
	 */
	private static function addAtsTicketStatusSelect(Container $container)
	{
		$container->blade->extend(function ($template, BladeCompiler $compiler) {
			$pattern = $compiler->createMatcher('atsTicketStatusSelect');

			return preg_replace($pattern, '<?php echo \\Akeeba\\TicketSystem\\Admin\\Helper\\Select::ticketstatuses$2; ?>', $template);
		});
	}

	/**
	 * Add `@atsCategorySelect`. Shows a dropdown to select the ticket category.
	 *
	 * @param   Container  $container
	 *
	 * @return  void
	 *
	 * @since   3.0.0
	 */
	private static function addAtsCategorySelect(Container $container)
	{
		$container->blade->extend(function ($template, BladeCompiler $compiler) {
			$pattern = $compiler->createMatcher('atsCategorySelect');

			return preg_replace($pattern, '<?php echo \\Akeeba\\TicketSystem\\Admin\\Helper\\Select::getCategories$2; ?>', $template);
		});
	}

	/**
	 * Add `@atsPrioritySelect`. Shows a dropdown to select the ticket priority.
	 *
	 * @param   Container  $container
	 *
	 * @return  void
	 *
	 * @since   3.0.0
	 */
	private static function addAtsPrioritySelect(Container $container)
	{
		$container->blade->extend(function ($template, BladeCompiler $compiler) {
			$pattern = $compiler->createMatcher('atsPrioritySelect');

			return preg_replace($pattern, '<?php echo \\Akeeba\\TicketSystem\\Admin\\Helper\\Select::priorities$2; ?>', $template);
		});
	}

	/**
	 * Add `@atsFormatDate`. Formats a generic date string.
	 *
	 * @param   Container  $container
	 *
	 * @return  void
	 *
	 * @since   3.0.0
	 */
	private static function addAtsFormatDate(Container $container)
	{
		$container->blade->extend(function ($template, BladeCompiler $compiler) {
			$pattern = $compiler->createMatcher('atsFormatDate');

			return preg_replace($pattern, '<?php echo \\Akeeba\\TicketSystem\\Admin\\Helper\\Format::date$2; ?>', $template);
		});
	}

	/**
	 * Add `@atsFormatDateTZ`. Formats a generic date string with the user's timezone.
	 *
	 * @param   Container  $container
	 *
	 * @return  void
	 *
	 * @since   3.0.0
	 */
	private static function addAtsFormatDate2(Container $container)
	{
		$container->blade->extend(function ($template, BladeCompiler $compiler) {
			$pattern = $compiler->createMatcher('atsFormatDateTZ');

			return preg_replace($pattern, '<?php echo \\Akeeba\\TicketSystem\\Admin\\Helper\\Format::date2$2; ?>', $template);
		});
	}

	/**
	 * Add `@parseBBCode`. Echoes the input after parsing it for bbCode
	 *
	 * @param   Container  $container
	 *
	 * @return  void
	 *
	 * @since   3.0.0
	 */
	private static function addParseBBCode(Container $container)
	{
		$container->blade->extend(function ($template, BladeCompiler $compiler) {
			$pattern = $compiler->createMatcher('parseBBCode');

			return preg_replace($pattern, '<?php echo \\Akeeba\\TicketSystem\\Admin\\Helper\\Bbcode::parseBBCode$2; ?>', $template);
		});
	}

}