<?php
/**
 * @package   ats
 * @copyright Copyright (c)2011-2022 Nicholas K. Dionysopoulos / Akeeba Ltd
 * @license   GNU General Public License version 3, or later
 */

namespace Akeeba\TicketSystem\Admin\Helper;

defined('_JEXEC') or die;

use FOF40\Container\Container;
use Joomla\CMS\Component\ComponentHelper;
use Joomla\CMS\Factory;
use Joomla\CMS\Mail\Mail;
use Joomla\CMS\Uri\Uri;
use phpmailerException;

class Email
{
	/**
	 * Allowed image file extensions to inline in sent emails
	 *
	 * @var   array
	 */
	private static $allowedImageExtensions = ['jpg', 'jpeg', 'png', 'gif', 'bmp', 'svg'];

	/**
	 * Load an email template from the database
	 *
	 * @param   string  $key  The email template key
	 *
	 * @return  array  A plain array: [$subject, $body, $loadedLanguage]
	 */
	public static function loadEmailTemplateFromDB($key)
	{
		// Initialise
		$templateText = '';
		$subject      = '';
		$container    = Container::getInstance('com_ats');

		// Look for desired languages
		$jLang     = Factory::getLanguage();
		$userLang  = Permissions::getUser()->getParam('language', '');
		$languages = [
			$userLang,
			$jLang->getTag(),
			$jLang->getDefault(),
			'en-GB',
			'*',
		];

		// Look for an override in the database
		$db    = $container->db;
		$query = $db->getQuery(true)
			->select('*')
			->from($db->qn('#__ats_emailtemplates'))
			->where($db->qn('key') . '=' . $db->q($key))
			->where($db->qn('enabled') . '=' . $db->q(1));
		$db->setQuery($query);
		$allTemplates = $db->loadObjectList('language');

		// Try finding the most relevant language override and load it
		$loadLanguage = null;

		foreach ($languages as $lang)
		{
			if (!array_key_exists($lang, $allTemplates))
			{
				continue;
			}

			if ($lang != '*')
			{
				$loadLanguage = $lang;
			}

			$subject      = $allTemplates[$lang]->subject;
			$templateText = $allTemplates[$lang]->template;

			// If I found a match for this $languages key, it has higher
			// affinity than the next entries on the list, therefore I needn't
			// look further.
			break;
		}

		// Because SpamAssassin blacklists our domain when it falsely thinks an email is spam.
		$replaceThat = [
			'<p style=\"text-align: right; font-size: 7pt; color: #ccc;\">Powered by <a href=\"https://www.akeeba.com/products/ats.html\" style=\"color: #ccf; text-decoration: none;\">Akeeba Ticket System</a></p>',
			'<p style="text-align: right; font-size: 7pt; color: #ccc;">Powered by <a href="https://www.akeeba.com/products/ats.html" style="color: #ccf; text-decoration: none;">Akeeba Ticket System</a></p>',
		];

		foreach ($replaceThat as $find)
		{
			$subject      = str_ireplace($find, '', $subject);
			$templateText = str_ireplace($find, '', $templateText);
		}

		// Because SpamAssassin demands there is a body and surrounding html tag even though it's not necessary.
		if (strpos($templateText, '<body') == false)
		{
			$templateText = '<body>' . $templateText . '</body>';
		}

		if (strpos($templateText, '<html') == false)
		{
			$templateText = <<< HTML
<html>
<head>
<title>{$subject}</title>
</head>
$templateText
</html>
HTML;

		}

		return [$subject, $templateText, $loadLanguage];
	}

	/**
	 * Returns a new JMail object, set up to send UTF-8 encoded, HTML emails.
	 *
	 * @return  Mail  The mailer object
	 *
	 * @throws  phpmailerException
	 */
	public static function &getMailer()
	{
		$mailer = Factory::getMailer();

		// We always send HTML emails
		$mailer->isHtml(true);

		// Required to not get broken characters in emails
		$mailer->CharSet = 'UTF-8';

		return $mailer;
	}

	/**
	 * Parses template text and subject with email variables
	 *
	 * @param   string  $templateText  Body text
	 * @param   string  $subject       Subject text
	 * @param   array   $mailInfo      Associative array of variables to replace in the boby and subject text
	 * @param   Mail    $mailer        If set to null, this function will return values, instead of setting them to the
	 *                                 mailer
	 *
	 * @return  bool|array  If a mailer is passed, it returns true, otherwise it will return parsed subject and body
	 */
	public static function parseTemplate($templateText, $subject, $mailInfo, &$mailer = null)
	{
		// Email variables
		$return    = [];
		$container = Container::getInstance('com_ats');
		$isCli     = $container->platform->isCli();

		if ($isCli)
		{
			$baseURL    = ComponentHelper::getParams('com_ats')->get('siteurl', 'http://www.example.com');
			$temp       = str_replace('http://', '', $baseURL);
			$temp       = str_replace('https://', '', $temp);
			$parts      = explode($temp, '/', 2);
			$subpathURL = count($parts) > 1 ? $parts[1] : '';
		}
		else
		{
			$baseURL    = URI::base();
			$subpathURL = URI::base(true);
		}

		$baseURL  = str_replace('/administrator', '', $baseURL);
		$sitename = $container->platform->getConfig()->get('sitename');

		$emailVars = [
			'sitename' => $sitename,
			'siteurl'  => $baseURL,
		];

		if (is_array($mailInfo) && !empty($mailInfo))
		{
			$emailVars = array_merge($emailVars, $mailInfo);
		}

		// Perform substitutions
		foreach ($emailVars as $key => $value)
		{
			$tag          = '[' . strtoupper($key) . ']';
			$templateText = str_replace($tag, $value, $templateText);
			$subject      = str_replace($tag, $value, $subject);
		}

		if ($mailer)
		{
			$templateText = self::inlineImages($templateText, $mailer);
			$mailer->setSubject($subject);
			$mailer->setBody($templateText);
		}
		else
		{
			$return['body'] = $templateText;
			$return['subject'] = $subject;
		}

		return $return;
	}

	/**
	 * Attach and inline the referenced images in the email message
	 *
	 * @param   string  $templateText
	 * @param   Mail    $mailer
	 *
	 * @return  string
	 *
	 * @since   3.2.0
	 */
	private static function inlineImages($templateText, Mail $mailer)
	{
		// RegEx patterns to detect images
		$patterns = [
			// srcset="**URL**" e.g. source tags
			'/srcset=\"?([^"]*)\"?/i',
			// src="**URL**" e.g. img tags
			'/src=\"?([^"]*)\"?/i',
			// url(**URL**) nad url("**URL**") i.e. inside CSS
			'/url\(\"?([^"\(\)]*)\"?\)/i',
		];

		// Cache of images so we don't inline them multiple times
		$foundImages = [];
		// Running counter of images, used to create the attachment IDs in the message
		$imageIndex = 0;

		// Run a RegEx search & replace for each pattern
		foreach ($patterns as $pattern)
		{
			// $matches[0]: the entire string matched by RegEx; $matches[1]: just the path / URL
			$templateText = preg_replace_callback($pattern, function (array $matches) use ($mailer, &$foundImages, &$imageIndex): string {
				// Abort if it's not a file type we can inline
				if (!self::isInlineableFileExtension($matches[1]))
				{
					return $matches[0];
				}

				// Try to get the local absolute filesystem path of the referenced media file
				$localPath = self::getLocalAbsolutePath(self::normalizeURL($matches[1]));

				// Abort if this was not a relative / absolute URL pointing to our own site
				if (empty($localPath))
				{
					return $matches[0];
				}

				// Abort if the referenced file does not exist
				if (!@file_exists($localPath) || !@is_file($localPath))
				{
					return $matches[0];
				}

				// Make sure the inlined image is cached; prevent inlining the same file multiple times
				if (!array_key_exists($localPath, $foundImages))
				{
					$imageIndex++;
					$mailer->AddEmbeddedImage($localPath, 'img' . $imageIndex, basename($localPath));
					$foundImages[$localPath] = $imageIndex;
				}

				return str_replace($matches[1], $toReplace = 'cid:img' . $foundImages[$localPath], $matches[0]);
			}, $templateText);
		}

		// Return the processed email content
		return $templateText;
	}

	/**
	 * Does this file / URL have an allowed image extension for inlining?
	 *
	 * @param   string  $fileOrUri
	 *
	 * @return  bool
	 *
	 * @since   3.2.0
	 */
	private static function isInlineableFileExtension($fileOrUri)
	{
		$dot = strrpos($fileOrUri, '.');

		if ($dot === false)
		{
			return false;
		}

		$extension = substr($fileOrUri, $dot + 1);

		return in_array(strtolower($extension), self::$allowedImageExtensions);
	}

	/**
	 * Normalizes an image relative or absolute URL as an absolute URL
	 *
	 * @param   string  $fileOrUri
	 *
	 * @return  string
	 *
	 * @since   3.2.0
	 */
	private static function normalizeURL($fileOrUri)
	{
		// Empty file / URIs are returned as-is (obvious screw up)
		if (empty($fileOrUri))
		{
			return $fileOrUri;
		}

		// Remove leading / trailing slashes
		$fileOrUri = trim($fileOrUri, '/');

		// HTTPS URLs are returned as-is
		if (substr($fileOrUri, 0, 8) == 'https://')
		{
			return $fileOrUri;
		}

		// HTTP URLs are returned upgraded to HTTPS
		if (substr($fileOrUri, 0, 7) == 'http://')
		{
			return 'https://' . substr($fileOrUri, 7);
		}

		// Normalize URLs with a partial schema as HTTPS
		if (substr($fileOrUri, 0, 3) == '://')
		{
			return 'https://' . substr($fileOrUri, 3);
		}

		// This is a file. We assume it's relative to the site's root
		return rtrim(Uri::base(), '/') . '/' . $fileOrUri;
	}

	/**
	 * Return the path to the local file referenced by the URL, provided it's internal.
	 *
	 * @param   string  $url
	 *
	 * @return  string|null  The local file path. NULL if the URL is not internal.
	 *
	 * @since   3.2.0
	 */
	private static function getLocalAbsolutePath($url)
	{
		$base = rtrim(Uri::base(), '/');

		if (strpos($url, $base) !== 0)
		{
			return null;
		}

		return JPATH_ROOT . '/' . ltrim(substr($url, strlen($base) + 1), '/');
	}

}
