<?php
/**
 * @package   ats
 * @copyright Copyright (c)2011-2022 Nicholas K. Dionysopoulos / Akeeba Ltd
 * @license   GNU General Public License version 3, or later
 */

namespace Akeeba\TicketSystem\Admin\Helper;

defined('_JEXEC') or die;

use Akeeba\TicketSystem\Admin\Model\Categories;
use DateTimeZone;
use FOF40\Container\Container;
use FOF40\Date\Date;
use Joomla\CMS\Language\Text;
use Vamsi\HTMLToBBCode\HtmlConverter;

class Format
{
	/**
	 * Format a date. Default is to format as a GMT date in YYYY-mm-dd HH:MM format.
	 *
	 * @param   string  $date    The date to format.
	 * @param   string  $format  The format to use, default is 'Y-m-d H:i'
	 * @param   bool    $local   Should I display the date in the user's local timezone? Default: no.
	 *
	 * @return  string  The formatted date
	 */
	public static function date($date, $format = null, $local = false)
	{
		$container = Container::getInstance('com_ats');
		$date      = new Date($date, 'GMT');

		if (empty($format))
		{
			$format = 'Y-m-d H:i';
		}

		if ($local)
		{
			$zone = Permissions::getUser()->getParam('timezone', $container->platform->getConfig()->get('offset', 'UTC'));
			$tz   = new DateTimeZone($zone);
			$date->setTimezone($tz);
		}

		return $date->format($format, $local);
	}

	/**
	 * Format a date. Default is Joomla's DATE_FORMAT_LC2 plus timezone, e.g. Thursday, 27 April 2017 20:26 EEST.
	 *
	 * @param   string  $date    The date to format.
	 * @param   string  $format  The format to use. Leave empty for Joomla's DATE_FORMAT_LC2 plus timezone.
	 * @param   bool    $local   Should I display the date in the user's local timezone? Default: no.
	 *
	 * @return  string  The formatted date
	 */
	public static function date2($date, $format = null, $local = false)
	{
		if (empty($format))
		{
			$format = Text::_('DATE_FORMAT_LC2') . ' T';
		}

		return self::date($date, $format, $local);
	}

	/**
	 * Converts a fancy shorthand size notation (e.g. 8M) to bytes
	 *
	 * @param   string  $fancy  The fancy notation to convert
	 *
	 * @return  int
	 */
	public static function toBytes($fancy)
	{
		if (empty($fancy))
		{
			return 0;
		}

		$value = (float) $fancy;
		$fancy = trim($fancy);
		$units = [
			'K' => 1024, 'M' => 1048576, 'G' => 1073741824,
		];

		$u = substr($fancy, -1);

		if ($u === false)
		{
			return (int) $value;
		}

		if (!array_key_exists($u, $units))
		{
			return (int) $value;
		}

		return (int) ($value * $units[$u]);
	}

	/**
	 * Convert bytes to human readable format
	 *
	 * @see http://codeaid.net/php/convert-size-in-bytes-to-a-human-readable-format-(php)
	 *
	 * @param   int  $bytes
	 * @param   int  $precision
	 *
	 * @return string
	 */
	public static function bytesToSize($bytes, $precision = 2)
	{
		$kilobyte = 1024;
		$megabyte = $kilobyte * 1024;
		$gigabyte = $megabyte * 1024;
		$terabyte = $gigabyte * 1024;

		if (($bytes >= 0) && ($bytes < $kilobyte))
		{
			return $bytes . ' B';

		}
		elseif (($bytes >= $kilobyte) && ($bytes < $megabyte))
		{
			return round($bytes / $kilobyte, $precision) . ' KB';

		}
		elseif (($bytes >= $megabyte) && ($bytes < $gigabyte))
		{
			return round($bytes / $megabyte, $precision) . ' MB';

		}
		elseif (($bytes >= $gigabyte) && ($bytes < $terabyte))
		{
			return round($bytes / $gigabyte, $precision) . ' GB';
		}
		elseif ($bytes >= $terabyte)
		{
			return round($bytes / $terabyte, $precision) . ' TB';
		}
		else
		{
			return $bytes . ' B';
		}
	}

	public static function categoryName($id)
	{
		static $categories = null;

		if (is_null($categories))
		{
			$container = Container::getInstance('com_ats');

			/** @var Categories $catModel */
			$catModel = $container->factory->model('Categories')->tmpInstance();
			$items    = $catModel->removeBehaviour('Language')->ignoreUser(1)->get();

			foreach ($items as $item)
			{
				$categories[$item->id] = $item->title;
			}
		}

		if (isset($categories[$id]))
		{
			return $categories[$id];
		}
		else
		{
			return '';
		}
	}

	/**
	 * Returns a fancy formatted time lapse code
	 *
	 * @param   int     $referenceTime  Timestamp of the reference date / time
	 * @param   string  $currentTime    Timestamp of the current date / time
	 * @param   string  $quantifyBy     Quantify by unit, one of s (seconds), m (minutes), h (yours), d (days) or y
	 *                                  (years)
	 * @param   bool    $autoText       Automatically appent text
	 *
	 * @return string
	 */
	public static function timeAgo($referenceTime = 0, $currentTime = '', $quantifyBy = '', $autoText = true)
	{
		if ($currentTime == '')
		{
			$currentTime = time();
		}

		// Raw time difference
		$rawTimeDifference      = $currentTime - $referenceTime;
		$absoluteTimeDifference = abs($rawTimeDifference);

		$uomMap = [
			['s', 60],
			['m', 60 * 60],
			['h', 60 * 60 * 24],
			['d', 60 * 60 * 24 * 365],
			['y', 0],
		];

		$textMap = [
			's' => [1, 'COM_ATS_TIME_SECOND'],
			'm' => [60, 'COM_ATS_TIME_MINUTE'],
			'h' => [60 * 60, 'COM_ATS_TIME_HOUR'],
			'd' => [60 * 60 * 24, 'COM_ATS_TIME_DAY'],
			'y' => [60 * 60 * 24 * 365, 'COM_ATS_TIME_YEAR'],
		];

		if ($quantifyBy == '')
		{
			$uom = 's';

			for ($i = 0; $i < count($uomMap); $i++)
			{
				if ($absoluteTimeDifference <= $uomMap[$i][1])
				{
					$uom = $uomMap[$i][0];

					break;
				}
			}

			if ($absoluteTimeDifference > 60 * 60 * 24 * 365)
			{
				$uom = 'y';
			}
		}
		else
		{
			$uom = $quantifyBy;
		}

		$dateDifference = floor($absoluteTimeDifference / $textMap[$uom][0]);

		$prefix = '';
		$suffix = '';

		if ($autoText == true && ($currentTime == time()))
		{
			if ($rawTimeDifference < 0)
			{
				$prefix = Text::_('COM_ATS_TIME_AFTER_PRE');
				$suffix = Text::_('COM_ATS_TIME_AFTER_POST');
			}
			else
			{
				$prefix = Text::_('COM_ATS_TIME_AGO_PRE');
				$suffix = Text::_('COM_ATS_TIME_AGO_POST');
			}
		}

		if ($prefix)
		{
			$prefix = trim($prefix) . ' ';
		}

		if ($suffix)
		{
			$suffix = ' ' . trim($suffix);
		}

		return $prefix . $dateDifference . ' ' . Text::_($textMap[$uom][1]) . $suffix;
	}

	public static function html2bbCode($html)
	{
		// Make sure the converter class is loaded, or try to load it
		if (!class_exists('\\Vamsi\\HTMLToBBCode\\HtmlConverter'))
		{
			$container      = Container::getInstance('com_ats');
			$autoloaderFile = $container->backEndPath . '/vendor/autoload.php';

			// We can't find the Composer autoloader :/
			if (!@is_file($autoloaderFile))
			{
				return "";
			}

			$autoloader = @include($autoloaderFile);

			// The autoloader failed for any reason :(
			if (empty($autoloader))
			{
				return "";
			}

			// The autoloader got loaded but the class is not there :@
			if (!class_exists('\\Vamsi\\HTMLToBBCode\\HtmlConverter'))
			{
				return "";
			}
		}

		$converter = new HtmlConverter([
			'strip_tags'   => true,
			'remove_nodes' => 'meta style script object embed',
		]);

		return $converter->convert($html);
	}

	/**
	 * Converts an ISO country code to an emoji flag.
	 *
	 * This takes advantage of how Unicode Regional Indicator Symbol Letter glyphs (added to Unicode back in 2010) work.
	 *
	 * On really old implementations (before 2015) the appear as boxed uppercase letters, e.g. 🇺 🇸. However, starting
	 * around 2015, a combination of two glyphs that would spell an ISO country name is defined as a multi-byte glyph of
	 * that country's flag. So 🇺 directly followed by 🇸 would give us the US flag: 🇺🇸.
	 *
	 * This is the basis of our implementation. We convert the ISO country name to uppercase and map the ASCII letters
	 * to the respective Unicode Regional Indicator Symbol Letter glyphs. In newer browsers that's a colour flag. In
	 * older browsers that's a monochrome flag. In even older browsers it's two boxed letters spelling out the country's
	 * ISO code. In even older browsers you get question marks. However, if you have a browser released over a decade
	 * ago not seeing an Emoji flag is the least of your concerns so I'l willing to take those chances.
	 *
	 * How do we avoid invalid combinations that don't make a flag, such as XU? We compare them against the list of ISO
	 * country codes, of course. Invalid combos get black flagged (&#x1F3F4; 🏴). Empty strings and nulls get a white
	 * flag (&#x1F3F3; 🏳).
	 *
	 * @param   string|null  $cCode
	 *
	 * @return  string
	 *
	 * @since   3.1.0
	 */
	public static function countryToEmoji($cCode = null)
	{
		// Convert the country code to all uppercase
		$cCode = strtoupper(is_null($cCode) ? '' : trim($cCode));

		// No country? White flag.
		if (empty($cCode))
		{
			return '&#x1F3F3;';
		}

		// Valid country codes
		$countryCodes = [
			'AD', 'AE', 'AF', 'AG', 'AI', 'AL', 'AM', 'AO', 'AQ', 'AR', 'AS', 'AT', 'AU', 'AW', 'AX', 'AZ', 'BA', 'BB',
			'BD', 'BE', 'BF', 'BG', 'BH', 'BI', 'BJ', 'BL', 'BM', 'BN', 'BO', 'BQ', 'BR', 'BS', 'BT', 'BV', 'BW', 'BY',
			'BZ', 'CA', 'CC', 'CD', 'CF', 'CG', 'CH', 'CI', 'CK', 'CL', 'CM', 'CN', 'CO', 'CR', 'CU', 'CV', 'CW', 'CX',
			'CY', 'CZ', 'DE', 'DJ', 'DK', 'DM', 'DO', 'DZ', 'EC', 'EE', 'EG', 'EH', 'ER', 'ES', 'ET', 'FI', 'FJ', 'FK',
			'FM', 'FO', 'FR', 'GA', 'GB', 'GD', 'GE', 'GF', 'GG', 'GH', 'GI', 'GL', 'GM', 'GN', 'GP', 'GQ', 'GR', 'GS',
			'GT', 'GU', 'GW', 'GY', 'HK', 'HM', 'HN', 'HR', 'HT', 'HU', 'ID', 'IE', 'IL', 'IM', 'IN', 'IO', 'IQ', 'IR',
			'IS', 'IT', 'JE', 'JM', 'JO', 'JP', 'KE', 'KG', 'KH', 'KI', 'KM', 'KN', 'KP', 'KR', 'KW', 'KY', 'KZ', 'LA',
			'LB', 'LC', 'LI', 'LK', 'LR', 'LS', 'LT', 'LU', 'LV', 'LY', 'MA', 'MC', 'MD', 'ME', 'MF', 'MG', 'MH', 'MK',
			'ML', 'MM', 'MN', 'MO', 'MP', 'MQ', 'MR', 'MS', 'MT', 'MU', 'MV', 'MW', 'MX', 'MY', 'MZ', 'NA', 'NC', 'NE',
			'NF', 'NG', 'NI', 'NL', 'NO', 'NP', 'NR', 'NU', 'NZ', 'OM', 'PA', 'PE', 'PF', 'PG', 'PH', 'PK', 'PL', 'PM',
			'PN', 'PR', 'PS', 'PT', 'PW', 'PY', 'QA', 'RE', 'RO', 'RS', 'RU', 'RW', 'SA', 'SB', 'SC', 'SD', 'SE', 'SG',
			'SH', 'SI', 'SJ', 'SK', 'SL', 'SM', 'SN', 'SO', 'SR', 'SS', 'ST', 'SV', 'SX', 'SY', 'SZ', 'TC', 'TD', 'TF',
			'TG', 'TH', 'TJ', 'TK', 'TL', 'TM', 'TN', 'TO', 'TR', 'TT', 'TV', 'TW', 'TZ', 'UA', 'UG', 'UM', 'US', 'UY',
			'UZ', 'VA', 'VC', 'VE', 'VG', 'VI', 'VN', 'VU', 'WF', 'WS', 'YE', 'YT', 'ZA', 'ZM', 'ZW',
		];

		// Invalid country? Black flag.
		if (!in_array($cCode, $countryCodes))
		{
			return '&#x1F3F4;';
		}

		// Uppercase letter to Unicode Regional Indicator Symbol Letter
		$letterToRISL = [
			'A' => "&#x1F1E6;",
			'B' => "&#x1F1E7;",
			'C' => "&#x1F1E8;",
			'D' => "&#x1F1E9;",
			'E' => "&#x1F1EA;",
			'F' => "&#x1F1EB;",
			'G' => "&#x1F1EC;",
			'H' => "&#x1F1ED;",
			'I' => "&#x1F1EE;",
			'J' => "&#x1F1EF;",
			'K' => "&#x1F1F0;",
			'L' => "&#x1F1F1;",
			'M' => "&#x1F1F2;",
			'N' => "&#x1F1F3;",
			'O' => "&#x1F1F4;",
			'P' => "&#x1F1F5;",
			'Q' => "&#x1F1F6;",
			'R' => "&#x1F1F7;",
			'S' => "&#x1F1F8;",
			'T' => "&#x1F1F9;",
			'U' => "&#x1F1FA;",
			'V' => "&#x1F1FB;",
			'W' => "&#x1F1FC;",
			'X' => "&#x1F1FD;",
			'Y' => "&#x1F1FE;",
			'Z' => "&#x1F1FF;",
		];

		return $letterToRISL[substr($cCode, 0, 1)] . $letterToRISL[substr($cCode, 1, 1)];
	}
}
