<?php
/**
 * @package   ats
 * @copyright Copyright (c)2011-2022 Nicholas K. Dionysopoulos / Akeeba Ltd
 * @license   GNU General Public License version 3, or later
 */

namespace Akeeba\TicketSystem\Admin\Helper;

use FOF40\Container\Container;
use Joomla\CMS\Component\ComponentHelper;
use Joomla\CMS\Filesystem\Folder as JFolder;
use Joomla\CMS\Language\Text;
use Joomla\CMS\User\User;

defined('_JEXEC') or die;

class Html
{
	/**
	 * Uses the ATS plugins to fetch ana avatar for the user
	 *
	 * @param   User  $user  The user for which to fetch an avatar for
	 * @param   int   $size  The size (in pixels), defaults to 64
	 *
	 * @return string The URL to the avatar image
	 */
	public static function getAvatarURL($user, $size = 64)
	{
		$container = Container::getInstance('com_ats');
		$container->platform->importPlugin('ats');

		$jResponse = $container->platform->runPlugins('onATSAvatar', [$user, $size]);

		$url = '';

		if (!empty($jResponse))
		{
			foreach ($jResponse as $response)
			{
				if ($response)
				{
					$url = $response;
				}
			}
		}

		return $url;
	}

	public static function getSignature(User $user)
	{
		static $signatures = [];

		if ($user->username == 'system')
		{
			return '';
		}

		$container = Container::getInstance('com_ats');

		if (!array_key_exists($user->username, $signatures))
		{
			$db = $container->db;

			$query     = $db->getQuery(true)
				->select($db->qn('profile_value'))
				->from($db->qn('#__user_profiles'))
				->where($db->qn('user_id') . ' = ' . $db->q($user->id))
				->where($db->qn('profile_key') . ' = ' . $db->q('ats.signature'));
			$signature = $db->setQuery($query)->loadResult();

			if (!is_null($signature))
			{
				$signature = Filter::filterText($signature, $user);
			}

			/**
			 * If there is no linebreak, paragraph or table tag we assume a plain text signature, therefore in need for
			 * nl2br(). This is backwards compatibility to older versions of ATS.
			 */
			if ((stripos($signature, '<br') === false) && (stripos($signature, '< br') === false) && (stripos($signature, '</p') === false) && (stripos($signature, '<td') === false))
			{
				$signature = nl2br($signature);
			}

			$signatures[$user->username] = empty($signature) || is_null($signature) ? '' : $signature;
		}

		return $signatures[$user->username];
	}

	/**
	 * Fetches the correct user that posted the ticket
	 *
	 * @param   int  $userid
	 *
	 * @return User
	 */
	public static function getPostUser($userid)
	{
		$container = Container::getInstance('com_ats');

		if ($userid == -1)
		{
			$postUser = clone $container->platform->getUser();

			$postUser->username = 'system';
			$postUser->email    = 'noreply@example.com';
			$postUser->name     = Text::_('COM_ATS_CLI_SYSTEMUSERLABEL');
		}
		else
		{
			$postUser = $container->platform->getUser($userid);
		}

		return $postUser;
	}

	/**
	 * Gets the total amount of tickets created by a specific user
	 *
	 * @param   integer  $userid  ID of the user
	 *
	 * @return  integer Number of created tickets
	 */
	public static function getTicketsCount($userid)
	{
		static $count = [];

		if (!isset($count[$userid]))
		{
			$container = Container::getInstance('com_ats');

			// Let's directly use the db, so I can be faster
			$db = $container->db;

			$query          = $db->getQuery(true)
				->select('COUNT(*)')
				->from($db->qn('#__ats_tickets'))
				->where($db->qn('created_by') . ' = ' . $db->q($userid));
			$count[$userid] = $db->setQuery($query)->loadResult();
		}

		return $count[$userid];
	}

	/**
	 * Gets the total amount of time spent supporting a specific user
	 *
	 * @param   integer  $userid  ID of the user
	 *
	 * @return  integer  Total time spent supporting the user, in minutes
	 */
	public static function getTimeSpentPerUser($userid)
	{
		static $timePerUser = [];

		if (!isset($timePerUser[$userid]))
		{
			$container = Container::getInstance('com_ats');

			// Let's directly use the db, so I can be faster
			$db = $container->db;

			$query = $db->getQuery(true);
			$query
				->select('SUM(' . $query->qn('timespent') . ')')
				->from($db->qn('#__ats_tickets'))
				->where($db->qn('created_by') . ' = ' . $db->q($userid));
			$timePerUser[$userid] = $db->setQuery($query)->loadResult() ?? 0;
		}

		return $timePerUser[$userid];
	}

	/**
	 * Returns the maximum allowed size for an attachment in human format (ie 1Mb)
	 *
	 * @return  string
	 */
	public static function getUploadLimits()
	{
		$mediaLimit = ComponentHelper::getParams('com_media')->get('upload_maxsize', 0) * 1024 * 1024;

		$umf = Format::toBytes(ini_get('upload_max_filesize'));
		$pms = Format::toBytes(ini_get('post_max_size'));

		if (($umf * $pms) == 0)
		{
			$uploadLimit = max($umf, $pms);
		}
		else
		{
			$uploadLimit = min($umf, $pms);
		}

		$uploadLimit = min($uploadLimit, $mediaLimit);
		$uploadLimit = Format::bytesToSize($uploadLimit);

		return $uploadLimit;
	}

	/**
	 * Returns the allowed file extensions for uploads
	 *
	 * @return array
	 */
	public static function allowedExtensions()
	{
		$mediaParams     = ComponentHelper::getParams('com_media');

		if (version_compare(JVERSION, '4.0', 'lt'))
		{
			$extensions = $mediaParams->get('upload_extensions');
		}
		else
		{
			$restrictUploads = $mediaParams->get('restrict_uploads', 1);

			if ($restrictUploads != 1)
			{
				return [];
			}

			$extensions = $mediaParams->get('restrict_uploads_extensions');
		}

		$extensions = array_map('strtolower', explode(',', $extensions));

		return array_unique($extensions);
	}

	/**
	 * Decodes the type of a custom field
	 *
	 * @param   string  $selected
	 *
	 * @return  string  The type of the custom field
	 */
	public static function getFieldType($selected)
	{
		static $fieldTypes = [];

		if (!$fieldTypes)
		{
			$basepath = JPATH_ADMINISTRATOR . '/components/com_ats/CustomField';

			$files = JFolder::files($basepath, '.php');

			foreach ($files as $file)
			{
				if ($file === 'Base.php')
				{
					continue;
				}

				$type      = basename($file, '.php');
				$className = 'Akeeba\\TicketSystem\\Admin\\CustomField\\' . $type;

				if (class_exists($className))
				{
					$fieldTypes[strtolower($type)] = Text::_('COM_ATS_CUSTOMFIELDS_FIELD_TYPE_' . strtoupper($type));
				}
			}
		}

		if (!isset($fieldTypes[$selected]))
		{
			return '';
		}

		return $fieldTypes[$selected];
	}

	/**
	 * Get the ATS tags for the given user ID
	 *
	 * @param   int  $userId
	 *
	 * @return  array
	 *
	 * @since   4.0.0
	 */
	public static function getUserTags(int $userId): array
	{
		static $tags = [];

		if (isset($tags[$userId]))
		{
			return $tags[$userId];
		}

		$container = Container::getInstance('com_ats');
		/** @var \Akeeba\TicketSystem\Site\Model\Jusers $juserModel */
		$juserModel    = $container->factory->model('Jusers')->tmpInstance();
		$juserModel    = $juserModel->find($userId);
		$tags[$userId] = $juserModel->atstags->toArray();

		return $tags[$userId];
	}
}
