<?php
/**
 * @package   ats
 * @copyright Copyright (c)2011-2022 Nicholas K. Dionysopoulos / Akeeba Ltd
 * @license   GNU General Public License version 3, or later
 */

namespace Akeeba\TicketSystem\Admin\Helper;

use FOF40\Container\Container;
use Joomla\CMS\Application\ApplicationHelper as JApplicationHelper;
use Joomla\CMS\Date\Date as JDate;
use Joomla\CMS\Factory as JFactory;
use Joomla\CMS\Language\Text;
use Joomla\CMS\Router\Route;
use Joomla\CMS\String\PunycodeHelper;
use Joomla\CMS\Uri\Uri as JUri;
use Joomla\CMS\User\User as JUser;
use Joomla\CMS\User\UserHelper as JUserHelper;
use RuntimeException;
use UnexpectedValueException;

defined('_JEXEC') or die;

class JoomlaUsers
{
	/**
	 * This flag is set when I've initialized a dummy application when running this code through CLI
	 *
	 * @var   bool
	 *
	 * @since 2.4.1
	 */
	private static $initialisedDummyAppForCli = false;
	/** @var Container */
	private $container;

	public function __construct(Container $container)
	{
		$this->container = $container;
	}

	public static function getUserGroups(JUser $user)
	{
		static $cache = [];

		if ($user->username == 'system')
		{
			return [];
		}

		if (!array_key_exists($user->id, $cache) && ($user->id <= 0))
		{
			$cache[$user->id] = [];
		}

		if (!array_key_exists($user->id, $cache))
		{
			$container = Container::getInstance('com_ats');
			$platform  = $container->platform;

			$platform->importPlugin('ats');
			$groups = $platform->runPlugins('onAtsGetUserGroupsList', [$user]);
			$groups = array_pop($groups);

			$cache[$user->id] = $groups;
		}

		return $cache[$user->id];
	}

	/**
	 * Returns the user ID, if a user exists, given an email address.
	 *
	 * @param   string  $email  The email to search on.
	 *
	 * @return  integer  The user id or 0 if not found.
	 */
	public function getUserIdByEmail($email)
	{
		// Initialise some variables
		$db    = $this->container->db;
		$query = $db->getQuery(true)
			->select($db->qn('id'))
			->from($db->qn('#__users'))
			->where($db->qn('email') . ' = ' . $db->q($email));
		$db->setQuery($query, 0, 1);

		return $db->loadResult();
	}

	/**
	 * Create a new user account.
	 *
	 * The username will be the name of the user: converted to all lowercase, all non-word characters squashed into a
	 * maximum of one consecutive dots. For example "Bill W. Gates 3rd" will be converted to "bill.w.gates.3rd". If
	 * that
	 * is taken the email will be used instead.
	 *
	 * If the email is already present or we cannot find a username that works we will throw UnexpectedValueException.
	 *
	 * Any other error from the model should result in a RuntimeException.
	 *
	 * @param   string  $email          The email address of the user to create
	 * @param   string  $name           The full name of the user
	 * @param   bool    $forceActivate  If true, force account creation and activation, no matter what Joomla! is
	 *                                  configured to do
	 *
	 * @return  array  ($activation, $userID) where $activation 'useractivate', 'adminactivate' if activation is
	 *                 required, null otherwise; $userID is the generated user ID
	 *
	 * @throws  UnexpectedValueException  If the email or username already exists
	 * @throws  RuntimeException          User registration is disabled and $forceActivate is false
	 */
	public function createUser($email, $name, $forceActivate)
	{
		// Am I allowed to create new users...?
		if (!$forceActivate)
		{
			$comUsersContainer     = Container::getInstance('com_users');
			$usersConfig           = $comUsersContainer->params;
			$allowUserRegistration = $usersConfig->get('allowUserRegistration');

			if ($allowUserRegistration == 0)
			{
				throw new RuntimeException('User registration is disabled');
			}
		}

		// Does the email already exist?
		if (empty($email) || self::getUserIdByEmail($email))
		{
			throw new UnexpectedValueException();
		}

		// Try to create a username from the full name
		$username = preg_replace('/[^\w]/us', '.', $name);
		$username = strtolower($username);

		while (strpos($username, '..') !== false)
		{
			$username = str_replace('..', '.', $username);
		}

		// If the username already exists try using the email as the username
		if (JUserHelper::getUserId($username))
		{
			$username = $email;

			// If that exists too throw an exception
			if (JUserHelper::getUserId($username))
			{
				throw new UnexpectedValueException();
			}
		}

		$data = [
			'name'     => $name,
			'username' => $username,
			'email'    => $email,
		];

		return $this->register($data, $forceActivate);
	}

	/**
	 * Method to register a new user account. Based on UsersModelRegistration::register().
	 *
	 * @param   array  $data                The user data to save.
	 * @param   bool   $skipUserActivation  Should I forcibly skip user activation?
	 *
	 * @return  array  ($activation, $userID) where $activation 'useractivate', 'adminactivate' if activation is
	 *                 required, null otherwise; $userID is the generated user ID
	 */
	private function register(array $data, $skipUserActivation = false)
	{
		// Load com_users language files
		$lang = $this->container->platform->getLanguage();
		$lang->load('com_users', JPATH_BASE . '/components/com_users', 'en-GB', false, false);
		$lang->load('com_users', JPATH_BASE, 'en-GB', false, false);
		$lang->load('com_users', JPATH_BASE . '/components/com_users', null, false, false);
		$lang->load('com_users', JPATH_BASE, null, false, false);

		$comUsersContainer = Container::getInstance('com_users');
		$params            = $comUsersContainer->params;

		$data = array_merge([
			'name'     => '',
			'username' => '',
			'password' => '',
			'email'    => '',
			'groups'   => [],
		], $data);

		// Initialise the table with JUser.
		$user = new JUser;

		// If no password was specified create a random one
		if (!isset($data['password']) || empty($data['password']))
		{
			$data['password'] = JUserHelper::genRandomPassword(24);
		}

		// Convert the email to punycode if necessary
		$data['email'] = PunycodeHelper::emailToPunycode($data['email']);

		// Get the groups the user should be added to after registration.
		$data['groups'] = [$params->get('new_usertype', 2)];

		/**
		 * Get the dispatcher and load the users plugins.
		 *
		 * IMPORTANT: We cannot go through the JApplicationCms object directly since user plugins will set the error
		 * message on the dispatcher instead of throwing an exception. See the plugins/user/profile/profile.php plugin
		 * file's onContentPrepareData method to understand this questionable approach. Until Joomla! stops supporting
		 * legacy error handling we cannot switch to the FOF Platform's runPlugins and a regular exceptions handler around it :(
		 */
		$this->container->platform->importPlugin('user');

		// Trigger the data preparation event.
		try
		{
			$results = $this->container->platform->runPlugins('onContentPrepareData', [
				'com_users.registration', $data,
			]);
		}
		catch (\Exception $e)
		{
			throw new RuntimeException($e->getMessage());
		}

		// Check for errors encountered while preparing the data. YOU CANNOT REMOVE THIS. READ THE BIG COMMENT ABOVE.
		if (version_compare(JVERSION, '3.999.999', 'le') && count($results) && in_array(false, $results, true))
		{
			throw new RuntimeException(\JEventDispatcher::getInstance()->getError());
		}

		// Get the parameters affecting the behavior
		$userActivation   = $params->get('useractivation');
		$sendPassword     = $params->get('sendpassword', 1);
		$sendEmailToAdmin = $params->get('mail_to_admin');

		// Do we have to forcibly skip user activation?
		if ($skipUserActivation)
		{
			$userActivation = 0;
		}

		// Check if the user needs to activate their account.
		if (($userActivation == 1) || ($userActivation == 2))
		{
			$data['activation'] = JApplicationHelper::getHash(JUserHelper::genRandomPassword());
			$data['block']      = 1;
		}

		// Set the user parameters
		$data['params'] = [];

		// Bind the data.
		if (!$user->bind($data))
		{
			throw new RuntimeException(Text::sprintf('COM_USERS_REGISTRATION_BIND_FAILED', $user->getError()));
		}

		// Do not remove! This is a necessary trick to create users through CLI, where a JApplicationCms instance doesn't exist.
		$this->initialiseDummyApplicationForCLI();

		// Store the data.
		if (!$user->save())
		{
			throw new RuntimeException(Text::sprintf('COM_USERS_REGISTRATION_SAVE_FAILED', $user->getError()));
		}

		// Get the base URI
		$app      = JFactory::getApplication();
		$liveSite = $app->get('live_site', '');
		$liveSite = trim($liveSite);

		// If the live_site is not set in Global Configuration get it from the component's configuration options
		if (empty($liveSite))
		{
			$liveSite = $this->container->params->get('siteurl');
		}

		$uri  = JUri::getInstance($liveSite);
		$base = $uri->toString(['scheme', 'user', 'pass', 'host', 'port']);

		$db    = $this->container->db;
		$query = $db->getQuery(true);

		// Compile the notification mail values.
		$data             = $user->getProperties();
		$data['fromname'] = $app->get('fromname');
		$data['mailfrom'] = $app->get('mailfrom');
		$data['sitename'] = $app->get('sitename');
		$data['siteurl']  = JUri::base();

		// Handle account activation/confirmation emails.
		$isAdmin = $this->container->platform->isBackend();

		switch ($userActivation)
		{
			// Self-activation of user account
			default:
			case 2:
				// Set the link to confirm the user email.
				$data['activate'] = Route::_('index.php?option=com_users&task=registration.activate&token=' . $data['activation'], false, Route::TLS_IGNORE, true);

				// Remove administrator/ from activate url in case this method is called from admin
				if ($isAdmin)
				{
					$adminPos         = strrpos($data['activate'], 'administrator/');
					$data['activate'] = substr_replace($data['activate'], '', $adminPos, 14);
				}

				$emailSubject = Text::sprintf(
					'COM_USERS_EMAIL_ACCOUNT_DETAILS',
					$data['name'],
					$data['sitename']
				);

				$emailBody = Text::sprintf(
					'COM_USERS_EMAIL_REGISTERED_WITH_ADMIN_ACTIVATION_BODY_NOPW',
					$data['name'],
					$data['sitename'],
					$data['activate'],
					$data['siteurl'],
					$data['username']
				);

				if ($sendPassword)
				{
					$emailBody = Text::sprintf(
						'COM_USERS_EMAIL_REGISTERED_WITH_ADMIN_ACTIVATION_BODY',
						$data['name'],
						$data['sitename'],
						$data['activate'],
						$data['siteurl'],
						$data['username'],
						$data['password_clear']
					);
				}
				break;

			// Administrator activation of user account
			case 1:
				// Set the link to activate the user account.
				$data['activate'] = Route::_('index.php?option=com_users&task=registration.activate&token=' . $data['activation'], false, Route::TLS_IGNORE, true);

				// Remove administrator/ from activate url in case this method is called from admin
				if ($isAdmin)
				{
					$adminPos         = strrpos($data['activate'], 'administrator/');
					$data['activate'] = substr_replace($data['activate'], '', $adminPos, 14);
				}

				$emailSubject = Text::sprintf(
					'COM_USERS_EMAIL_ACCOUNT_DETAILS',
					$data['name'],
					$data['sitename']
				);

				$emailBody = Text::sprintf(
					'COM_USERS_EMAIL_REGISTERED_WITH_ACTIVATION_BODY_NOPW',
					$data['name'],
					$data['sitename'],
					$data['activate'],
					$data['siteurl'],
					$data['username']
				);

				if ($sendPassword)
				{
					$emailBody = Text::sprintf(
						'COM_USERS_EMAIL_REGISTERED_WITH_ACTIVATION_BODY',
						$data['name'],
						$data['sitename'],
						$data['activate'],
						$data['siteurl'],
						$data['username'],
						$data['password_clear']
					);
				}

				break;

			// No activation required
			case 0:
				$emailSubject = Text::sprintf(
					'COM_USERS_EMAIL_ACCOUNT_DETAILS',
					$data['name'],
					$data['sitename']
				);

				$emailBody = Text::sprintf(
					'COM_USERS_EMAIL_REGISTERED_BODY_NOPW',
					$data['name'],
					$data['sitename'],
					$data['siteurl']
				);

				if ($sendPassword)
				{
					$emailBody = Text::sprintf(
						'COM_USERS_EMAIL_REGISTERED_BODY',
						$data['name'],
						$data['sitename'],
						$data['siteurl'],
						$data['username'],
						$data['password_clear']
					);
				}

				break;
		}

		// Send the registration email.
		$return = JFactory::getMailer()
			->sendMail($data['mailfrom'], $data['fromname'], $data['email'], $emailSubject, $emailBody);

		// Send Notification mail to administrators
		if (($userActivation < 2) && ($sendEmailToAdmin == 1))
		{
			$emailSubject = Text::sprintf(
				'COM_USERS_EMAIL_ACCOUNT_DETAILS',
				$data['name'],
				$data['sitename']
			);

			$emailBodyAdmin = Text::sprintf(
				'COM_USERS_EMAIL_REGISTERED_NOTIFICATION_TO_ADMIN_BODY',
				$data['name'],
				$data['username'],
				$data['siteurl']
			);

			// Get all admin users
			$query->clear()
				->select($db->quoteName(['name', 'email', 'sendEmail']))
				->from($db->quoteName('#__users'))
				->where($db->quoteName('sendEmail') . ' = 1')
				->where($db->quoteName('block') . ' = 0');

			$db->setQuery($query);

			try
			{
				$rows = $db->loadObjectList();
			}
			catch (RuntimeException $e)
			{
				throw new RuntimeException(Text::sprintf('COM_USERS_DATABASE_ERROR', $e->getMessage()), 500);
			}

			// Send mail to all Super Users
			foreach ($rows as $row)
			{
				$return = JFactory::getMailer()
					->sendMail($data['mailfrom'], $data['fromname'], $row->email, $emailSubject, $emailBodyAdmin);

				// Check for an error.
				if ($return !== true)
				{
					throw new RuntimeException(Text::_('COM_USERS_REGISTRATION_ACTIVATION_NOTIFY_SEND_MAIL_FAILED'));
				}
			}
		}

		// Check for an error.
		if ($return !== true)
		{
			// Send a system message to administrators receiving system mails
			$db = JFactory::getDbo();
			$query->clear()
				->select($db->qn('id'))
				->from($db->qn('#__users'))
				->where($db->qn('block') . ' = ' . (int) 0)
				->where($db->qn('sendEmail') . ' = ' . (int) 1);
			$db->setQuery($query);

			try
			{
				$userids = $db->loadColumn();
			}
			catch (RuntimeException $e)
			{
				throw new RuntimeException(Text::sprintf('COM_USERS_DATABASE_ERROR', $e->getMessage()), 500);
			}

			if (count($userids) > 0)
			{
				$jdate = new JDate;

				// Build the query to add the messages
				foreach ($userids as $userid)
				{
					$values = [
						$db->quote($userid),
						$db->quote($userid),
						$db->quote($jdate->toSql()),
						$db->quote(Text::_('COM_USERS_MAIL_SEND_FAILURE_SUBJECT')),
						$db->quote(Text::sprintf('COM_USERS_MAIL_SEND_FAILURE_BODY', $return, $data['username'])),
					];
					$query->clear()
						->insert($db->quoteName('#__messages'))
						->columns($db->quoteName([
							'user_id_from',
							'user_id_to',
							'date_time',
							'subject',
							'message',
						]))
						->values(implode(',', $values));
					$db->setQuery($query);

					try
					{
						$db->execute();
					}
					catch (RuntimeException $e)
					{
						throw new RuntimeException(Text::sprintf('COM_USERS_DATABASE_ERROR', $e->getMessage()), 500);
					}
				}
			}

			throw new RuntimeException(Text::_('COM_USERS_REGISTRATION_SEND_MAIL_FAILED'));
		}

		if ($userActivation == 1)
		{
			return ['useractivate', $user->id];
		}
		elseif ($userActivation == 2)
		{
			return ['adminactivate', $user->id];
		}
		else
		{
			return [null, $user->id];
		}
	}

	/**
	 * If we are under CLI we need to fake the application in JFactory. The PlgUserJoomla plugin has a private $app
	 * property defined which will be populated with JFactory::getApplication() by the JPlugin constructor. This
	 * property is used to get the mailfrom Global Configuration parameter. Therefore we have to create a fake
	 * front-end application.
	 *
	 * Moreover we need to inject JUri::$instances['SERVER'] with the site's front-end URL. That's because the frontend
	 * of com_users, used for user registration, will use JUri::getInstance() to get the site's base URL for the site.
	 * Moreover we need to set live_site, if it's not already set, to the same URL to force JUri::base() to return the
	 * same base URL.
	 *
	 * @since    2.4.1
	 */
	private function initialiseDummyApplicationForCLI()
	{
		// This is only required for CLI scripts
		if (!$this->container->platform->isCli())
		{
			return;
		}

		/**
		 * If we've already done this, let's go away.
		 *
		 * Why not use JFactory::getApplication() to see if we've already created an application? Because doing that without an
		 * application actually set up results in an exception and we're hosed. As a result we have to use a flag of our own.
		 */
		if (static::$initialisedDummyAppForCli)
		{
			return;
		}

		// Set the flag
		static::$initialisedDummyAppForCli = true;

		// JFactory::getApplication results in an application object without a proper configuration object...
		$app      = JFactory::getApplication('site');
		$liveSite = $app->get('live_site', '');
		$liveSite = trim($liveSite);

		// If the live_site is not set in Global Configuration get it from the component's configuration options
		if (empty($liveSite))
		{
			$liveSite = $this->container->params->get('siteurl');
		}

		/**
		 * Set the application's live_site parameter.
		 *
		 * We have to set it both through the appand  JFactory::getConfig(). Normally, they should be the same object.
		 * This happens because JApplicationCms::initalizeApp will use JFactory::getConfig to pull in the configuration
		 * object dependency (which is necessary since JFactory::getApplication and JApplicationCms::getInstance ignore any
		 * injected dependency * groan *). However, what weI observed, is that they are separate objects. Since JUri uses
		 * JFactory::getConfig() we need to set the live_site there too, otherwise all email links end up being broken, leaking
		 * information about the site's filesystem installation path as well.
		 *
		 * Also note that the configuration object must be initialized before you reach this code. Currently, we do that in
		 * the CLI script itself with JFactory::getConfig(JPATH_CONFIGURATION . '/configuration.php') since Joomla! 3.5 and
		 * later will NOT initialize the configuration when you just do JFactory::getConfig(), i.e. what the aforementioned
		 * JApplicationCms::initalizeApp does.
		 */
		$app->set('live_site', $liveSite);
		JFactory::getConfig()->set('live_site', $liveSite);

		// Auto-load JUri. DO NOT REMOVE.
		class_exists('JUri', true);
		// Inject JUri::$instances['SERVER'] using Reflection
		$refClass = new \ReflectionClass('JUri');
		$refProp  = $refClass->getProperty('instances');
		$refProp->setAccessible(true);
		$instances           = $refProp->getValue();
		$instances['SERVER'] = new JUri($liveSite);
		$refProp->setValue($instances);
		unset($instances);

		// Clear JUri::$base using Reflection - Forces JUri::base() to go through live_site next time it is called.
		$refProp = $refClass->getProperty('base');
		$refProp->setAccessible(true);
		$refProp->setValue([]);
		unset($refProp);

		// Clear JUri::$root using Reflection - Forces JUri::root() to go through live_site next time it is called.
		$refProp = $refClass->getProperty('root');
		$refProp->setAccessible(true);
		$refProp->setValue([]);
		unset($refProp, $refClass);
	}
}
