<?php
/**
 * @package   ats
 * @copyright Copyright (c)2011-2022 Nicholas K. Dionysopoulos / Akeeba Ltd
 * @license   GNU General Public License version 3, or later
 */

namespace Akeeba\TicketSystem\Admin\Helper;

defined('_JEXEC') or die;

use Akeeba\TicketSystem\Admin\Model\Attachments;
use Akeeba\TicketSystem\Admin\Model\Posts;
use Akeeba\TicketSystem\Admin\Model\Tickets;
use Akeeba\TicketSystem\Site\Model\Categories;
use FOF40\Container\Container;
use Joomla\CMS\Access\Access as JAccess;
use Joomla\CMS\User\User;

class Permissions
{
	private static $identityUserCache = null;

	private static $usersCache = [];

	/**
	 * Returns the all the privileges linked with a specific ticket
	 *
	 * @param   Tickets  $ticket
	 *
	 * @return  array
	 */
	public static function getPrivileges($ticket)
	{
		$ret = [
			'view'       => false,
			'post'       => false,
			'edit'       => false,
			'edit.state' => false,
			'admin'      => false,
			'close'      => false,
			'attachment' => false,
		];

		$container = Container::getInstance('com_ats');
		$actions   = self::getAclPrivileges($ticket->catid);
		$user      = self::getUser();

		// Can I change the visibility of the ticket?
		$ret['edit.state'] = $actions['core.edit.state'];

		// If I am the onwer of this ticket, I can see it and post to it
		if (!$user->guest && ($ticket->created_by == $user->id))
		{
			$ret['view']  = true;
			$ret['post']  = true;
			$ret['close'] = true;
			$ret['edit']  = $actions['core.edit.own'];
		}

		// If I am the manager I can do anything
		if (self::isManager($ticket->catid))
		{
			$ret = [
				'view'       => true,
				'post'       => true,
				'edit'       => true,
				'edit.state' => true,
				'admin'      => true,
				'close'      => true,
				'attachment' => true,
			];
		}

		// If it's a public ticket, I can view it
		if ($ticket->public)
		{
			$ret['view'] = true;
		}

		// What about attachments?
		$ret['attachment'] = $actions['ats.attachment'];

		return $ret;
	}

	/**
	 * Gets a list of the actions that can be performed.
	 *
	 * @param   int  $categoryId  The category ID.
	 * @param   int  $userId      User id
	 *
	 * @return  array
	 */
	public static function getAclPrivileges($categoryId = 0, $userId = null)
	{
		$container = Container::getInstance('com_ats');
		$user      = self::getUser($userId);

		$result = [];

		$actions = [
			'core.admin', 'core.manage', 'core.create', 'core.edit',
			'core.edit.own', 'core.edit.state', 'core.delete',
			'ats.private', 'ats.attachment',
		];

		foreach ($actions as $action)
		{
			if ($categoryId)
			{
				$result[$action] = $user->authorise($action, 'com_ats') ||
					$user->authorise($action, 'com_ats.category.' . (int) $categoryId);
			}
			else
			{
				$result[$action] = $user->authorise($action, 'com_ats');
			}
		}

		return $result;
	}

	/**
	 * Is the user a manger?
	 *
	 * @param   int  $category  Category id (opt.)
	 * @param   int  $userid    Userid (opt. current user if null)
	 *
	 * @return  bool  Is manager?
	 */
	public static function isManager($category = null, $userid = null)
	{
		$container = Container::getInstance('com_ats');

		// Automatically fetches the current user if the id is null
		$user = self::getUser($userid);

		if ($user->authorise('core.admin', 'com_ats') ||
			$user->authorise('core.manage', 'com_ats') ||
			($category ? $user->authorise('core.manage', 'com_ats.category.' . $category) : false))
		{
			return true;
		}

		return false;
	}

	/**
	 * Returns the categories where the user is a manager
	 *
	 * @param   int  $userid
	 *
	 * @return  array
	 */
	public static function getManagerCategories($userid = null)
	{
		static $cache = [];

		$container = Container::getInstance('com_ats');
		$userid    = self::getUser($userid)->id;

		if (!isset($cache[$userid]))
		{
			// First of all let's fetch all the categories the user has access to
			/** @var Categories $catModel */
			$catModel   = $container->factory->model('Categories')->tmpInstance();
			$categories = $catModel->userid($userid)->get();

			$allowed = [];

			/** @var Categories $category */
			foreach ($categories as $category)
			{
				if (Permissions::isManager($category->id, $userid))
				{
					$allowed[] = $category->id;
				}
			}

			$cache[$userid] = $allowed;
		}

		return $cache[$userid];
	}

	/**
	 * Fetches all the managers of the given category
	 *
	 * @param int|null $category Ticket category
	 *
	 * @return  array   List of ids and names of managers (indexed by id) of the category
	 */
	public static function getManagers(?int $category)
	{
		static $cache = [];

		if (isset($cache[$category]))
		{
			return $cache[$category];
		}

		//AFAIK there is no way to get the list of users enabled to do something, so we have to improvise
		$container = Container::getInstance('com_ats');

		$db       = $container->db;

		// First, let's get the whole list of groups
		$query  = $db->getQuery(true)
			->select('id')
			->from('#__usergroups')
			->order('id DESC');
		$groups = $db->setQuery($query)->loadColumn();

		$validGroups = [];

		// Then check if they can admin tickets
		foreach ($groups as $group)
		{
			if (JAccess::checkGroup($group, 'core.admin', 'com_ats') ||
				JAccess::checkGroup($group, 'core.manage', 'com_ats') ||
				JAccess::checkGroup($group, 'core.manage', 'com_ats.category.' . $category))
			{
				$validGroups[] = $group;
			}
		}

		// Get all users in those groups
		$query            = $db->getQuery(true)
			->select([
				$db->qn('id'),
				$db->qn('name'),
			])->from($db->qn('#__users', 'u'))
			->leftJoin($db->qn('#__user_usergroup_map', 'm')
				. 'ON(' . $db->qn('u.id') . ' = ' . $db->qn('m.user_id') . ')'
			)->where($db->qn('m.group_id') .
				'IN (' . implode(',', array_map([$db, 'q'], $validGroups)) . ')'
			);
		$cache[$category] = $db->setQuery($query)->loadObjectList('id') ?? [];

		$cache[$category] = array_filter($cache[$category], function ($def) use ($container, $category) {
			$juser = self::getUser($def->id);

			return $juser->authorise('core.admin', 'com_ats') ||
				$juser->authorise('core.manage', 'com_ats') ||
				$juser->authorise('core.manage', 'com_ats.category.' . $category);
		});

		return $cache[$category];
	}

	public static function getUser(?int $id = null): ?User
	{
		if (is_null($id))
		{
			if (empty(self::$identityUserCache))
			{
				self::$identityUserCache = Container::getInstance('com_ats')->platform->getUser();
			}

			return self::$identityUserCache;
		}

		if (isset(self::$usersCache[$id]))
		{
			return self::$usersCache[$id];
		}

		self::$usersCache[$id] = Container::getInstance('com_ats')->platform->getUser($id);

		return self::$usersCache[$id];
	}

	/**
	 * Checks if a post if editable since we're inside the grace time
	 *
	 * @param   Posts  $post  Post to check
	 *
	 * @return  bool    Are we within the grace time?
	 */
	public static function editGraceTime($post)
	{
		$container = Container::getInstance('com_ats');
		$graceTime = $container->params->get('editeableforxminutes', 15);
		$result    = false;
		$user      = self::getUser();
		$userid    = $user->id;

		if ((($post->created_by == $userid) || ($post->modified_by == $userid)) && !$user->guest)
		{
			if (($post->modified_by == $userid))
			{
				$editedOn = $container->platform->getDate($post->modified_on);
			}
			else
			{
				$editedOn = $container->platform->getDate($post->created_on);
			}

			$now = $container->platform->getDate();

			$editedAgo = abs($now->toUnix() - $editedOn->toUnix());

			$result = $editedAgo < 60 * $graceTime;
		}

		return $result;
	}

	/**
	 * Is the post attachment visible?
	 *
	 * @param   Attachments  $attachment  Attachment record
	 * @param   bool         $isManager   Am I a manager? Calculated in the ticket view since I need the category ID.
	 * @param   Posts        $post        Post record
	 *
	 * @return  bool
	 */
	public static function attachmentVisible($attachment, $isManager, $post)
	{
		/*
		 * This was the original if statement:
		 * if($attachment->ats_attachment_id && !empty($attachment->original_filename) && ($attachment->enabled || (!$attachment->enabled && $user->authorise('core.manage','com_ats')) || $this->isManager || (!$attachment->enabled && $user->authorise('core.edit.own','com_ats') && ($user->id == $item->created_by)))):
		 *
		 * it was split into several different checks for better readability
		 */
		// Always cast the incoming value to object
		if (is_array($attachment))
		{
			$attachment = (object) $attachment;
		}

		// No attachment id? Stop here
		if (!$attachment->ats_attachment_id)
		{
			return false;
		}

		// No attachment file name? Stop here
		if (!$attachment->original_filename)
		{
			return false;
		}

		// Well, if it's enabled everyone can see it
		if ($attachment->enabled)
		{
			return true;
		}

		// If it's not enabled and I'am a manager, I can see it anyway
		if ($isManager)
		{
			return true;
		}

		$container = Container::getInstance('com_ats');
		$user      = self::getUser();

		// Mhm... I'm not a manager and the attachment is not enabled. Can I still see it?
		// If I can edit my own tickets and I'm the owner of the post, I can do that
		if ($user->authorise('core.edit.own', 'com_ats') && ($user->id == $post->created_by))
		{
			return true;
		}

		return false;
	}

	/**
	 * Is the attachment private?
	 *
	 * @param   Attachments  $attachment    Attachment record
	 * @param   bool         $isManager     Am I a manager?
	 * @param   int          $ticket_owner  Ticket owner
	 *
	 * @return  bool
	 */
	public static function attachmentPrivate(Attachments $attachment, $isManager, $ticket_owner)
	{
		$container = Container::getInstance('com_ats');

		// Is the "Attachments private" flag enabled?
		if (!$container->params->get('attachments_private', 0))
		{
			return false;
		}

		$user = self::getUser();

		if (
			// If this is not a manager
			!$isManager
			// and he doesn't own the attachment
			&& ($user->id != $attachment->created_by)
			// and he doesn't own the ticket
			&& ($user->id != $ticket_owner)
		)
		{
			// The attachment is private
			return true;
		}

		// Otherwise I is not private
		return false;
	}
}
