<?php
/**
 * @package   ats
 * @copyright Copyright (c)2011-2022 Nicholas K. Dionysopoulos / Akeeba Ltd
 * @license   GNU General Public License version 3, or later
 */

namespace Akeeba\TicketSystem\Admin\Helper;

use Exception;
use FOF40\Container\Container;
use FOF40\JoomlaAbstraction\CacheCleaner;
use Joomla\CMS\Plugin\PluginHelper;
use Joomla\Registry\Registry;

/**
 * Handles plugin parameters. Also migrates component options into the plugins which now implement them.
 *
 * @since  3.2.0
 */
class PluginParameters
{
	/**
	 * Maps ATS plugins to the respective parameters that got migrated off the component
	 *
	 * @var   array
	 * @since 3.2.0
	 */
	protected static $map = [
		// ATS plugins
		'ats'             => [
			'mailfetch'         => [
				'replybyemail',
				'emailadminonly',
				'newbyemail',
				'track_email_id',
				'text_version',
				'email_default_category',
				'mailboxtype',
				'email_server',
				'email_port',
				'email_ssl',
				'email_tls',
				'email_validate_certificate',
				'email_username',
				'email_password',
				'email_folder',
				'email_delete',
				'email_instantreply',
				'email_lastcheck',
			],
			'removeattachments' => [
				'maxattachmentage',
			],
			'autoclose'         => [
				'maxinactiveticket',
				'silentticketcloseage',
			],
			'autoreply'         => [
				'botusers',
			],
		],
		// ATS InstantReply plugins
		'atsinstantreply' => [
			'tickets' => [
				'instantreplies_daylimit',
			],
		],
	];

	/**
	 * Automatic migration of component parameters to the corresponding ATS plugins
	 *
	 * @return  void
	 * @since   3.2.0
	 */
	public static function migrate()
	{
		// Get the component parameters and mark them as unchanged
		$container       = Container::getInstance('com_ats');
		$componentParams = $container->params;
		$componentDirty  = false;

		// For each plugin in the map
		foreach (self::$map as $folder => $mappedPlugins)
		{
			foreach ($mappedPlugins as $plugin => $keys)
			{
				// Get the parameters for the plugin in question and mark them as unchanged
				$pluginParams = self::getPluginParams($plugin);
				$dirty        = false;

				// Loop all migrated option keys
				foreach ($keys as $key)
				{
					// Get the value from the component
					$v = $componentParams->get($key, null);

					// Non-null values get transcribed to the plugin parameters which are then marked as changed
					if (!is_null($v))
					{
						$dirty = true;
						$pluginParams->set($key, $v);
					}
				}

				// If we didn't change any plugin parameters let's move on to the next plugin
				if (!$dirty)
				{
					continue;
				}

				// Try to save the plugin parameters
				if (self::savePluginParams($plugin, $pluginParams))
				{
					// If the plugin params saved, nullify the respective keys in the component params and mark them changed
					$componentDirty = true;
					$componentParams->setParams(array_combine($keys, array_fill(0, count($keys), null)));
				}
			}
		}

		// If the component params are changed save them back to the database
		if ($componentDirty)
		{
			$componentParams->save();
		}
	}

	/**
	 * Get the parameters of a specific ATS plugin as a Registry object
	 *
	 * @param   string  $plugin  The name of the ATS plugin to get parameters for
	 *
	 * @return  Registry
	 * @since   3.2.0
	 */
	public static function getPluginParams($plugin)
	{
		$pluginObject = PluginHelper::getPlugin('ats', $plugin);
		$params       = (is_object($pluginObject) && property_exists($pluginObject, 'params')) ? $pluginObject->params : [];

		return new Registry($params);
	}

	/**
	 * Save the modified plugin parameters back to the database
	 *
	 * @param   string    $plugin  The name of the ATS plugin where parameters will be written to
	 * @param   Registry  $params  A Registry object with the plugin parameters
	 *
	 * @return  bool  True on success
	 * @since   3.2.0
	 */
	public static function savePluginParams($plugin, Registry $params)
	{
		$container = Container::getInstance('com_ats');
		$db        = $container->db;
		$data      = $params->toString();

		$sql = $db->getQuery(true)
			->update($db->qn('#__extensions'))
			->set($db->qn('params') . ' = ' . $db->q($data))
			->where($db->qn('type') . ' = ' . $db->q('plugin'))
			->where($db->qn('element') . ' = ' . $db->q($plugin))
			->where($db->qn('folder') . ' = ' . $db->q('ats'));

		$db->setQuery($sql);

		try
		{
			$db->execute();
			// We changed the plugin parameters. We must reset the system cache which holds them to prevent inconsistencies.
			CacheCleaner::clearCacheGroups(['_system'], $container->platform->isBackend() ? [0] : [1]);
		}
		catch (Exception $e)
		{
			return false;
		}

		return true;
	}
}