<?php
/**
 * @package   ats
 * @copyright Copyright (c)2011-2022 Nicholas K. Dionysopoulos / Akeeba Ltd
 * @license   GNU General Public License version 3, or later
 */

namespace Akeeba\TicketSystem\Admin\Helper;

use FOF40\Container\Container;
use Joomla\CMS\Application\SiteApplication;
use Joomla\CMS\Factory as JFactory;
use Joomla\CMS\Filesystem\Folder as JFolder;
use Joomla\CMS\HTML\HTMLHelper as JHtml;
use Joomla\CMS\Language\Language as JLanguage;
use Joomla\CMS\Language\LanguageHelper as JLanguageHelper;
use Joomla\CMS\Language\Text;
use Joomla\CMS\Plugin\PluginHelper;
use Joomla\Registry\Registry;

defined('_JEXEC') or die;

/**
 * A helper class for drop-down selection boxes
 */
abstract class Select
{
	/**
	 * Returns a list of custom field types
	 *
	 * @return  array  type => description
	 */
	public static function getFieldTypes()
	{
		$fieldTypes = [];

		$basepath = JPATH_ADMINISTRATOR . '/components/com_ats/CustomField';

		$files = JFolder::files($basepath, '.php');

		foreach ($files as $file)
		{
			if ($file === 'Base.php')
			{
				continue;
			}

			$type      = basename($file, '.php');
			$className = 'Akeeba\\TicketSystem\\Admin\\CustomField\\' . $type;

			if (class_exists($className))
			{
				$fieldTypes[strtolower($type)] = Text::_('COM_ATS_CUSTOMFIELDS_FIELD_TYPE_' . strtoupper($type));
			}
		}

		return $fieldTypes;
	}

	public static function fieldTypes($id, $selected, array $attribs = [])
	{
		$options[] = JHtml::_('FEFHelp.select.option', '', '- ' . Text::_('COM_ATS_CUSTOMFIELDS_FIELD_TYPE') . ' -');

		$basepath = JPATH_ADMINISTRATOR . '/components/com_ats/CustomField';

		$files = JFolder::files($basepath, '.php');

		foreach ($files as $file)
		{
			if ($file === 'Base.php')
			{
				continue;
			}

			$type      = basename($file, '.php');
			$className = 'Akeeba\\TicketSystem\\Admin\\CustomField\\' . $type;

			if (class_exists($className))
			{
				$options[] = JHtml::_('FEFHelp.select.option', strtolower($type), Text::_('COM_ATS_CUSTOMFIELDS_FIELD_TYPE_' . strtoupper($type)));
			}
		}

		return static::genericlist($options, $id, $attribs, $selected, $id);
	}

	public static function fieldShow($id, $selected, array $attribs = [])
	{
		$options[] = JHtml::_('FEFHelp.select.option', 'all', Text::_('COM_ATS_CUSTOMFIELDS_FIELD_SHOW_ALL'));
		$options[] = JHtml::_('FEFHelp.select.option', 'category', Text::_('COM_ATS_CUSTOMFIELDS_FIELD_SHOW_CATEGORY'));

		return static::genericlist($options, $id, $attribs, $selected, $id);
	}

	/**
	 * Generates a yes/no drop-down list.
	 *
	 * @param   string  $name      The value of the HTML name attribute
	 * @param   array   $attribs   Additional HTML attributes for the <select> tag
	 * @param   string  $selected  The key that is selected
	 *
	 * @return  string  HTML for the list
	 */
	public static function booleanlist($name, $attribs = null, $selected = null)
	{
		$options = [
			JHtml::_('FEFHelp.select.option', '', '---'),
			JHtml::_('FEFHelp.select.option', '0', Text::_('JNo')),
			JHtml::_('FEFHelp.select.option', '1', Text::_('JYes')),
		];

		return self::genericlist($options, $name, $attribs, $selected, $name);
	}

	/**
	 * Generates a Published/Unpublished drop-down list.
	 *
	 * @param   string  $selected  The key that is selected (0 = unpublished / 1 = published)
	 * @param   string  $id        The value of the HTML name attribute
	 * @param   array   $attribs   Additional HTML attributes for the <select> tag
	 *
	 * @return  string  HTML for the list
	 */
	public static function published($selected = null, $id = 'enabled', $attribs = [])
	{
		$options   = [];
		$options[] = JHtml::_('FEFHelp.select.option', null, '- ' . Text::_('COM_ATS_COMMON_SELECTSTATE') . ' -');
		$options[] = JHtml::_('FEFHelp.select.option', 0, Text::_('JUNPUBLISHED'));
		$options[] = JHtml::_('FEFHelp.select.option', 1, Text::_('JPUBLISHED'));

		return self::genericlist($options, $id, $attribs, $selected, $id);
	}

	/**
	 * Generates a the options for the drop-down list of the available languages on a multi-language site.
	 * In this way we can use it in a F0FForm list field.
	 *
	 * @param   string  $selected  The key that is selected
	 * @param   string  $id        The value of the HTML name attribute
	 * @param   array   $attribs   Additional HTML attributes for the <select> tag
	 *
	 * @return  string  HTML for the list
	 */
	public static function languages($selected = null, $id = 'language', $attribs = [])
	{
		$languages = JLanguageHelper::getLanguages('lang_code');
		$options   = [];
		$options[] = ['value' => '*', 'text' => Text::_('JALL_LANGUAGE')];

		if (!empty($languages))
		{
			foreach ($languages as $key => $lang)
			{
				$options[] = ['value' => $key, 'text' => $lang->title];
			}
		}

		return $options;
	}

	/**
	 * Get the categories selection box
	 *
	 * @param   null    $selected        Currently selected item
	 * @param   string  $name            Name of the HTML select element
	 * @param   array   $attribs         HTML attributes of the select element
	 * @param   null    $default_option  The no option text
	 * @param   bool    $filtered        Apply language filtering? (Ignored in the back-end)
	 * @param   bool    $accessAware     Should the list be aware of the user's access level?
	 *
	 * @return mixed
	 *
	 * @since version
	 */
	public static function getCategories($selected = null, $name = 'category', $attribs = [], $default_option = null, $filtered = false, $accessAware = false)
	{
		$options = self::getCategoriesOptions($default_option, true, $filtered, $accessAware);

		return JHTML::_('FEFHelp.select.genericlist', $options, $name, $attribs, 'value', 'text', $selected, $name);
	}

	/**
	 * Static method that return only the options for a category select list.
	 * In this way we can use it in a F0FForm list field.
	 *
	 * @param   null  $default_option  The no option text
	 * @param   bool  $asObject        Do I want to create an array of array or of objects?
	 * @param   bool  $filtered        Apply language filtering?
	 * @param   bool  $accessAware     Should the list be filtered by the user's access level?
	 *
	 * @return array
	 */
	public static function getCategoriesOptions($default_option = null, $asObject = false, $filtered = false, $accessAware = false)
	{
		$options = [];

		// Do I want a default option, at all?
		if ($default_option !== false)
		{
			if (is_null($default_option))
			{
				$default_option = 'COM_ATS_CONFIG_DEFAULTCAT_MSG';
			}

			if ($asObject)
			{
				$options[] = JHTML::_('FEFHelp.select.option', '', Text::_($default_option));
			}
			else
			{
				$options[] = ['value' => '', 'text' => Text::_($default_option)];
			}
		}

		$className = 'CategoriesModelCategories';
		$filePath  = JPATH_ADMINISTRATOR . '/components/com_categories/models/categories.php';

		if (version_compare(JVERSION, '3.999.999', 'gt'))
		{
			$className = 'Joomla\Component\Categories\Administrator\Model\CategoriesModel';
			$filePath  = JPATH_ADMINISTRATOR . '/components/com_categories/src/Model/CategoriesModel.php';
		}

		if (!class_exists($className))
		{
			require_once $filePath;
		}

		$model      = new $className();
		$modelState = $model->getState();

		// The default "language title" is not consistent, we have to use the language code for filtering
		$modelState->set('list.select', 'a.id, a.title, a.level, l.lang_code');
		$modelState->set('filter.extension', 'com_ats');
		$modelState->set('filter.access', null);
		$modelState->set('filter.language', null);
		$modelState->set('filter.published', 1);
		$modelState->set('list.start', 0);
		$modelState->set('list.limit', 0);

		$items = $model->getItems();

		/**
		 * Joomla's model doesn't let you filter by language correctly. You can filter by no language at all, only items
		 * with "any language" (*) or only items with a specific language (e.g. el_GR). In the front-end we need to
		 * show all items that have EITHER any language OR the specific front-end language set. Moreover, language
		 * filtering must only take place when the front-end language filtering is enabled, i.e. you let the user select
		 * which language to display the site in.
		 */
		if ($filtered && !empty($items))
		{
			$items = self::filterCategoriesByLanguage($items);
		}

		/**
		 * Joomla's model doesn't let you filter by ACL privileges. If $accessAware is true we need to only ever show
		 * categories the user can file a ticket in. Therefore we need to check if the current user is granted the
		 * core.create privilege for the category. If not, we get to filter it out.
		 */
		if ($accessAware)
		{
			$items = array_filter($items, function ($item) {
				$privs = Permissions::getAclPrivileges($item->id);

				return in_array('core.create', $privs);
			});
		}

		if (count($items))
		{
			foreach ($items as $item)
			{
				$id    = $item->id;
				$title = $item->title;

				if ($item->level > 1)
				{
					$title = str_repeat('–', $item->level - 1) . ' ' . $title;
				}

				if ($asObject)
				{
					$options[] = JHTML::_('FEFHelp.select.option', $id, $title);
				}
				else
				{
					$options[] = ['value' => $id, 'text' => $title];
				}

			}
		}

		return $options;
	}

	public static function ticketstatuses($selected = '', $name = 'status', $attribs = [], $id = 'status')
	{
		$options = [];

		$addUnselect = true;

		if (isset($attribs['unselect_option']))
		{
			$addUnselect = $attribs['unselect_option'];
		}

		if ($addUnselect)
		{
			$options[] = JHTML::_('FEFHelp.select.option', '', '- ' . Text::_('COM_ATS_TICKETS_STATUS_SELECT') . ' -');
		}

		$options = array_merge($options, self::getTicketStatus());

		return self::genericlist($options, $name, $attribs, $selected, $id);
	}

	public static function getTicketStatus()
	{
		$options[] = ['value' => 'O', 'text' => Text::_('COM_ATS_TICKETS_STATUS_O')];
		$options[] = ['value' => 'P', 'text' => Text::_('COM_ATS_TICKETS_STATUS_P')];

		$statuses = self::getCustomTicketStatuses();

		foreach ($statuses as $value => $text)
		{
			$options[] = ['value' => $value, 'text' => $text];
		}

		$options[] = ['value' => 'C', 'text' => Text::_('COM_ATS_TICKETS_STATUS_C')];

		return $options;
	}

	public static function publicstate($selected = '', $id = 'status', $attribs = [])
	{
		$options = [
			JHTML::_('FEFHelp.select.option', '', '- ' . Text::_('COM_ATS_TICKETS_PUBLIC_SELECT') . ' -'),
			JHTML::_('FEFHelp.select.option', '0', Text::_('COM_ATS_TICKETS_PUBLIC_PRIVATE')),
			JHTML::_('FEFHelp.select.option', '1', Text::_('COM_ATS_TICKETS_PUBLIC_PUBLIC')),
		];

		return self::genericlist($options, $id, $attribs, $selected, $id);
	}

	/**
	 * Ticket priorities select list
	 *
	 * @param   string        $name      Select list name and id
	 * @param   string        $selected  Selected value
	 * @param   array|string  $attribs   Associative array of attributes
	 *
	 * @return  string   Select list HTML
	 */
	public static function priorities($name = 'priority', $selected = '', $attribs = [])
	{
		$options = [];

		// Priority is a 1-5-10 list, so in the future we can add more items for fine-tuning
		$options[] = JHTML::_('FEFHelp.select.option', '', '- ' . Text::_('COM_ATS_COMMON_SELECT') . ' -');
		$options[] = JHTML::_('FEFHelp.select.option', '1', Text::_('COM_ATS_PRIORITIES_HIGH'));
		$options[] = JHTML::_('FEFHelp.select.option', '5', Text::_('COM_ATS_PRIORITIES_NORMAL'));
		$options[] = JHTML::_('FEFHelp.select.option', '10', Text::_('COM_ATS_PRIORITIES_LOW'));

		return self::genericlist($options, $name, $attribs, $selected, $name);
	}

	public static function getManagers($selected = '', $id = 'user_id', $attribs = [], $category = null)
	{
		$options[] = JHTML::_('FEFHelp.select.option', '', '- ' . Text::_('COM_ATS_TICKETS_ASSIGN_TO') . ' -');

		$managers = Permissions::getManagers($category);

		foreach ($managers as $manager)
		{
			$options[] = JHTML::_('FEFHelp.select.option', $manager->id, $manager->name);
		}

		return self::genericlist($options, $id, $attribs, $selected, $id);
	}

	public static function buckets($selected = '', $id = 'ats_bucket_id', $attribs = [])
	{
		$container = Container::getInstance('com_ats');

		$options[] = JHTML::_('FEFHelp.select.option', null, '- ' . Text::_('COM_ATS_TICKETS_BUCKET_SELECT') . ' -');

		/** @var \Akeeba\TicketSystem\Admin\Model\Buckets $model */
		$model   = $container->factory->model('Buckets')->tmpInstance();
		$buckets = $model->status('O', 'P')->get(true);

		foreach ($buckets as $bucket)
		{
			$options[] = JHTML::_('FEFHelp.select.option', $bucket->ats_bucket_id, $bucket->title);
		}

		return self::genericlist($options, $id, $attribs, $selected, $id);
	}

	/**
	 * User tag select list
	 *
	 * @param   string  $name      Select list name and id
	 * @param   array   $selected  Selected value
	 * @param   array   $attribs   Associative array of attributes
	 *
	 * @return string   Select list HTML
	 */
	public static function usertags($name, $selected, $attribs = [])
	{
		$container = Container::getInstance('com_ats');

		/** @var \Akeeba\TicketSystem\Admin\Model\UserTags $model */
		$model = $container->factory->model('UserTags')->tmpInstance();

		$options = [];

		$tags = $model
			->enabled(1)
			->filter_order('ordering')
			->filter_order_Dir('ASC')
			->limit(0)
			->get();

		foreach ($tags as $tag)
		{
			$options[] = JHTML::_('FEFHelp.select.option', $tag->ats_usertag_id, $tag->title);
		}

		return self::genericlist($options, $name, $attribs, $selected, $name);
	}

	public static function weekdays($name, $selected, $attribs = [])
	{
		$options[] = JHtml::_('FEFHelp.select.option', '*', '*');
		$options[] = JHtml::_('FEFHelp.select.option', '0', Text::_('SUNDAY'));
		$options[] = JHtml::_('FEFHelp.select.option', '1', Text::_('MONDAY'));
		$options[] = JHtml::_('FEFHelp.select.option', '2', Text::_('TUESDAY'));
		$options[] = JHtml::_('FEFHelp.select.option', '3', Text::_('WEDNESDAY'));
		$options[] = JHtml::_('FEFHelp.select.option', '4', Text::_('THURSDAY'));
		$options[] = JHtml::_('FEFHelp.select.option', '5', Text::_('FRIDAY'));
		$options[] = JHtml::_('FEFHelp.select.option', '6', Text::_('SATURDAY'));

		return self::genericlist($options, $name, $attribs, $selected);
	}

	public static function dateInteger($name, $selected, $attribs = [])
	{
		$options   = [];
		$options[] = JHtml::_('FEFHelp.select.option', '*');

		$monthNames = [
			1  => 'JANUARY',
			2  => 'FEBRUARY',
			3  => 'MARCH',
			4  => 'APRIL',
			5  => 'MAY',
			6  => 'JUNE',
			7  => 'JULY',
			8  => 'AUGUST',
			9  => 'SEPTEMBER',
			10 => 'OCTOBER',
			11 => 'NOVEMBER',
			12 => 'DECEMBER',
		];

		$first = 0;
		$last  = 0;
		$step  = 0;

		if (isset($attribs['first']))
		{
			$first = (int) $attribs['first'];

			unset($attribs['first']);
		}

		if (isset($attribs['last']))
		{
			$last = (int) $attribs['last'];

			unset($attribs['last']);
		}

		if (isset($attribs['step']))
		{
			$step = (int) $attribs['step'];

			unset($attribs['step']);
		}

		if (isset($attribs['year']))
		{
			unset($attribs['year']);

			if (($first < 2010) && ($last < 2010))
			{
				$first = gmdate('Y') + $first;
				$last  = gmdate('Y') + $last;
			}
		}

		// Sanity checks.
		if ($step == 0)
		{
			// Step of 0 will create an endless loop.
			return '';
		}
		elseif ($first < $last && $step < 0)
		{
			// A negative step will never reach the last number.
			return '';
		}
		elseif ($first > $last && $step > 0)
		{
			// A position step will never reach the last number.
			return '';
		}

		$month = false;

		if (isset($attribs['month']))
		{
			$month = true;

			unset($attribs['month']);
		}

		// Build the options array.
		for ($i = $first; $i <= $last; $i += $step)
		{
			if ($month)
			{
				$options[] = JHtml::_('FEFHelp.select.option', $i, Text::_($monthNames[$i]));
			}
			else
			{
				$options[] = JHtml::_('FEFHelp.select.option', $i);
			}
		}

		return self::genericlist($options, $name, $attribs, $selected);
	}

	public static function runManagerActions($name, $selected, $attribs = [], $optionsOnly = false)
	{
		$options[] = JHTML::_('FEFHelp.select.option', '', '- ' . Text::_('COM_ATS_AUTOREPLIES_RUN_AFTER_MANAGER') . ' -');
		$options[] = JHTML::_('FEFHelp.select.option', '0', Text::_('COM_ATS_AUTOREPLIES_RUN_AFTER_MANAGER_NO'));
		$options[] = JHTML::_('FEFHelp.select.option', '1', Text::_('COM_ATS_AUTOREPLIES_RUN_AFTER_MANAGER_YES'));
		$options[] = JHTML::_('FEFHelp.select.option', '2', Text::_('COM_ATS_AUTOREPLIES_RUN_AFTER_MANAGER_ALWAYS'));

		if ($optionsOnly)
		{
			return $options;
		}

		return self::genericlist($options, $name, $attribs, $selected);
	}

	public static function autoRepliesAttachment($name, $selected, $attribs = [], $optionsOnly = false)
	{
		$options[] = JHTML::_('FEFHelp.select.option', '', '- ' . Text::_('COM_ATS_AUTOREPLIES_ATTACHMENT') . ' -');
		$options[] = JHTML::_('FEFHelp.select.option', '0', Text::_('COM_ATS_AUTOREPLIES_NO_ATTACHMENT'));
		$options[] = JHTML::_('FEFHelp.select.option', '1', Text::_('COM_ATS_AUTOREPLIES_WITH_ATTACHMENT'));
		$options[] = JHTML::_('FEFHelp.select.option', '2', Text::_('COM_ATS_AUTOREPLIES_ATTACHMENT_ALWAYS'));

		if ($optionsOnly)
		{
			return $options;
		}

		return self::genericlist($options, $name, $attribs, $selected);
	}

	/**
	 * Return a generic drop-down list
	 *
	 * @param   array   $list      An array of objects, arrays, or scalars.
	 * @param   string  $name      The value of the HTML name attribute.
	 * @param   mixed   $attribs   Additional HTML attributes for the <select> tag. This
	 *                             can be an array of attributes, or an array of options. Treated as options
	 *                             if it is the last argument passed. Valid options are:
	 *                             Format options, see {@see JHtml::$formatOptions}.
	 *                             Selection options, see {@see JHtmlSelect::options()}.
	 *                             list.attr, string|array: Additional attributes for the select
	 *                             element.
	 *                             id, string: Value to use as the select element id attribute.
	 *                             Defaults to the same as the name.
	 *                             list.select, string|array: Identifies one or more option elements
	 *                             to be selected, based on the option key values.
	 * @param   mixed   $selected  The key that is selected (accepts an array or a string).
	 * @param   string  $idTag     Value of the field id or null by default
	 *
	 * @return  string  HTML for the select list
	 */
	protected static function genericlist($list, $name, $attribs, $selected, $idTag = null)
	{
		$idTag   = $idTag ?? $name;
		$attribs = empty($attribs) ? [] : $attribs;
		$class   = $attribs['class'] ?? '';

		if (strpos($class, 'advancedSelect') !== false)
		{
			return JHtml::_('FEFHelp.select.smartlist', $list, $name, $attribs, 'value', 'text', $selected, $idTag);
		}

		return JHtml::_('FEFHelp.select.genericlist', $list, $name, $attribs, 'value', 'text', $selected, $idTag);
	}

	/**
	 * Generates an HTML radio list.
	 *
	 * @param   array    $list      An array of objects
	 * @param   string   $name      The value of the HTML name attribute
	 * @param   string   $attribs   Additional HTML attributes for the <select> tag
	 * @param   string   $selected  The name of the object variable for the option text
	 * @param   boolean  $idTag     Value of the field id or null by default
	 *
	 * @return  string  HTML for the select list
	 */
	protected static function genericradiolist($list, $name, $attribs = null, $selected = null, $idTag = null)
	{
		if (empty($attribs))
		{
			$attribs = [];
		}

		return JHtml::_('FEFHelp.select.radiolist', $list, $name, $attribs, 'value', 'text', $selected, $idTag);
	}

	/**
	 * Filters an array of categories by the currently active language, including items marked as "Any language"
	 * (language = *).
	 *
	 * @param   array  $items  The array of categories to filter
	 *
	 * @return  array  The filtered array
	 */
	protected static function filterCategoriesByLanguage($items)
	{
		/** @var SiteApplication $app */
		$app               = JFactory::getApplication();
		$hasLanguageFilter = false;

		if (method_exists($app, 'getLanguageFilter'))
		{
			$hasLanguageFilter = $app->getLanguageFilter();
		}

		if (!$hasLanguageFilter)
		{
			return $items;
		}

		// Ask Joomla for the plugin only if we don't already have it. Useful for tests
		$lang_filter_plugin = PluginHelper::getPlugin('system', 'languagefilter');
		$lang_filter_params = new Registry($lang_filter_plugin->params);

		if ($lang_filter_params->get('remove_default_prefix'))
		{
			// Get default site language
			$lg       = JFactory::getLanguage();
			$language = $lg->getTag();
		}
		else
		{
			// We have to use JInput since the language fragment is not set in the $_REQUEST
			$langCode = $app->input->getCmd('language', '*');
			$lg       = JLanguage::getInstance($langCode);
			$language = $lg->getTag();
		}

		$temp = [];

		foreach ($items as $item)
		{
			if (empty($item->lang_code) || ($item->lang_code == $language))
			{
				$temp[] = $item;
			}
		}

		return $temp;
	}

	/**
	 * Returns the custom ticket statuses, defined in the component configuration.
	 *
	 * @return array
	 *
	 * @since  3.4.0
	 */
	public static function getCustomTicketStatuses()
	{
		static $statuses = '';

		if (is_array($statuses))
		{
			return $statuses;
		}

		$statuses = [];
		$container = Container::getInstance('com_ats');

		// Let's prime the statuses variable as an array, so next time I can simply skip the following code
		$statuses = [];
		$custom   = $container->params->get('customStatuses', '');
		$custom   = str_replace("\\n", "\n", $custom);
		$custom   = str_replace("\r", "\n", $custom);
		$custom   = str_replace("\n\n", "\n", $custom);

		$lines = explode("\n", $custom);

		foreach ($lines as $line)
		{
			$parts = explode('=', $line);

			if (count($parts) != 2)
			{
				continue;
			}

			$parts[0] = trim($parts[0]);
			$parts[1] = trim($parts[1]);

			if (!is_numeric($parts[0]))
			{
				continue;
			}

			$statusId = (int) $parts[0];

			if (($statusId <= 0) || ($statusId > 99))
			{
				continue;
			}

			$description = $parts[1];

			if (empty($description))
			{
				continue;
			}

			$statuses[$statusId] = $description;
		}

		return $statuses;
	}

}
