<?php
/**
 * @package   ats
 * @copyright Copyright (c)2011-2022 Nicholas K. Dionysopoulos / Akeeba Ltd
 * @license   GNU General Public License version 3, or later
 */

namespace Akeeba\TicketSystem\Admin\Model;

defined('_JEXEC') or die;

use Akeeba\TicketSystem\Admin\Helper\Permissions;
use Exception;
use FOF40\Container\Container;
use FOF40\Date\Date;
use Joomla\CMS\Factory;
use Joomla\CMS\Filesystem\File as JFile;
use Joomla\CMS\Filesystem\Path as JPath;
use Joomla\CMS\Helper\MediaHelper;
use Joomla\CMS\Language\Text;

/**
 * Class Attachments
 *
 * @property    int       ats_attachment_id   Primary key
 * @property    string    original_filename   Original name of the attachment
 * @property    string    mangled_filename    Hashed filename
 * @property    string    mime_type           File mime-type
 * @property    int       ats_post_id         Link to the post containing the attachment
 *
 * Filters:
 *
 * @method  $this   ids_search($ids)        Search for a set of attachment ids
 * @method  $this   created_on($search)     Search on created_on field
 *
 * @property-read   Posts post                Post linked to the attachment
 *
 * @package Akeeba\TicketSystem\Admin\Model
 */
class Attachments extends DefaultDataModel
{
	private static $attachmentsDirectory;

	public function __construct(Container $container, array $config = [])
	{
		parent::__construct($container, $config);

		$this->addBehaviour('Filters');

		$this->belongsTo('post', 'Posts', 'ats_post_id', 'ats_post_id');

		$this->autoChecks = false;
	}

	/**
	 * Returns the absolute filesystem path where attachments are stored.
	 *
	 * It tries to understand the configured attachments directory in the following ways:
	 *
	 * - A path relative to the Joomla `media` folder
	 * - A path relative to the Joomla installation root folder
	 * - An absolute filesystem path
	 *
	 * If all checks fail to yield an existing directory it returns the default attachments directory
	 *
	 * @return  string
	 *
	 * @since   3.2.0
	 */
	public static function getAttachmentsDirectory()
	{
		if (!empty(self::$attachmentsDirectory))
		{
			return self::$attachmentsDirectory;
		}

		$container = Container::getInstance('com_ats');
		$directory = $container->params->get('attachments_folder', 'com_ats/attachments');
		$directory = trim($directory);
		$directory = rtrim($directory, '/\\');

		// 0. Empty setting will result in the default being used
		if (empty($directory))
		{
			self::$attachmentsDirectory = JPATH_SITE . '/media/com_ats/attachments';

			return self::$attachmentsDirectory;
		}

		// 1. Is this a subdirectory of the media folder?
		try
		{
			$possiblePath = JPATH_ROOT . '/media/' . $directory;
			$cleanDir     = JPath::check($possiblePath);

			if (is_dir($cleanDir))
			{
				self::$attachmentsDirectory = $cleanDir;

				return $cleanDir;
			}
		}
		catch (Exception $e)
		{
			// Nope. Let's fall through.
		}

		// 2. Is this a subdirectory of the site's root?
		try
		{
			$possiblePath = JPATH_ROOT . $directory;
			$cleanDir     = JPath::check($possiblePath);

			if (is_dir($cleanDir))
			{
				self::$attachmentsDirectory = $cleanDir;

				return $cleanDir;
			}
		}
		catch (Exception $e)
		{
			// Nope. Let's fall through.
		}

		// 3. Is this an absolute path (try cleaning the path first)?
		try
		{
			$cleanDir = JPath::clean($directory);

			if (is_dir($cleanDir))
			{
				self::$attachmentsDirectory = $cleanDir;

				return $cleanDir;
			}
		}
		catch (Exception $e)
		{
			// WTF? This shouldn't happen.
		}

		// 4. Is this an absolute path?
		if (is_dir($directory))
		{
			self::$attachmentsDirectory = $cleanDir;

			return $directory;
		}

		// 5. Everything failed, return the default.
		self::$attachmentsDirectory = JPATH_SITE . '/media/com_ats/attachments';

		return self::$attachmentsDirectory;
	}

	public static function secureAttachmentsDirectory()
	{
		$dir = self::getAttachmentsDirectory();

		// Create a .htaccess file to prevent all web access (Apache 1.3+, Lightspeed, Lighttpd, ...)
		if (!is_file($dir . '/.htaccess'))
		{
			$htaccess = <<< APACHE
<IfModule !mod_authz_core.c>
Order deny,allow
Deny from all
</IfModule>
<IfModule mod_authz_core.c>
  <RequireAll>
    Require all denied
  </RequireAll>
</IfModule>
APACHE;

			@file_put_contents($dir . '/.htaccess', $htaccess);
		}

		// Create a web.config to prevent all web access (IIS 7+)
		if (!is_file($dir . '/web.config'))
		{
			$webConfig = <<< XML
<?xml version="1.0"?>
<!--
    This only works on IIS 7 or later. See https://www.iis.net/configreference/system.webserver/security/requestfiltering/fileextensions
-->
<configuration>
    <system.webServer>
        <security>
            <requestFiltering>
                <fileExtensions allowUnlisted="false" >
                    <clear />
                    <add fileExtension=".html" allowed="true"/>
                </fileExtensions>
            </requestFiltering>
        </security>
    </system.webServer>
</configuration>
XML;
			@file_put_contents($dir . '/web.config', $webConfig);
		}

		// Create a blank index.html or index.htm to prevent directory listings (all servers)
		$blankHtml = <<< HTML
<!DOCTYPE HTML PUBLIC "-//W3C//DTD HTML 4.01 Transitional//EN">
<html>
  <head>
    <title>Access Denied</title>
  </head>
  <body>
	  <h1>Access Denied</h1>
  </body>
</html>
HTML;

		if (!is_file($dir . '/index.html'))
		{
			@file_put_contents($dir . '/index.html', $blankHtml);
		}

		if (!is_file($dir . '/index.htm'))
		{
			@file_put_contents($dir . '/index.htm', $blankHtml);
		}

		// Create a default index.php to prevent directory listings with an error (all servers)
		if (!is_file($dir . '/index.php'))
		{
			$deadPHP = '<' . '?' . 'php header(\'HTTP/1.1 403 Forbidden\');';
			@file_put_contents($dir . '/index.php', $deadPHP);
		}
	}

	/**
	 * @param   null  $oid
	 *
	 * @return bool
	 */
	public function deleteFile($oid = null)
	{
		if ($oid)
		{
			$this->load($oid);
		}

		// Delete the physical file before deleting the record
		if ($this->mangled_filename)
		{
			$filename = self::getAttachmentsDirectory() . '/' . $this->mangled_filename;

			if (JFile::exists($filename))
			{
				return JFile::delete($filename);
			}
		}

		return true;
	}

	public function buildQuery($override = false)
	{
		$db = $this->getDbo();

		$query = parent::buildQuery($override);

		if ($ids = $this->getState('ids_search', [], 'array'))
		{
			$ids = array_map([$db, 'quote'], $ids);
			$query->where($db->qn('ats_attachment_id') . ' IN (' . implode(',', $ids) . ')');
		}

		return $query;
	}

	public function updateSavedAttachments($attachments, $post)
	{
		if (!is_array($attachments))
		{
			$attachments = explode(',', $attachments);
		}

		$db = $this->getDbo();

		$quoted_attachments = array_map([$db, 'quote'], $attachments);

		$query = $db->getQuery(true)
			->update($db->qn('#__ats_attachments'))
			->set($db->qn('ats_post_id') . ' = ' . $db->q($post))
			->where($db->qn('ats_attachment_id') . ' IN(' . implode(',', $quoted_attachments) . ')');
		$db->setQuery($query)->execute();

		// Let's emit an event so our plugins can hook to the "attachment saved" event
		$this->container->platform->runPlugins('onAttachmentPostUpdate', [$attachments, $post]);
	}

	/**
	 * Process all the uploaded files
	 *
	 * @param   array  $files   Uploaded files. Please note that this array should be already organized following
	 *                          JInputFiles behavior, ie $files = array(0 => array(file details), 1 => array(file
	 *                          details))
	 *
	 * @return array|bool
	 */
	public function manageUploads($files)
	{
		if (!is_array($files))
		{
			return false;
		}

		if (!isset($files[0]['name']))
		{
			return false;
		}

		$w_files = [];
		$errors  = [];
		$return  = [];

		// We allow only 10 attachments maximum
		for ($i = 0; $i < 10; $i++)
		{
			if (!isset($files[$i]))
			{
				break;
			}

			// Remove whitespaces or MediaManager will report an error
			$files[$i]['name'] = str_replace(' ', '_', $files[$i]['name']);

			$w_files[] = $files[$i];
		}

		$files = $w_files;

		foreach ($files as $file)
		{
			try
			{
				$filedef = $this->uploadFile($file, true);
			}
			catch (Exception $e)
			{
				$errors[] = Text::sprintf('COM_ATS_ATTACHMENTS_ERR_UPLOAD', $file['name'], $e->getMessage());

				continue;
			}

			$this->reset();
			$this->ats_attachment_id = null;

			$date = new Date();

			$filedef['ats_post_id'] = 0;
			$filedef['created_by']  = Permissions::getUser()->id;
			$filedef['created_on']  = $date->toSql();
			$filedef['enabled']     = 1;

			try
			{
				$this->save($filedef);
			}
			catch (Exception $e)
			{
				$errors[] = Text::sprintf('COM_ATS_ATTACHMENTS_ERR_UPLOAD', $file['name'], $e->getMessage());

				continue;
			}

			$return[] = $this->ats_attachment_id;
		}

		return [
			implode(',', $return),
			$errors,
		];
	}

	/**
	 * Moves an uploaded file to the configured attachments directory under a random name and returns a full file
	 * definition array, or false if the upload failed for any reason.
	 *
	 * @param   array  $file         The file descriptor returned by PHP
	 * @param   bool   $checkUpload  Should I check if upload is allowed for this file?
	 *
	 * @return  array
	 *
	 * @throws  Exception
	 */
	public function uploadFile($file, $checkUpload = true)
	{
		if (!isset($file['name']))
		{
			throw new Exception(Text::_('COM_ATS_ATTACHMENTS_ERR_NOFILE'));
		}

		// Can we upload this file type?
		$paths = [JPATH_ROOT, JPATH_ADMINISTRATOR];

		$jlang = $this->container->platform->getLanguage();
		$jlang->load('com_media', $paths[0], 'en-GB', true);
		$jlang->load('com_media', $paths[0], null, true);
		$jlang->load('com_media', $paths[1], 'en-GB', true);
		$jlang->load('com_media', $paths[1], null, true);

		$mediaHelper = new MediaHelper;

		if (!$mediaHelper->canUpload($file, 'com_media'))
		{
			$app    = Factory::getApplication();
			$errors = $app->getMessageQueue();

			if (count($errors))
			{
				$error = array_pop($errors);
				$err   = $error['message'];
			}
			else
			{
				$err = '';
			}

			if (!empty($err))
			{
				throw new Exception(Text::_('COM_ATS_ATTACHMENTS_ERR_MEDIAHELPERERROR') . ' ' . $err);
			}
			else
			{
				throw new Exception(Text::_('COM_ATS_POSTS_ERR_ATTACHMENTERRORGENERIC'));
			}
		}

		// Get a (very!) randomised name
		$serverkey   = $this->container->platform->getConfig()->get('secret', '');
		$sig         = $file['name'] . microtime() . $serverkey;
		$mangledname = md5($sig);

		if (function_exists('sha256'))
		{
			$mangledname = sha256($sig);
		}
		elseif (function_exists('sha1'))
		{
			$mangledname = sha1($sig);
		}

		// ...and its full path
		$filepath = JPath::clean(self::getAttachmentsDirectory() . '/' . $mangledname);

		// If we have a name clash, abort the upload
		if (JFile::exists($filepath))
		{
			throw new Exception(Text::_('COM_ATS_ATTACHMENTS_ERR_NAMECLASH'));
		}

		// Do the upload
		if ($checkUpload)
		{
			if (!JFile::upload($file['tmp_name'], $filepath))
			{
				throw new Exception(Text::_('COM_ATS_ATTACHMENTS_ERR_CANTJFILEUPLOAD'));
			}
		}
		else
		{
			$copied = @copy($file['tmp_name'], $filepath);

			if (!$copied)
			{
				$copied = JFile::copy($file['tmp_name'], $filepath);
			}

			if (!$copied)
			{
				throw new Exception(Text::_('COM_ATS_ATTACHMENTS_ERR_CANTJFILEUPLOAD'));
			}
		}

		// Get the MIME type
		$mime = 'application/octet-stream';

		if (function_exists('mime_content_type'))
		{
			$mime = mime_content_type($filepath);
		}
		elseif (function_exists('finfo_open'))
		{
			$finfo = finfo_open(FILEINFO_MIME_TYPE);
			$mime  = finfo_file($finfo, $filepath);
		}

		// Return the file info
		return [
			'original_filename' => $file['name'],
			'mangled_filename'  => $mangledname,
			'mime_type'         => $mime,
		];
	}

	public function doDownload()
	{
		// Calculate the Etag
		$etagContent = $this->mangled_filename . $this->mime_type . $this->original_filename . $this->created_on . $this->created_by;

		if (function_exists('sha1'))
		{
			$eTag = sha1($etagContent);
		}
		else
		{
			$eTag = md5($etagContent);
		}

		// Do we have an If-None-Match header?
		$inm = '';

		if (function_exists('apache_request_headers'))
		{
			$headers = apache_request_headers();

			if (array_key_exists('If-None-Match', $headers))
			{
				$inm = $headers['If-None-Match'];
			}
		}

		if (empty($inm))
		{
			if (array_key_exists('HTTP-IF-NONE-MATCH', $_SERVER))
			{
				$inm = $_SERVER['HTTP-IF-NONE-MATCH'];
			}
		}

		if ($inm == $eTag)
		{
			while (@ob_end_clean())
			{
				;
			}
			header('HTTP/1.0 304 Not Modified');

			$this->container->platform->closeApplication();
		}

		$filepath = JPath::clean(self::getAttachmentsDirectory() . '/' . $this->mangled_filename);
		$basename = $this->original_filename;

		if (!\JFile::exists($filepath))
		{
			header('HTTP/1.0 404 Not Found');
			$this->container->platform->closeApplication();
		}

		$this->input->set('format', 'raw');

		// Disable error reporting and error display
		if (function_exists('error_reporting'))
		{
			error_reporting(0);
		}

		if (function_exists('ini_set'))
		{
			@ini_set('display_error', 0);
		}

		// Clear cache
		while (@ob_end_clean())
		{
			;
		}

		// Fix IE bugs
		if (isset($_SERVER['HTTP_USER_AGENT']) && strstr($_SERVER['HTTP_USER_AGENT'], 'MSIE'))
		{
			$header_file = preg_replace('/\./', '%2e', $basename, substr_count($basename, '.') - 1);

			if (ini_get('zlib.output_compression'))
			{
				ini_set('zlib.output_compression', 'Off');
			}
		}
		else
		{
			$header_file = $basename;
		}

		@clearstatcache();

		// Disable caching for regular attachment disposition
		if ($this->getState('disposition', 'attachment') !== 'attachment')
		{
			header("Pragma: public");
			header("Expires: 0");
			header("Cache-Control: must-revalidate, post-check=0, pre-check=0");
			header("Cache-Control: public", false);
		}

		// Send a Date header
		$date = new Date($this->created_on);

		header('Date: ' . $date->toRFC822());

		// Send an Etag
		header('Etag: ' . $eTag);

		// Send MIME headers
		header("Content-Description: File Transfer");

		if (empty($this->mime_type))
		{
			header('Content-Type: application/octet-stream');
		}
		else
		{
			header('Content-Type: ' . $this->mime_type);
		}

		header("Accept-Ranges: bytes");

		if ($this->getState('disposition', 'attachment') != 'attachment')
		{
			header('Content-Disposition: inline; filename="' . $header_file . '"');
		}
		else
		{
			header('Content-Disposition: attachment; filename="' . $header_file . '"');
		}

		header('Content-Transfer-Encoding: binary');

		// Notify of filesize, if this info is available
		$filesize = @filesize($filepath);

		if ($filesize > 0)
		{
			header('Content-Length: ' . (int) $filesize);
		}

		// Disable time limits
		if (!ini_get('safe_mode'))
		{
			set_time_limit(0);
		}

		// Use 1M chunks for echoing the data to the browser
		@flush();
		$chunksize = 1024 * 1024; //1M chunks

		$handle = @fopen($filepath, 'r');

		if ($handle !== false)
		{
			while (!feof($handle))
			{
				$buffer = fread($handle, $chunksize);

				echo $buffer;

				@ob_flush();
				flush();
			}

			@fclose($handle);
		}
		else
		{
			@readfile($filepath);
			@flush();
		}

		// Ungraceful application exit -- so that any plugins won't screw up the download...
		$this->container->platform->closeApplication();
	}

	protected function onBeforeDelete($oid)
	{
		// Delete the physical file before deleting the record
		$this->deleteFile();

		// Update post record removing the link to this attachment
		/** @var Posts $post */
		$post = $this->container->factory->model('Posts')->tmpInstance();
		$post->load($this->ats_post_id);

		if ($post->ats_post_id)
		{
			$attachments = array_diff($post->ats_attachment_id, [$oid]);

			// If it's an empty set, let's add 0 for consistency vs previous versions
			if (!$attachments)
			{
				$attachments[] = 0;
			}

			$data['ats_attachment_id']    = $attachments;
			$data['_do_not_check_ticket'] = 1;

			$post->save($data);
		}
	}
}
