<?php
/**
 * @package   ats
 * @copyright Copyright (c)2011-2022 Nicholas K. Dionysopoulos / Akeeba Ltd
 * @license   GNU General Public License version 3, or later
 */

namespace Akeeba\TicketSystem\Admin\Model;

defined('_JEXEC') or die;

use Akeeba\TicketSystem\Admin\Helper\Permissions;
use Akeeba\TicketSystem\Admin\Helper\PluginParameters;
use FOF40\Container\Container;
use FOF40\Timer\Timer;
use FOF40\Utils\ArrayHelper;
use Joomla\CMS\Language\Text;
use Joomla\CMS\User\UserHelper;
use RuntimeException;
use Traversable;

/**
 * @property    int          $ats_autoreply_id
 * @property    string       $title   Autoreply title
 * @property    string       $reply   Autoreply content
 * @property    int          $run_after_manager
 * @property    int          $num_posts
 * @property    int          $min_after
 * @property    int          $attachment
 * @property    string       $keywords_title
 * @property    string       $keywords_text
 * @property    string|array $params
 *
 * @package Akeeba\TicketSystem\Admin\Model
 */
class AutoReplies extends DefaultDataModel
{
	/**
	 * Processed tickets stack, in order to prevent double posting
	 *
	 * @var   array
	 */
	protected $processedTickets = [];

	/**
	 * List of bot user IDs
	 *
	 * @var   array
	 * @since 3.2.0
	 */
	protected $bots = [];

	/** @inheritDoc */
	public function __construct(Container $container, array $config = [])
	{
		parent::__construct($container, $config);

		$this->addBehaviour('Filters');

		$this->autoChecks = false;

		$this->populateBots();
	}

	/** @inheritDoc */
	public function check()
	{
		$this->assertNotEmpty($this->title, 'COM_ATS_AUTOREPLIES_ERR_TITLE');
		$this->assertNotEmpty($this->reply, 'COM_ATS_AUTOREPLY_ERR_REPLY_EMPTY');

		parent::check();
	}

	/**
	 * Run the auto-reply rules and return the number of tickets we replied to.
	 *
	 * If $tickets is set to a non-empty array or Traversable object then only these tickets will be taken into account,
	 * without ticket age limits, without querying the database for tickets in need of an automatic reply.
	 *
	 * @param   array|Traversable|null  $tickets  (array of integers) Only auto-reply to these ticket IDs
	 * @param   Timer|null              $timer    A timer object to pace ourselves
	 *
	 * @return  int  The number of tickets we replied to
	 */
	public function runAutoreplies($tickets = null, $timer = null)
	{
		// Sanity check: are there any bot users?
		if (count($this->bots) == 0)
		{
			throw new RuntimeException(Text::_('COM_ATS_AUTOREPLIES_ERR_NOBOTS'));
		}

		// Let's get all the enabled autoreply rules
		$rules = $this->tmpInstance()
			->enabled(1)
			->filter_order('ordering')
			->filter_order_Dir('ASC')
			->get(true);

		// Sanity check: are there any rules?
		if ($rules->count() == 0)
		{
			return 0;
		}

		if (!is_object($timer) || !($timer instanceof Timer))
		{
			$timer = null;
		}

		// And now let's start looking for suitable tickets
		foreach ($rules as $rule)
		{
			if (is_object($timer) && ($timer instanceof Timer))
			{
				if ($timer->getTimeLeft() <= 0.01)
				{
					break;
				}
			}

			// If I got an error let's bubble it out
			$this->runRule($rule, $tickets, $timer);
		}

		// If everything was ok, let's return the number of processed tickets
		return count($this->processedTickets);
	}

	/**
	 * Runs before binding data to the Model.
	 *
	 * Used to make sure the params array will always have a categories key
	 *
	 * @param   array|null  $data
	 *
	 * @return  void
	 */
	protected function onBeforeBind(&$data)
	{
		if (is_array($data) && !isset($data['params']))
		{
			$data['params']['categories'] = [];
		}
	}

	/**
	 * Runs a specific auto-reply rule
	 *
	 * If $tickets is set to a non-empty array or Traversable object then only these tickets will be taken into account,
	 * without ticket age limits, without querying the database for tickets in need of an automatic reply.
	 *
	 * @param   AutoReplies             $rule
	 * @param   array|Traversable|null  $tickets  (array of integers) Only auto-reply to these ticket IDs
	 * @param   Timer|null              $timer    A timer object to pace ourselves
	 *
	 * @return  void
	 */
	protected function runRule(AutoReplies $rule, $tickets = null, $timer = null)
	{
		$categories     = [];
		$keywords_title = $this->keywordsListToArray($rule->keywords_title);
		$keywords_text  = $this->keywordsListToArray($rule->keywords_text);
		$ruleId         = $rule->getId();

		// FOF will automatically decode the JSON string into an array
		$params = $rule->params;

		if (isset($params['categories']))
		{
			$categories = $params['categories'];
		}

		// $tickets could be null or an empty array. In either cases, we need to fetch the tickets from the cats
		$needsTickets = is_null($tickets);
		$needsTickets = $needsTickets || (is_array($tickets) && empty($tickets));

		if ($needsTickets)
		{
			$tickets = $this->getTickets($categories);
		}

		// Performs several checks vs the rule, narrowing the number of affected tickets
		$tickets = $this->checkMinAfter($rule->min_after, $tickets);
		$tickets = $this->checkNumPosts($rule->num_posts, $tickets);
		$tickets = $this->checkAttachments($rule->attachment, $tickets);
		$tickets = $this->checkManagerReply($rule->run_after_manager, $tickets);
		$tickets = $this->checkTitleKeywords($keywords_title, $tickets);
		$tickets = $this->checkTextKeywords($keywords_text, $tickets);

		if (empty($tickets))
		{
			return;
		}

		$jNow = $this->container->platform->getDate();

		// Don't get a tmpInstance, since that will detach the relations
		/** @var Posts $post */
		$post = $this->container->factory->model('Posts')->tmpInstance();

		/** @var Tickets $t */
		$t = $this->container->factory->model('Tickets')->tmpInstance();
		// Remove the modified behavior, we want to do that manually
		$t->removeBehaviour('Modified');

		foreach ($tickets as $ticket)
		{
			if (is_object($timer) && ($timer instanceof Timer))
			{
				if ($timer->getTimeLeft() <= 0.01)
				{
					break;
				}
			}

			$t->reset()->find($ticket);

			// Do I have any posts on this ticket from this auto-reply rule?
			$params      = $t->params;
			$autoReplies = [];

			if (isset($params['__ats__autoreplies']))
			{
				$autoReplies = $params['__ats__autoreplies'];
			}

			if (!is_array($autoReplies))
			{
				$autoReplies = [];
			}

			if (in_array($ruleId, $autoReplies))
			{
				continue;
			}

			$autoReplies[]                = $ruleId;
			$params['__ats__autoreplies'] = $autoReplies;

			$current_bot = $this->bots[array_rand($this->bots)];

			/**
			 * I'll leave empty the creation date, so FOF will populate it automatically.
			 * Otherwise it will populate the modified_ column set, resulting in an edited post (which is false)
			 */
			$data = [
				'ats_post_id'   => 0,
				'ats_ticket_id' => $ticket,
				'content_html'  => $rule->reply,
				'created_on'    => null,
				'created_by'    => $current_bot,
				'enabled'       => 1,
			];

			$post->save($data);

			$t->status      = 'P';
			$t->modified_on = $jNow->toSql();
			$t->modified_by = $current_bot;
			$t->params      = $params;

			$t::$noPermissionsCheck = 1;

			$t->save();

			$t::$noPermissionsCheck = 0;

			$this->processedTickets[] = $ticket;
		}
	}

	protected function setParamsAttribute($value)
	{
		return $this->setAttributeForJson($value);
	}

	protected function getParamsAttribute($value)
	{
		return $this->getAttributeForJson($value);
	}

	protected function keywordsListToArray($keywords)
	{
		$keywords = trim($keywords);
		$keywords = str_replace(["\\n", "\r", "\r\n", "\n\n"], ["\n", "\n", "\n", "\n"], $keywords);

		if ($keywords)
		{
			$ret = explode("\n", $keywords);

			return array_map('trim', $ret);
		}

		return [];
	}

	/**
	 * Populate $this->bots with the user IDs of valid bot users
	 *
	 * @return void
	 * @since  3.2.0
	 */
	private function populateBots()
	{
		$this->bots = [];

		// Let's check if I have any username
		$pluginParams = PluginParameters::getPluginParams('autoreply');
		$rawData      = trim($pluginParams->get('botusers', ''));
		$rawData      = empty($rawData) ? '' : str_replace("\\n", "\n", $rawData);
		$rawData      = empty($rawData) ? '' : str_replace("\r\n", "\n", $rawData);
		$rawData      = empty($rawData) ? '' : str_replace("\r", "\n", $rawData);

		if (empty($rawData))
		{
			return;
		}

		// Just in case something funky happened...
		$bots = array_map(function ($botUser) {
			if (empty($botUser))
			{
				return null;
			}

			$userId = UserHelper::getUserId($botUser);

			if (empty($userId))
			{
				return null;
			}

			$user = Permissions::getUser($userId);

			if (empty($user))
			{
				return null;
			}

			if (empty($user->id))
			{
				return null;
			}

			return $user->id;
		}, explode("\n", $rawData));

		$this->bots = array_filter($bots, function ($uID) {
			return !is_null($uID) && ($uID > 0);
		});
	}

	private function getTickets($categoryIDs = [])
	{
		static $tickets = [];

		$db = $this->getDbo();

		// Remove any empty value
		$categoryIDs = array_diff($categoryIDs, ['', '0']);
		$hash        = md5(implode('', $categoryIDs));

		if (!isset($tickets[$hash]))
		{
			// Get the ticket id where a bot already replied in previous iterations
			$query       = $db->getQuery(true)
				->select($db->qn('ats_ticket_id'))
				->from($db->qn('#__ats_posts'))
				->where($db->qn('created_by') . ' IN(' . implode(',', $this->bots) . ')');
			$autoReplied = $db->setQuery($query)->loadColumn();

			// Get only open tickets
			$query = $db->getQuery(true)
				->select($db->qn('ats_ticket_id'))
				->from($db->qn('#__ats_tickets'))
				->where($db->qn('status') . ' = ' . $db->quote('O'));

			// Skip tickets with an auto-reply
			if ($autoReplied)
			{
				$query->where($db->qn('ats_ticket_id') . ' NOT IN(' . implode($autoReplied) . ')');
			}

			if ($categoryIDs)
			{
				$query->where($db->qn('catid') . ' IN(' . implode(',', $categoryIDs) . ')');
			}

			$tickets[$hash] = $db->setQuery($query)->loadColumn();
		}

		// Subtracts the already processed tickets in order to avoid double posting
		$tickets[$hash] = array_diff($tickets[$hash], $this->processedTickets);

		return $tickets[$hash];
	}

	private function checkMinAfter($minutes, $tickets)
	{
		if (!$minutes || !$tickets)
		{
			return $tickets;
		}

		$db   = $this->getDbo();
		$time = $this->container->platform->getDate('-' . $minutes . ' minutes');

		$query = $db->getQuery(true)
			->select($db->qn('ats_ticket_id'))
			->from($db->qn('#__ats_tickets'))
			->where($db->qn('ats_ticket_id') . ' IN(' . implode(',', $tickets) . ')')
			->where($db->qn('modified_on') . ' <= ' . $db->q($time->toSql()));

		$tickets = $db->setQuery($query)->loadColumn();

		return $tickets;
	}

	private function checkNumPosts($numPost, $tickets)
	{
		// If I have no constraints on the number of posts, let's return the passed ticket array
		if (!$numPost || $numPost == 1 || !$tickets)
		{
			return $tickets;
		}

		$db = $this->getDbo();

		$query = $db->getQuery(true)
			->select($db->qn('ats_ticket_id') . ', COUNT(*) as total')
			->from($db->qn('#__ats_posts'))
			->where($db->qn('ats_ticket_id') . ' IN(' . implode(',', $tickets) . ')')
			->group($db->qn('ats_ticket_id'))
			->having($db->qn('total') . ' >= ' . $numPost);

		$tickets = $db->setQuery($query)->loadColumn();

		return $tickets;
	}

	private function checkAttachments($attach, $tickets)
	{
		// If the rule applies to posts with or without attachments (ie every ticket),
		// let's return the passed ticket array
		if ($attach == 2 || !$tickets)
		{
			return $tickets;
		}

		$db = $this->getDbo();

		$query = $db->getQuery(true)
			->select($db->qn('ats_ticket_id') . ', SUM(' . $db->qn('ats_attachment_id') . ') as attachments')
			->from($db->qn('#__ats_posts'))
			->where($db->qn('ats_ticket_id') . ' IN(' . implode(',', $tickets) . ')')
			->group($db->qn('ats_ticket_id'));

		// I want tickets where AT LEAST ONE post has an attachment
		if ($attach)
		{
			$query->having($db->qn('attachments') . ' >= 0');
		}
		// I want tickets where NO POSTS have an attachments
		else
		{
			$query->having($db->qn('attachments') . ' = 0');
		}

		$tickets = $db->setQuery($query)->loadColumn();

		return $tickets;
	}

	private function checkManagerReply($manager, $tickets)
	{
		// Rule applies regardless the manager reply, let's return the passed ticket array
		if ($manager == 2 || !$tickets)
		{
			return $tickets;
		}

		$db       = $this->getDbo();
		$managers = [];

		// Let's get all the affected categories
		$query = $db->getQuery(true)
			->select($db->qn('catid'))
			->from($db->qn('#__ats_tickets'))
			->where($db->qn('ats_ticket_id') . ' IN(' . implode(',', $tickets) . ')')
			->group($db->qn('catid'));

		$cats = $db->setQuery($query)->loadColumn();

		// Now let's get all the managers
		// The only problem that could arise, is if a manager posts as customer inside another category he's not manager
		// I really think this is a remote chance
		foreach ($cats as $cat)
		{
			$managers = array_merge($managers, Permissions::getManagers($cat));
		}

		$managers = array_map(function($x) {
			return $x->id;
		}, $managers);

		$query = $db->getQuery(true)
			->select($db->qn('ats_ticket_id'))
			->from($db->qn('#__ats_posts'))
			->where($db->qn('ats_ticket_id') . ' IN(' . implode(',', $tickets) . ')')
			->group($db->qn('ats_ticket_id'));

		// I want only ticket where the manager didn't reply
		if (!$manager)
		{
			$query->where($db->qn('created_by') . ' NOT IN(' . implode(',', $managers) . ')');
		}
		// Only where where he replied
		else
		{
			$query->where($db->qn('created_by') . ' IN(' . implode(',', $managers) . ')');
		}

		$tickets = $db->setQuery($query)->loadColumn();

		return $tickets;

	}

	private function checkTitleKeywords(array $keywords, $tickets)
	{
		if (!$keywords || !$tickets)
		{
			return $tickets;
		}

		$db    = $this->getDbo();
		$parts = [];

		$query = $db->getQuery(true)
			->select($db->qn('ats_ticket_id'))
			->from($db->qn('#__ats_tickets'))
			->where($db->qn('ats_ticket_id') . ' IN(' . implode(',', $tickets) . ')');

		foreach ($keywords as $keyword)
		{
			$parts[] = $db->qn('title') . ' LIKE ' . $db->q('%' . $keyword . '%');
		}

		$query->where('(' . implode(' OR ', $parts) . ')');

		$tickets = $db->setQuery($query)->loadColumn();

		return $tickets;
	}

	private function checkTextKeywords($keywords, $tickets)
	{
		if (!$keywords || !$tickets)
		{
			return $tickets;
		}

		$db    = $this->getDbo();
		$parts = [];

		$query = $db->getQuery(true)
			->select($db->qn('ats_ticket_id'))
			->from($db->qn('#__ats_posts'))
			->where($db->qn('ats_ticket_id') . ' IN(' . implode(',', $tickets) . ')');

		foreach ($keywords as $keyword)
		{
			$parts[] = $db->qn('content') . ' LIKE ' . $db->q('%' . $keyword . '%');
			$parts[] = $db->qn('content_html') . ' LIKE ' . $db->q('%' . $keyword . '%');
		}

		$query->where('(' . implode(' OR ', $parts) . ')');

		$tickets = $db->setQuery($query)->loadColumn();

		return $tickets;
	}
}
