<?php
/**
 * @package   ats
 * @copyright Copyright (c)2011-2022 Nicholas K. Dionysopoulos / Akeeba Ltd
 * @license   GNU General Public License version 3, or later
 */

namespace Akeeba\TicketSystem\Admin\Model;

defined('_JEXEC') or die;

use Akeeba\TicketSystem\Admin\Helper\Permissions;
use FOF40\Container\Container;
use Joomla\Registry\Registry;

/**
 * Class Categories
 *
 * @property       int      $id             Primary key
 * @property       string   $title
 * @property       string   $description    Category description
 * @property       int      $level          Category level (depth)
 * @property-read  Registry $params         Category parameters
 *
 * Filters:
 *
 * @method  $this   category($catid)    Filter by category ID
 * @method  $this   userid($id)         User used to filter categories by access
 * @method  $this   ignoreUser($flag)   Should I ignore the user access level?
 *
 * @package Akeeba\TicketSystem\Admin\Model
 */
class Categories extends DefaultDataModel
{
    public function __construct(Container $container, array $config = array())
    {
        $config['tableName']   = '#__categories';
        $config['idFieldName'] = 'id';

        parent::__construct($container, $config);

        $this->addBehaviour('Filters');
    }

    public function get($overrideLimits = false, $limitstart = 0, $limit = 0)
    {
        // Always show all the categories
        return parent::get(true, $limitstart, $limit);
    }

    public function buildQuery($override = false)
    {
        $db = $this->getDbo();

	    $userid        = $this->getState('userid', null, 'int');
	    $user          = Permissions::getUser($userid);
	    $fltIgnoreUser = $this->getState('ignoreUser', 0) != 0;

	    /**
	     * Take care of special users: Super Users and ATS backend staff. These users are exempt from
	     * authorized view level checks.
	     */
        if ($user && !$user->guest)
        {
	        $fltIgnoreUser |= $user->authorise('core.admin');
	        $fltIgnoreUser |= $user->authorise('core.manage', 'com_ats');
        }

        // Do I have a valid user (so I'm not in CLI) AND I want to apply access level rules?
        // Let's put the groups inside the model state, otherwise FOF will use the default value
        // creating a query like "WHERE access = 1"
        if ($user && !$fltIgnoreUser)
        {
            $this->setState('access', $user->getAuthorisedViewLevels());
        }

        $query = parent::buildQuery($override);

        $query->where($db->qn('extension').' = '.$db->q('com_ats'));
        $query->where($db->qn('published').' = '.$db->q(1));

        if($id = $this->getState('category', null, 'int'))
        {
            $query->where($db->qn('id').' = '.$db->q($id));
        }

        // Apply custom ordering
        $query->clear('order');

        $listOrdering = $this->getState('filter_order', 'lft');
        $listDirn     = $db->escape($this->getState('filter_order_Dir', 'ASC'));

        if ($listOrdering == 'access')
        {
            $query->order('access '.$listDirn.', lft '.$listDirn);
        }
        else
        {
            $query->order($db->escape($listOrdering).' '.$listDirn);
        }

        return $query;
    }

	protected function getParamsAttribute($value)
	{
		return new Registry($value);
	}

	/**
	 * @param   Registry|array|string  $value
	 *
	 * @return  string
	 *
	 * @since   3.0.0
	 */
	protected function setParamsAttribute($value)
	{
		if (is_string($value))
		{
			return $value;
		}

		if (is_array($value))
		{
			$value = new Registry($value);
		}

		return $value->toString('JSON');
	}
}
