<?php
/**
 * @package   ats
 * @copyright Copyright (c)2011-2022 Nicholas K. Dionysopoulos / Akeeba Ltd
 * @license   GNU General Public License version 3, or later
 */

namespace Akeeba\TicketSystem\Admin\Model;

defined('_JEXEC') or die;

use FOF40\Container\Container;
use FOF40\Model\DataModel\Collection;
use RuntimeException;

/**
 * Model for custom field defintions
 *
 * @property  int                      $ats_customfield_id  Primary key
 * @property  string                   $title               Field title
 * @property  string                   $slug                Field alias
 * @property  string                   $show                One of 'all', 'category', 'notlevel'
 * @property  string                   $type                Field type
 * @property  string                   $options             Field options
 * @property  string                   $default             Default value
 * @property  bool                     $allow_empty         Should we allow empty values?
 * @property  bool                     $public              Is this a public field?
 * @property  string                   $valid_label         Translation key to show next to a valid field
 * @property  string                   $invalid_label       Translation key to show next to an invalid field
 * @property  string                   $params              Field parameters
 *
 * Filters:
 *
 * @method  $this  ats_customfield_id()          akeebasubs_customfield_id(int $v)
 * @method  $this  title()                      title(string $v)
 * @method  $this  slug()                       slug(string $v)
 * @method  $this  show()                       show(string $v)
 * @method  $this  type()                       type(string $v)
 * @method  $this  options()                    options(string $v)
 * @method  $this  default()                    default(string $v)
 * @method  $this  allow_empty()                allow_empty(bool $v)
 * @method  $this  public ()                    public (bool $v)
 * @method  $this  valid_label()                valid_label(string $v)
 * @method  $this  invalid_label()              invalid_label(string $v)
 * @method  $this  enabled()                    enabled(bool $v)
 * @method  $this  ordering()                   ordering(int $v)
 * @method  $this  created_by()                 created_by(int $v)
 * @method  $this  created_on()                 created_on(string $v)
 * @method  $this  modified_by()                 modified_by(int $v)
 * @method  $this  modified_on()                 modified_on(string $v)
 *
 * Relations:
 *
 * @property   Collection|Categories[] cats                 List of categories where this custom field should be
 *             displayed
 *
 */
class CustomFields extends DefaultDataModel
{

	var $preventCatCheck = false;

	/**
	 * Public constructor.
	 *
	 * @param   Container  $container  The configuration variables to this model
	 * @param   array      $config     Configuration values for this model
	 *
	 * @throws \FOF40\Model\DataModel\Exception\NoTableColumns
	 */
	public function __construct(Container $container, array $config = [])
	{
		parent::__construct($container, $config);

		$this->addBehaviour('Filters');

		$this->belongsToMany('cats', 'Categories', 'ats_customfield_id', 'id',
			'#__ats_customfields_cats', 'ats_customfield_id', 'catid');

		$this->with(['cats']);
	}

	/**
	 * Check the data for validity.
	 *
	 * @return  static  Self, for chaining
	 *
	 * @throws RuntimeException  When the data bound to this record is invalid
	 */
	public function check()
	{
		$this->assertNotEmpty($this->show, 'COM_ATS_CUSTOMFIELDS_ERR_SHOW');
		$this->assertNotEmpty($this->slug, 'COM_ATS_ERR_SLUG_EMPTY');

		$pattern = '/^[a-z_][a-z0-9_\-]*$/';

		$this->assert(preg_match($pattern, $this->slug), 'COM_ATS_ERR_SLUG_INVALID');

		$this->slug = str_replace('-', '_', $this->slug);

		parent::check();

		// In some situation (ie reordering) I don't want to perform these check on categories
		if ($this->preventCatCheck)
		{
			return $this;
		}

		if (($this->cats->count()) && ($this->show != 'category'))
		{
			$this->show = 'category';
		}
		// I said that I want to show the field on specific categories, but I didn't say which ones...
		elseif (!$this->cats->count() && ($this->show == 'category'))
		{
			$this->show = 'all';
		}

		return $this;
	}

	/**
	 * Deletes every references from the category-customfield table
	 *
	 * @param   int  $oid  Custom field primary key
	 */
	protected function onAfterDelete($oid)
	{
		$db = $this->getDbo();

		$query = $db->getQuery(true)
			->delete($db->qn('#__ats_customfields_cats'))
			->where($db->qn('ats_customfield_id') . ' = ' . $db->q($oid));

		$db->setQuery($query)->execute();
	}
}
