<?php
/**
 * @package   ats
 * @copyright Copyright (c)2011-2022 Nicholas K. Dionysopoulos / Akeeba Ltd
 * @license   GNU General Public License version 3, or later
 */

namespace Akeeba\TicketSystem\Admin\Model;

defined('_JEXEC') or die;

use FOF40\Date\Date;
use FOF40\Model\Model;
use Joomla\CMS\Factory;
use Joomla\CMS\Filter\InputFilter;
use Joomla\CMS\HTML\HTMLHelper;
use Joomla\CMS\Language\Text;
use RuntimeException;

/**
 * ATS Log Viewer Model
 *
 * @since       3.2.0
 */
class Log extends Model
{
	/**
	 * Get an array with the names of all ATS log files as filename => language key
	 *
	 * @return  array
	 *
	 * @since   3.2.0
	 */
	public function getLogFiles()
	{
		return [
			'ats_cron.php'        => 'COM_ATS_LOG_TYPE_CRON',
			'ats_email_check.php' => 'COM_ATS_LOG_TYPE_EMAIL',
		];
	}

	/**
	 * Get JHtml options for the log files. Unavailable files are disabled.
	 *
	 * @return  array
	 *
	 * @since   3.2.0
	 */
	public function getLogOptions()
	{
		$logFolder = $this->getLogFolder();
		$options   = [
			HTMLHelper::_('select.option', '', '&mdash; ' . Text::_('COM_ATS_LOG_CHOOSE_FILE_TITLE') . ' &mdash;')
		];

		foreach ($this->getLogFiles() as $filename => $langKey)
		{
			$logPath   = $logFolder . '/' . $filename;
			$disabled  = !is_file($logPath) || !is_readable($logPath);
			$options[] = HTMLHelper::_('select.option', $filename, Text::_($langKey), 'value', 'text', $disabled);
		}

		return $options;
	}

	/**
	 * Get the absolute path to Joomla's log folder. Uses the default path if none is specified in JConfig.
	 *
	 * @return  string
	 *
	 * @since   3.2.0
	 */
	public function getLogFolder()
	{
		try
		{
			$logFolder = trim(Factory::getApplication()->get('log_path', ''));
		}
		catch (\Exception $e)
		{
			$logFolder = '';
		}

		return empty($logFolder) ? (JPATH_ADMINISTRATOR . '/logs') : $logFolder;
	}

	/**
	 * Output the raw text log file to the standard output
	 *
	 * @return  void
	 */
	public function echoRawLog()
	{
		try
		{
			$filePath = $this->getLogPath();
			$filename = basename($filePath);
		}
		catch (RuntimeException $e)
		{
			echo $e->getMessage();

			return;
		}

		if (!file_exists($filePath))
		{
			echo "Log file $filename is empty or does not exist.";

			return;
		}

		// Try to open our file for reading
		$fp = @fopen($filePath, 'r');

		if ($fp === false)
		{
			echo "ERROR: Cannot open $filename for reading.\r\n";

			return;
		}

		// All right. Let's start the output!
		echo "WARNING: Do not copy and paste lines from this file!\r\n";
		echo "You are supposed to ZIP and attach it in your support forum post.\r\n";
		echo "\r\n";
		echo "Log filename: $filename\r\n";
		echo "\r\n";
		echo "--- START OF LOG --\r\n";

		// Process each and every log line.
		while (!feof($fp))
		{
			$line = @fgets($fp);

			if ($line === false)
			{
				break;
			}

			$line = trim($line);

			// Ignore comments
			if (substr($line, 0, 1) == '#')
			{
				continue;
			}

			// Ignore PHP lines
			if (substr($line, 0, 5) == '<?php')
			{
				continue;
			}

			// Empty lines are also output as a blank line
			if (empty($line))
			{
				echo "\r\n";

				continue;
			}

			/**
			 * A log line looks like this
			 *
			 * 2020-02-10T11:19:39+00:00    DEBUG -    ats.cron    Starting the CRON job from CLI
			 *
			 * Split it by tabs into three parts: date, level/context and message
			 */
			$parts = explode("\t", $line, 4);

			// If there are fewer parts I have no idea what this line is, skip it over
			if (count($parts) != 4)
			{
				continue;
			}

			// Get the formatted date
			try
			{
				$date          = new Date($parts[0]);
				$formattedDate = $date->format('Y-m-d H:i:s P', true, false);
			}
			catch (\Exception $e)
			{
				continue;
			}

			// Get the log level
			$contextParts = explode(' ', $parts[1], 2);
			$logLevel     = $contextParts[0];

			/**
			 * Echo the formatted line.
			 *
			 * Example:
			 * 2020-02-10 13:19:39 +02:00 | DEBUG     |  Starting the CRON job from CLI
			 */
			echo sprintf("%-26s | %-9s | %s\r\n", $formattedDate, $logLevel, $parts[3]);
		}

		// Finally, close the log file.
		@fclose($fp);

		// That's the end of our log. Thank you for flying with us and see you soon!
		echo "--- END OF LOG ---\r\n";
	}

	/**
	 * Get the path to the log file based on the 'filename' model state variable
	 *
	 * @return  string
	 *
	 * @since   3.2.0
	 *
	 * @throws  RuntimeException  When the log file name is invalid.
	 */
	public function getLogPath()
	{
		// Get the filepath of the cleaned filename
		$filename = $this->getState('filename', null);

		if (is_null($filename) || empty(trim($filename)))
		{
			throw new RuntimeException("You have not selected a log file yet.");
		}

		$filter   = InputFilter::getInstance();
		$filename = $filter->clean($filename, 'cmd');

		// Is someone trying to abuse the log viewer?
		if ((substr($filename, 0, 4) != 'ats_') || (substr($filename, -4) != '.php'))
		{
			throw new RuntimeException("Nice try, buster.");
		}

		// More filtering, in case someone gets weird ideas about mulitple dots in the filename...
		$filename = str_replace('.', '_', substr($filename, 0, -4)) . '.php';

		// Get the absolute file path and make sure the file exists.
		return $this->getLogFolder() . '/' . $filename;
	}
}