<?php
/**
 * @package   ats
 * @copyright Copyright (c)2011-2022 Nicholas K. Dionysopoulos / Akeeba Ltd
 * @license   GNU General Public License version 3, or later
 */

namespace Akeeba\TicketSystem\Admin\Model;

defined('_JEXEC') or die;

use Akeeba\TicketSystem\Admin\Helper\PluginParameters;
use FOF40\Encrypt\Randval;
use FOF40\Model\Model;
use FOF40\JoomlaAbstraction\CacheCleaner;
use Joomla\CMS\Language\Text;
use Joomla\CMS\Uri\Uri;
use Joomla\CMS\User\User;
use Joomla\Filter\InputFilter;
use Joomla\Registry\Registry;
use RuntimeException;

class SchedulingInfo extends Model
{
	/**
	 * Cache of all of the installed ATS CRON plugins
	 *
	 * @var   array|null
	 * @since 3.2.0
	 */
	protected $atsCronPlugins = null;

	/**
	 * Cache of all enabled ATS CRON commands
	 *
	 * @var   array|null
	 * @since 3.2.0
	 */
	protected $atsCronCommands = null;

	/**
	 * Get the CRON information which is used to render the Automation page
	 *
	 * @return  object
	 * @since   3.2.0
	 */
	public function getCRONInfo()
	{
		$ret = (object) [
			'cli'      => (object) [
				'supported' => false,
				'path'      => false,
			],
			'frontend' => (object) [
				'supported' => false,
				'path'      => false,
			],
			'info'     => (object) [
				'windows'  => false,
				'php_path' => false,
				'root_url' => false,
				'secret'   => '',
			],
		];

		$siteRoot = rtrim(realpath(JPATH_ROOT), '/\\');

		$ret->info->windows  = (DIRECTORY_SEPARATOR == '\\') || (substr(strtoupper(PHP_OS), 0, 3) == 'WIN');
		$ret->info->php_path = $ret->info->windows ? 'c:\path\to\php.exe' : '/path/to/php';
		$ret->info->root_url = rtrim(Uri::base(), '/');
		$ret->info->secret   = $this->container->params->get('secret', '');

		if (substr($ret->info->root_url, -14) == '/administrator')
		{
			$ret->info->root_url = substr($ret->info->root_url, 0, -14);
		}

		$commandName = $this->getSelectedCRONCommand();

		// Get information for CLI CRON script
		$ret->cli->supported = true;
		$ret->cli->path      = implode(DIRECTORY_SEPARATOR, [$siteRoot, 'cli', 'ats-cron.php']);
		$ret->cli->path      .= ' --command=' . $commandName;

		// Get information for front-end backup
		if (trim($ret->info->secret))
		{
			$ret->frontend->supported = true;
			$ret->frontend->path      = 'index.php?option=com_ats&view=Cron&key=' . urlencode($ret->info->secret) .
				'&command=' . urlencode($commandName);
		}

		return $ret;
	}

	/**
	 * Get the currently selected CRON command
	 *
	 * @return  string
	 *
	 * @since   3.2.0
	 */
	public function getSelectedCRONCommand()
	{
		$commandName = $this->container->platform->getSessionVar('commandName', 'commandName', 'ats.cron');
		$allCommands = $this->getActiveCRONCommands();

		// Check for validity
		if (!array_key_exists($commandName, $allCommands))
		{
			$commandName = 'commandName';
		}

		return $commandName;
	}

	/**
	 * Return the information about the currently enabled ATS CRON commands
	 *
	 * @return  array
	 *
	 * @since   3.2.0
	 */
	public function getActiveCRONCommands()
	{
		if (!is_null($this->atsCronCommands))
		{
			return $this->atsCronCommands;
		}

		$current = $this->container->platform->isAllowPluginsInCli();

		if ($this->container->platform->isCli())
		{
			$this->container->platform->setAllowPluginsInCli(true);
		}

		$this->container->platform->importPlugin('ats');

		if ($this->container->platform->isCli())
		{
			$this->container->platform->setAllowPluginsInCli($current);
		}

		$responses = $this->container->platform->runPlugins('onAtsCronTaskInfo', []);

		$this->atsCronCommands = array_filter($responses, function ($response) {
			if (empty($response) || !is_array($response))
			{
				return false;
			}

			foreach (['command', 'label', 'description'] as $key)
			{
				if (!array_key_exists($key, $response))
				{
					return false;
				}

				if (empty($response[$key]))
				{
					return false;
				}
			}

			return true;
		});

		$keys = array_map(function ($info) {
			return $info['command'];
		}, $this->atsCronCommands);

		$this->atsCronCommands = array_combine($keys, $this->atsCronCommands);

		return $this->atsCronCommands;
	}

	/**
	 * Get a list of all of the installed ATS CRON plugins
	 *
	 * @return  array
	 *
	 * @since   3.2.0
	 */
	public function getAtsCronPlugins()
	{
		// Return the cached results if available
		if (!is_null($this->atsCronPlugins))
		{
			return $this->atsCronPlugins;
		}

		// Default: no plugins detected
		$this->atsCronPlugins = [];

		// Get the information for all installed ATS plugins
		try
		{
			$pluginsInfo = $this->getAtsPluginsFromDB();

			if (is_null($pluginsInfo))
			{
				throw new RuntimeException('No ats plugins are installed on this site');
			}
		}
		catch (RuntimeException $e)
		{
			return $this->atsCronPlugins;
		}

		// Filter by plugins with an onAtsCronTaskInfo method
		$pluginsInfo = array_filter($pluginsInfo, function (array $pluginInfo) {
			$pluginPath  = JPATH_PLUGINS . '/ats/' . $pluginInfo['element'] . '/' . $pluginInfo['element'] . '.php';
			$pluginClass = 'plgAts' . ucfirst($pluginInfo['element']);

			// Does the plugin even exist?
			if (!@file_exists($pluginPath))
			{
				return false;
			}

			// If the plugin is already loaded I can check if the method exists directly
			if (class_exists($pluginClass))
			{
				return method_exists($pluginClass, 'onAtsCronTaskInfo');
			}

			$fileContents = @file_get_contents($pluginPath);

			if ($fileContents === false)
			{
				return false;
			}

			return preg_match('/function(\s){1,}onAtsCronTaskInfo/', $fileContents) > 0;
		});

		if (empty($pluginsInfo))
		{
			return $this->atsCronPlugins;
		}

		// Process the plugin info into something usable
		$this->atsCronPlugins = array_map(function ($info) {
			$params           = new Registry($info['params']);
			$manifest         = new Registry($info['manifest_cache']);
			$cronValue        = $params->get('cron', null);
			$user             = new User();
			$publicViewLevels = $user->getAuthorisedViewLevels();

			$this->container->platform->getLanguage()->load('plg_ats_' . $info['element'] . '.sys', JPATH_ADMINISTRATOR, null, false, true);

			$ret = [
				'id'          => $info['extension_id'],
				'name'        => Text::_($info['name']),
				'command'     => $info['element'],
				'description' => Text::_($manifest->get('description', '')),
				'published'   => $info['enabled'] == 1,
				'public'      => in_array($info['access'], $publicViewLevels),
				'cronEnabled' => is_null($cronValue) || ($cronValue == 1),
			];

			return $ret;
		}, $pluginsInfo);

		return $this->atsCronPlugins;
	}

	/**
	 * Publish an ATS plugin with public access (or set the access to public if it's already published)
	 *
	 * @param   string  $pluginName  The name of the ats plugin
	 *
	 * @return  bool  True if the operation succeeded or the plugin was already published with a public access level.
	 *
	 * @since   3.2.0
	 */
	public function publishPlugin($pluginName)
	{
		// Get the filtered plugin name
		$filter     = new InputFilter();
		$pluginName = $filter->clean($pluginName, 'cmd');

		if (empty($pluginName) || is_array($pluginName))
		{
			return false;
		}

		// Make sure the plugin in question exists
		$atsCronPlugins = $this->getAtsCronPlugins();

		if (!array_key_exists($pluginName, $atsCronPlugins))
		{
			return false;
		}

		// Make sure the plugin in question is not already published with a public access level and CRON enabled
		$pluginInfo = $atsCronPlugins[$pluginName];
		$pluginId   = $pluginInfo['id'];

		if ($pluginInfo['published'] && $pluginInfo['public'] && $pluginInfo['cronEnabled'])
		{
			return true;
		}

		$db      = $this->container->db;
		$changes = [];

		// Do I need to publish the plugin?
		if (!$pluginInfo['published'])
		{
			$changes[] = $db->qn('enabled') . ' = 1';
		}

		// Do I need to make the plugin public?
		if (!$pluginInfo['public'])
		{
			// Get the first public view access level
			$user             = new User();
			$publicViewLevels = $user->getAuthorisedViewLevels();

			if (empty($publicViewLevels))
			{
				return false;
			}

			$publicLevel = min($publicViewLevels);

			if (empty($publicLevel))
			{
				return false;
			}

			$changes[] = $db->qn('access') . ' = ' . $db->q($publicLevel);
		}

		try
		{
			if (!empty($changes))
			{
				$query = $db->getQuery(true)
					->update($db->qn('#__extensions'))
					->set(
						implode(',', $changes)
					)
					->where($db->qn('extension_id') . ' = ' . $db->q($pluginId));

				$db->setQuery($query)->execute();
			}

			$retResult = true;

			// Do I need to Enable CRON for this plugin?
			if (!$pluginInfo['cronEnabled'])
			{
				$params = PluginParameters::getPluginParams($pluginName);
				$params->set('cron', 1);

				$retResult = PluginParameters::savePluginParams($pluginName, $params);
			}
		}
		catch (RuntimeException $e)
		{
			$retResult = false;
		}
		finally
		{
			CacheCleaner::clearPluginsCache();

			return $retResult;
		}
	}

	/**
	 * Has the user configured a Secret Word for CRON jobs?
	 *
	 * @return  bool
	 *
	 * @since   3.2.0
	 */
	public function hasSecretWord()
	{
		$secret = trim($this->container->params->get('secret', ''));

		return !empty($secret);
	}

	/**
	 * Generate a new, random Secret Word and store it in the session
	 *
	 * @return  void
	 *
	 * @since   3.2.0
	 */
	public function createNewSecretWord()
	{
		$random    = new Randval();
		$newSecret = $random->getRandomPassword(32);
		$this->container->platform->setSessionVar('newSecretWord', $newSecret, 'ats.cron');
	}

	/**
	 * Apply the given secret. If none is provided the already generated, random Secret Word will be used instead.
	 *
	 * @param   string  $secret  The secret word to apply. Empty to use the already generated, random Secret Word.
	 *
	 * @return  void
	 *
	 * @since   3.2.0
	 */
	public function applySecretWord($secret = '')
	{
		$secret = trim($secret);

		if (empty($secret))
		{
			$secret = $this->container->platform->getSessionVar('newSecretWord', '', 'ats.cron');
		}

		if (empty($secret))
		{
			$this->createNewSecretWord();

			$secret = $this->container->platform->getSessionVar('newSecretWord', '', 'ats.cron');
		}

		$this->container->params->set('secret', $secret);
		$this->container->params->save();
	}

	/**
	 * Get a list of installed ATS plugins from the database
	 *
	 * @return  array|null
	 *
	 * @since   3.2.0
	 */
	protected function getAtsPluginsFromDB()
	{
		$db = $this->container->db;

		$query = $db->getQuery(true)
			->select(array_map([$db, 'qn'], [
				'extension_id', 'name', 'element', 'enabled', 'access', 'params', 'manifest_cache',
			]))
			->from($db->qn('#__extensions'))
			->where($db->qn('type') . ' = ' . $db->q('plugin'))
			->where($db->qn('folder') . ' = ' . $db->q('ats'));

		return $db->setQuery($query)->loadAssocList('element');
	}
}