<?php
/**
 * @package   ats
 * @copyright Copyright (c)2011-2022 Nicholas K. Dionysopoulos / Akeeba Ltd
 * @license   GNU General Public License version 3, or later
 */

namespace Akeeba\TicketSystem\Admin\Model;

defined('_JEXEC') or die;

use Exception;
use FOF40\Container\Container;
use FOF40\Update\Update;
use Joomla\CMS\Factory as JFactory;
use Joomla\CMS\Filesystem\File as JFile;

class Updates extends Update
{
	/**
	 * Public constructor. Initialises the protected members as well.
	 *
	 * @param   array  $config
	 */
	public function __construct($config = [])
	{
		$container = Container::getInstance('com_ats');

		$config['update_component'] = 'pkg_ats';
		$config['update_sitename']  = 'Akeeba Ticket System Core';
		$config['update_site']      = 'https://cdn.akeeba.com/updates/atscore.xml';
		$config['update_paramskey'] = 'downloadid';
		$config['update_container'] = $container;

		$isPro = defined('ATS_PRO') && ATS_PRO;

		if ($isPro)
		{
			$config['update_sitename'] = 'Akeeba Ticket System Professional';
			$config['update_site']     = 'https://cdn.akeeba.com/updates/atspro.xml';
		}

		if (defined('ATS_VERSION') && !in_array(substr(ATS_VERSION, 0, 3), ['dev', 'rev']))
		{
			$config['update_version'] = ATS_VERSION;
		}

		parent::__construct($config);

		$this->container    = $container;
		$this->extension_id = $this->findExtensionId('pkg_ats', 'package');

		if (empty($this->extension_id))
		{
			$this->createFakePackageExtension();
			$this->extension_id = $this->findExtensionId('pkg_ats', 'package');
		}
	}

	/**
	 * Sends an update notification email
	 *
	 * @param   string  $version  The newest available version
	 * @param   string  $email    The email address of the recipient
	 *
	 * @return  boolean  The result from JMailer::send()
	 */
	public function sendNotificationEmail($version, $email)
	{
		$email_subject = <<<ENDSUBJECT
THIS EMAIL IS SENT FROM YOUR SITE "[SITENAME]" - Update available
ENDSUBJECT;

		$email_body = <<<ENDBODY
This email IS NOT sent by the authors of Akeeba Ticket System. It is sent automatically
by your own site, [SITENAME]

================================================================================
UPDATE INFORMATION
================================================================================

Your site has determined that there is an updated version of Akeeba Ticket System
available for download.

New version number: [VERSION]

This email is sent to you by your site to remind you of this fact. The authors
of the software will never contact you about available updates.

================================================================================
WHY AM I RECEIVING THIS EMAIL?
================================================================================

This email has been automatically sent by a CLI script you, or the person who built
or manages your site, has installed and explicitly activated. This script looks
for updated versions of the software and sends an email notification to all
Super Users. You will receive several similar emails from your site, up to 6
times per day, until you either update the software or disable these emails.

To disable these emails, please contact your site administrator.

If you do not understand what this means, please do not contact the authors of
the software. They are NOT sending you this email and they cannot help you.
Instead, please contact the person who built or manages your site.

================================================================================
WHO SENT ME THIS EMAIL?
================================================================================

This email is sent to you by your own site, [SITENAME]

ENDBODY;

		$jconfig  = $this->container->platform->getConfig();
		$sitename = $jconfig->get('sitename');

		$substitutions = [
			'[VERSION]'  => $version,
			'[SITENAME]' => $sitename,
		];

		$email_subject = str_replace(array_keys($substitutions), array_values($substitutions), $email_subject);
		$email_body    = str_replace(array_keys($substitutions), array_values($substitutions), $email_body);

		try
		{
			$mailer = JFactory::getMailer();

			$mailfrom = $jconfig->get('mailfrom');
			$fromname = $jconfig->get('fromname');

			$mailer->setSender([$mailfrom, $fromname]);
			$mailer->addRecipient($email);
			$mailer->setSubject($email_subject);
			$mailer->setBody($email_body);

			return $mailer->Send();
		}
		catch (\Exception $e)
		{
			return false;
		}
	}

	/**
	 * Gets the ID of an extension
	 *
	 * @param   string  $element  Extension element, e.g. com_foo, mod_foo, lib_foo, pkg_foo or foo (CAUTION: plugin,
	 *                            file!)
	 * @param   string  $type     Extension type: component, module, library, package, plugin or file
	 * @param   null    $folder   Plugins: plugin folder. Modules: admin/site
	 *
	 * @return  int  Extension ID or 0 on failure
	 */
	private function findExtensionId($element, $type = 'component', $folder = null)
	{
		$db    = $this->container->db;
		$query = $db->getQuery(true)
			->select($db->qn('extension_id'))
			->from($db->qn('#__extensions'))
			->where($db->qn('element') . ' = ' . $db->q($element))
			->where($db->qn('type') . ' = ' . $db->q($type));

		// Plugin? We should look for a folder
		if ($type == 'plugin')
		{
			$folder = empty($folder) ? 'system' : $folder;

			$query->where($db->qn('folder') . ' = ' . $db->q($folder));
		}

		// Module? Use the folder to determine if it's site or admin module.
		if ($type == 'module')
		{
			$folder = empty($folder) ? 'site' : $folder;

			$query->where($db->qn('client_id') . ' = ' . $db->q(($folder == 'site') ? 0 : 1));
		}

		try
		{
			$id = $db->setQuery($query, 0, 1)->loadResult();
		}
		catch (Exception $e)
		{
			$id = 0;
		}

		return empty($id) ? 0 : (int) $id;
	}

	private function createFakePackageExtension()
	{
		$db = $this->container->db;

		$query = $db->getQuery(true)
			->insert($db->qn('#__extensions'))
			->columns([
				$db->qn('name'), $db->qn('type'), $db->qn('element'), $db->qn('folder'), $db->qn('client_id'),
				$db->qn('enabled'), $db->qn('access'), $db->qn('protected'), $db->qn('manifest_cache'),
				$db->qn('params'), $db->qn('custom_data'), $db->qn('system_data'), $db->qn('checked_out'),
				$db->qn('checked_out_time'), $db->qn('ordering'), $db->qn('state'),
			])
			->values([
				$db->q('Akeeba Ticket System package') . ',' .
				$db->q('package') . ',' .
				$db->q('pkg_ats') . ',' .
				$db->q('') . ',' .
				$db->q(0) . ',' .
				$db->q(1) . ',' .
				$db->q(1) . ',' .
				$db->q(0) . ',' .
				$db->q('{"name":"Akeeba Ticket System package","type":"package","creationDate":"2016-04-21","author":"Nicholas K. Dionysopoulos","copyright":"Copyright (c)2011-2017 Akeeba Ltd \/ Nicholas K. Dionysopoulos","authorEmail":"","authorUrl":"","version":"' . $this->version . '","description":"Akeeba Ticket System installation package, for updating from versions older than 2.2.x only","group":"","filename":"pkg_ats"}') . ',' .
				$db->q('{}') . ',' .
				$db->q('') . ',' .
				$db->q('') . ',' .
				$db->q(0) . ',' .
				$db->q($db->getNullDate()) . ',' .
				$db->q(0) . ',' .
				$db->q(0),
			]);

		try
		{
			$db->setQuery($query)->execute();
		}
		catch (\Exception $e)
		{
			// Your database if FUBAR.
			return;
		}

		$this->createFakePackageManifest();
	}

	private function createFakePackageManifest()
	{
		$path = JPATH_ADMINISTRATOR . '/manifests/packages/pkg_ats.xml';

		if (file_exists($path))
		{
			return;
		}

		$isPro = defined('ATS_PRO') && ATS_PRO;
		$dlid  = $isPro ? '<dlid prefix="dlid=" suffix=""/>' : '';

		if ($isPro)
		{
			$extensions = <<< XML
    <files>
        <!-- Component -->
        <file type="component" id="com_ats">com_ats-pro.zip</file>
		<!-- CLI scripts -->
        <file type="file" id="file_ats">file_ats-pro.zip</file>

        <!-- Modules -->
        <file type="module" client="administrator" id="mod_atsstats">mod_atsstats.zip</file>

        <file type="module" client="site" id="mod_atscredits">mod_atscredits.zip</file>
        <file type="module" client="site" id="mod_atstickets">mod_atstickets.zip</file>

        <!-- Plugins: actionlog -->
        <file type="plugin" group="actionlog" id="ats">plg_actionlog_ats.zip</file>

        <!-- Plugins: ats -->
        <file type="plugin" group="ats" id="akeebasubs">plg_ats_akeebasubs.zip</file>
        <file type="plugin" group="ats" id="autoclose">plg_ats_autoclose.zip</file>
        <file type="plugin" group="ats" id="autoreply">plg_ats_autoreply.zip</file>
        <file type="plugin" group="ats" id="customfields">plg_ats_customfields.zip</file>
        <file type="plugin" group="ats" id="deletenotes">plg_ats_deletenotes.zip</file>
        <file type="plugin" group="ats" id="geolocation">plg_ats_geolocation.zip</file>
        <file type="plugin" group="ats" id="gravatar">plg_ats_gravatar.zip</file>
        <file type="plugin" group="ats" id="easyavatar">plg_ats_easyavatar.zip</file>
        <file type="plugin" group="ats" id="mailfetch">plg_ats_mailfetch.zip</file>
        <file type="plugin" group="ats" id="postemail">plg_ats_postemail.zip</file>
        <file type="plugin" group="ats" id="removeattachments">plg_ats_removeattachments.zip</file>
        <file type="plugin" group="ats" id="sociallike">plg_ats_sociallike.zip</file>
        <file type="plugin" group="ats" id="usergroups">plg_ats_usergroups.zip</file>

        <!-- Plugins: content -->
		<file type="plugin" group="content" id="atscredits">plg_content_atscredits.zip</file>

        <!-- Plugins: installer -->
		<file type="plugin" group="installer" id="ats">plg_installer_ats.zip</file>

        <!-- Plugins: finder -->
		<file type="plugin" group="finder" id="ats">plg_finder_ats.zip</file>

		<!-- Plugins: search -->
		<file type="plugin" group="search" id="ats">plg_search_ats.zip</file>

		<!-- Plugins: sh404sefextplugins -->
		<file type="plugin" group="sh404sefextplugins" id="com_ats">plg_sh404sefextplugins_com_ats.zip</file>

		<!-- Plugins: user -->
		<file type="plugin" group="user" id="ats">plg_user_ats.zip</file>
	</files>
XML;
		}
		else
		{
			$extensions = <<< XML
    <files>
        <!-- Component -->
        <file type="component" id="com_ats">com_ats-core.zip</file>

        <!-- Modules -->
        <file type="module" client="admin" id="mod_atsstats">mod_atsstats.zip</file>

        <file type="module" client="site" id="mod_atscredits">mod_atscredits.zip</file>
        <file type="module" client="site" id="mod_atstickets">mod_atstickets.zip</file>

        <!-- Plugins: actionlog -->
        <file type="plugin" group="actionlog" id="ats">plg_actionlog_ats.zip</file>

        <!-- Plugins: ats -->
        <file type="plugin" group="ats" id="akeebasubs">plg_ats_akeebasubs.zip</file>
        <file type="plugin" group="ats" id="customfields">plg_ats_customfields.zip</file>
        <file type="plugin" group="ats" id="gravatar">plg_ats_gravatar.zip</file>

        <!-- Plugins: finder -->
		<file type="plugin" group="finder" id="ats">plg_finder_ats.zip</file>

		<!-- Plugins: search -->
		<file type="plugin" group="search" id="ats">plg_search_ats.zip</file>

        <!-- Plugins: sh404sefextplugins -->
        <file type="plugin" group="sh404sefextplugins" id="com_ats">plg_sh404sefextplugins_com_ats.zip</file>

        <!-- Plugins: user -->
		<file type="plugin" group="user" id="ats">plg_user_ats.zip</file>
    </files>
XML;
		}

		$content = <<< XML
<?xml version="1.0" encoding="utf-8"?>
<extension version="3.9.0" type="package" method="upgrade">
	$dlid
    <name>Akeeba Ticket System package</name>
    <author>Nicholas K. Dionysopoulos</author>
    <creationDate>2016-04-20</creationDate>
    <packagename>ats</packagename>
    <version>{$this->version}</version>
    <url>https://www.akeeba.com</url>
    <packager>Akeeba Ltd</packager>
    <packagerurl>https://www.akeeba.com</packagerurl>
    <copyright>Copyright (c)2011-2019 Nicholas K. Dionysopoulos</copyright>
    <license>GNU GPL v3 or later</license>
    <description>Akeeba Ticket System installation package v.{$this->version}</description>

    <!-- List of extensions to install -->
{$extensions}

    <!-- Installation script -->
    <scriptfile>script.ats.php</scriptfile>

    <!-- Update servers -->
    <updateservers>
        <server type="extension" priority="1" name="Akeeba Ticket System">https://cdn.akeeba.com/updates/atspro.xml</server>
    </updateservers>
</extension>
XML;

		if (!@file_put_contents($content, $path))
		{
			JFile::write($path, $content);
		}
	}
}
