<?php
/**
 * @package   ats
 * @copyright Copyright (c)2011-2022 Nicholas K. Dionysopoulos / Akeeba Ltd
 * @license   GNU General Public License version 3, or later
 */

namespace Akeeba\TicketSystem\Site\View\NewTickets;

defined('_JEXEC') or die;

use Akeeba\TicketSystem\Admin\Helper\Bbcode;
use Akeeba\TicketSystem\Admin\Helper\ComponentParams;
use Akeeba\TicketSystem\Admin\Helper\Credits;
use Akeeba\TicketSystem\Admin\Helper\Editor;
use Akeeba\TicketSystem\Admin\Helper\Permissions;
use Akeeba\TicketSystem\Site\Model\Categories;
use Akeeba\TicketSystem\Site\Model\Tickets;
use Akeeba\TicketSystem\Site\View\PageMetaAware;
use Joomla\CMS\Factory;
use Joomla\CMS\Router\Route;

class Html extends \FOF40\View\DataView\Html
{
	use PageMetaAware;

	/**
	 * Is the current user a manager?
	 *
	 * @var   bool
	 */
	public $isManager = false;

	/**
	 * The record loaded (read, edit, add views)
	 *
	 * @var   Tickets
	 */
	public $item = null;

	/**
	 * The ticket category in which we're filing a new ticket
	 *
	 * @var   Categories
	 */
	public $category;

	/**
	 * Should I show credits information?
	 *
	 * @var   bool
	 */
	public $show_credits = false;

	/**
	 * Do I have enough credits for a public ticket?
	 *
	 * @var   bool
	 */
	public $enough_credits_public = true;

	/**
	 * Do I have enough credits for a private ticket?
	 *
	 * @var   bool
	 */
	public $enough_credits_private = true;

	/**
	 * Credits required for filing a public ticket
	 *
	 * @var   int
	 */
	public $credits_public = 0;

	/**
	 * Credits required for filing a private ticket
	 *
	 * @var   int
	 */
	public $credits_private = 0;

	/**
	 * Joomla menu item ID
	 *
	 * @var   int
	 * @since 3.0.0
	 */
	public $itemID;

	/**
	 * Forced ticket type: PUB, PRIV or empty if it's not forced
	 *
	 * @var   string
	 * @since 3.0.0
	 */
	public $forceType;

	/**
	 * Is this ticket private by default?
	 *
	 * @var   bool
	 * @since 3.0.0.
	 */
	public $ticketPrivate;

	/**
	 * Custom field definitions
	 *
	 * @var   array[]
	 * @since 3.0.0
	 */
	public $customFields;

	/**
	 * Is the default setting of this category to create private tickets?
	 *
	 * @var   bool
	 * @since 3.0.0
	 */
	public $defaultPrivate;

	/**
	 * Pre-populated post content
	 *
	 * @var   string
	 * @since 3.0.0
	 */
	public $post_content;

	/**
	 * The state of the category filter for thelanding page
	 *
	 * @var   int
	 * @since 3.0.0
	 */
	public $categoryFilter;

	/**
	 * Should I display the ticker priority selection?
	 *
	 * @var   bool
	 * @since 3.0.0
	 */
	public $displayPriorities;

	/**
	 * Total amount of credits currently available to the user
	 *
	 * @var   int
	 * @since 3.0.0
	 */
	public $totalCreditsLeft;

	/**
	 * Default priority for new tickets
	 *
	 * @var   int
	 */
	public $ticketPriority;

	protected function onBeforeBrowse()
	{
		/** @var \JApplicationSite $app */
		$app              = Factory::getApplication();
		$params           = $app->getParams();
		$this->pageParams = $params;

		parent::onBeforeBrowse();

		$this->itemID         = $this->input->getInt('Itemid', 0);
		$this->categoryFilter = $this->getModel()->getState('category');

		// Is the "No new tickets" killswitch activated?
		if ($this->container->params->get('nonewtickets', 0))
		{
			$this->setLayout('nonewtickets');

			return false;
		}

		if ($this->getLayout() == 'landing')
		{
			return true;
		}
	}

	protected function onBeforeAdd()
	{
		$container = $this->container;
		$container->template->addJS('media://com_ats/js/NewTicket.min.js', true, false, $container->mediaVersion);
		$container->template->addJS('media://com_ats/js/TicketOverlays.min.js', true, false, $container->mediaVersion);

		/** @var \JApplicationSite $app */
		$app              = Factory::getApplication();
		$params           = $app->getParams();
		$this->pageParams = $params;

		// Assign the ItemID before any sanity check, so we will fetch the correct menu
		$this->itemID = $this->input->getInt('Itemid', 0);

		parent::onBeforeAdd();

		// Is the "No new tickets" killswitch activated?
		if ($container->params->get('nonewtickets', 0))
		{
			$this->setLayout('nonewtickets');

			return false;
		}

		if ($this->getLayout() == 'landing')
		{
			return true;
		}

		$container = $container;
		$platform  = $container->platform;
		$user      = Permissions::getUser();
		$container->platform->importPlugin('ats');

		// Fire onTicketsAdd event, maybe I have anything that changes the
		// user credit balance (ie AlphaUserPoint)?
		$container->platform->runPlugins('onTicketsBeforeAdd', [&$this->item, $this->category, Permissions::getUser()]);

		if ($container->params->get('showcredits', 0) && $container->params->get('ticketPriorities', 0))
		{
			$container->platform->addScriptOptions('ats.PriorityCredits.enabled', true);
		}

		if ($this->category->params->get('instantreply', 0))
		{
			$proxyURL = Route::_('index.php?option=com_ats&view=InstantReply&task=browse&catid=' . $this->category->id, false);
			$proxyURL .= strpos($proxyURL, '?') !== false ? '&' : '?';
			$proxyURL .= 'format=json';

			$container->template->addJS('media://com_ats/js/InstantReply.min.js', true, false, $container->mediaVersion);

			$container->platform->addScriptOptions('ats.InstantReply.enabled', true);
			$container->platform->addScriptOptions('ats.InstantReply.catID', $this->category->id);

		}

		$this->isManager = Permissions::isManager($this->category->id);

		$this->show_credits           = $container->params->get('showcredits', 0);
		$this->enough_credits_public  = true;
		$this->enough_credits_private = true;

		// Perform checks only if requested to
		$this->forceType = $this->category->params->get('forcetype');

		if (!$this->isManager && $this->show_credits)
		{
			/** @var Categories $catModel */
			$catModel = $container->factory->model('Categories')->tmpInstance();
			$catModel->find($this->category->id);

			// First of all, let's check if the user can post anything with the lower credit cost
			// (ie public ticket with low priority)
			if ($container->params->get('ticketPriorities'))
			{
				$this->enough_credits_public  = Credits::haveEnoughCredits($user->id, $this->category->id, true, true, false);
				$this->enough_credits_private = Credits::haveEnoughCredits($user->id, $this->category->id, true, false, false);
			}
			else
			{
				$this->enough_credits_public  = Credits::haveEnoughCredits($user->id, $this->category->id, true, true, 5);
				$this->enough_credits_private = Credits::haveEnoughCredits($user->id, $this->category->id, true, false, 1);
			}

			// Let's get credit cost on "default" values
			$this->credits_public  = Credits::creditsRequired($this->category->id, true, true, 5);
			$this->credits_private = Credits::creditsRequired($this->category->id, true, false, 1);

			// If the user can't do anything, neither post on low priority, let's stop here
			// Those tests are run vs default visibility, not the one coming from the session. However, if the user could select it, it means
			// that can use it (checks are performed before saving, too)
			if (
				(!$this->enough_credits_public && !$this->enough_credits_private) ||                         // Not enough credits for anything
				(!$this->enough_credits_public && $this->enough_credits_private && !$this->allow_private) || // Not enough credits for public posting
				($this->forceType == 'PRIV' && !$this->enough_credits_private) || // Forced private but not enough credits
				($this->forceType == 'PUB' && !$this->enough_credits_public)                                      // Forced public but not enough credits
			)
			{
				$this->setLayout('notenoughcredits');
			}
		}
		elseif ($this->show_credits)
		{
			$this->enough_credits_public  = true;
			$this->enough_credits_private = true;
			$this->credits_public         = 0;
			$this->credits_private        = 0;
		}

		// Common variables
		$this->defaultPrivate    = (bool) $this->category->params->get('defaultprivate', 0);
		$this->ticketPrivate     = !is_null($this->ticket_public) ? !$this->ticket_public : $this->defaultPrivate;
		$this->ticketPrivate     = $this->forceType ? (int) ($this->forceType == 'PRIV') : $this->ticketPrivate;
		$this->customFields      = $this->item->getCustomFieldDefinitions($this->category->id);
		$this->displayPriorities = $container->params->get('ticketPriorities', 0) && ATS_PRO;
		$this->totalCreditsLeft  = Credits::creditsLeft(Permissions::getUser()->id, true);

		// Fetch default priority for new tickets (private HIGH, public NORMAL)
		$this->ticketPriority    = $this->defaultPrivate ? 1 : 5;
		$defaultCategoryPriority = (int) $this->category->params->get('default_priority', '');

		// Do I have a specific value for this category? If so, let's use it
		if ($defaultCategoryPriority)
		{
			$this->ticketPriority = $defaultCategoryPriority;
		}

		if (empty($this->post_content))
		{
			$this->post_content = $this->category->params->get('defposttext', '');
			$this->post_content = Bbcode::parseBBCodeConditional($this->post_content);
		}

		$this->setPageMeta();

		return true;
	}
}
