<?php
/**
 * @package   ats
 * @copyright Copyright (c)2011-2022 Nicholas K. Dionysopoulos / Akeeba Ltd
 * @license   GNU General Public License version 3, or later
 */

namespace Akeeba\TicketSystem\MailFetch;

defined('_JEXEC') or die;

use Akeeba\TicketSystem\Admin\Helper\AddonEmails;
use Akeeba\TicketSystem\Admin\Helper\Credits;
use Akeeba\TicketSystem\Admin\Helper\Debug;
use Akeeba\TicketSystem\Admin\Helper\Email;
use Akeeba\TicketSystem\Admin\Helper\Filter;
use Akeeba\TicketSystem\Admin\Helper\JoomlaUsers as JoomlaUsersHelper;
use Akeeba\TicketSystem\Admin\Helper\Permissions;
use Akeeba\TicketSystem\Admin\Helper\PluginParameters;
use Akeeba\TicketSystem\Admin\Model\Attachments;
use Akeeba\TicketSystem\Admin\Model\Categories;
use Akeeba\TicketSystem\Admin\Model\Posts;
use Akeeba\TicketSystem\Admin\Model\Tickets;
use Akeeba\TicketSystem\Site\Helper\Text2Html;
use Akeeba\TicketSystem\Site\Model\InstantReplies;
use Exception;
use FOF40\Container\Container;
use FOF40\Date\Date;
use FOF40\Model\DataModel\Collection as DataCollection;
use FOF40\Timer\Timer;
use Joomla\CMS\Language\Text;
use Joomla\CMS\Log\Log as JLog;
use Joomla\CMS\User\User as JUser;
use Joomla\Registry\Registry;

class EmailCheck
{
	/**
	 * Have I already registered a logger for the email checks?
	 *
	 * @var   bool
	 * @since 3.2.0
	 */
	protected static $hasRegisteredLogger = false;

	/**
	 * The mail fetcher backend
	 *
	 * @var   MailClientInterface
	 * @since 3.2.0
	 */
	protected $backend;

	/**
	 * The ATS container
	 *
	 * @var   Container
	 * @since 3.2.0
	 */
	protected $container;

	/**
	 * The options of the mailfetch plugin
	 *
	 * @var   Registry
	 * @since 3.2.0
	 */
	protected $pluginOptions;

	/**
	 * @inheritDoc
	 */
	public function __construct(Container $container)
	{
		$this->container = $container;

		// Register a log file for email fetching
		if (!self::$hasRegisteredLogger)
		{
			Debug::registerFileLogger('ats.emails', 'ats_email_check.php');

			self::$hasRegisteredLogger = true;
		}

		// Populate the plugin options
		$this->populatePluginOptions();
	}

	/**
	 * Get the timestamp of the last email check
	 *
	 * @return int
	 */
	public function getLastTimestamp()
	{
		return $this->getPluginOption('email_lastcheck', 0);
	}

	/**
	 * Set the timestamp of the last email check to now
	 *
	 * @return  void
	 */
	public function setLastTimestamp($timestamp = null)
	{
		if (empty($timestamp))
		{
			$timestamp = time();
		}

		$this->pluginOptions->set('email_lastcheck', $timestamp);

		PluginParameters::savePluginParams('mailfetch', $this->pluginOptions);
	}

	/**
	 * Check for new email and process the messages
	 *
	 * @param   Timer  $timer  The timer object which controls execution time
	 *
	 * @since   3.2.0
	 */
	public function checkEmail(Timer $timer)
	{
		JLog::add("Beginning checking for emails", JLog::INFO, 'ats.emails');

		// Get parameters
		$replyByEmail = $this->getPluginOption('replybyemail', '0');
		$newByEmail   = $this->getPluginOption('newbyemail', '0');

		// Bail out if no email feature is set at all
		if (!$replyByEmail && !$newByEmail)
		{
			JLog::add("Reply by Email and Create Ticket by Email are both disabled.", JLog::WARNING, 'ats.emails');

			return;
		}

		// Get the mail backend
		try
		{
			JLog::add("Attempting to connect to mail server", JLog::INFO, 'ats.emails');

			$mailbox = $this->getMailbox();
		}
		catch (Exception $e)
		{
			JLog::add("Mail server connection failed.", JLog::ERROR, 'ats.emails');
			JLog::add($e->getMessage(), JLog::DEBUG, 'ats.emails');

			return;
		}

		// Check for emails since the last mail check
		$lastCheckDate = $this->getLastTimestamp();

		// If there is no last check date assume it was 24 hours ago
		if (empty($lastCheckDate))
		{
			$lastCheckDate = time() - 86400;
		}

		// Set the last check date to prevent this loop running in parallel several times
		$this->setLastTimestamp();

		$formattedSinceDate = date("r", $lastCheckDate - 24 * 3600);
		JLog::add(sprintf("Checking for unread email sent after %s", $formattedSinceDate), JLog::INFO, 'ats.emails');

		try
		{
			$mailbox->iterateUnread(function (MailMessage $email) use ($timer) {
				static $iterations = 0;

				// Check if we have enough time left to process another email. DO NOT REMOVE.
				if (($iterations > 0) && ($timer->getTimeLeft() <= 0.01))
				{
					throw new TimeoutException();
				}

				$debugLine = sprintf("%s. %s <%s> — %s", $email->mID, $email->fromName, $email->fromEmail, $email->subject);
				JLog::add($debugLine, JLog::INFO, 'ats.emails');

				return $this->processEmail($email);
			}, $formattedSinceDate);
		}
		catch (TimeoutException $e)
		{
			// Reset last check time so the next page load / CRON task will pick up where we left off.
			$this->setLastTimestamp($lastCheckDate);

			// The rethrow the exception so the plugin code knows we ran out of time
			throw $e;
		}
		catch (Exception $e)
		{
			JLog::add(sprintf("Failed to process email messages: [#%d] %s", $e->getCode(), $e->getMessage()), JLog::ERROR, 'ats.emails');
		}
	}

	/**
	 * Get a configuration parameter of the plg_ats_mailfetch plugin.
	 *
	 * @param   string  $key
	 * @param   mixed   $default
	 *
	 * @return  mixed
	 * @since   3.2.0
	 */
	public function getPluginOption($key, $default = null)
	{
		return $this->pluginOptions->get($key, $default);
	}

	/**
	 * Populate $this->pluginOptions with the options of the plg_ats_mailfetch plugin
	 *
	 * @return  void
	 * @since   3.2.0
	 */
	protected function populatePluginOptions()
	{
		$this->pluginOptions = PluginParameters::getPluginParams('mailfetch');
	}

	/**
	 * Return the mail client backend used to fetch and process incoming mail messages
	 *
	 * @return  MailClientInterface
	 *
	 * @throws  Exception  When connecting to the mail server fails for any reason
	 * @since   3.2.0
	 */
	private function getMailbox()
	{
		if (is_null($this->backend))
		{
			$mailFetcherParams = [
				'mailbox_type'         => $this->getPluginOption('mailboxtype', 'imap'),
				'server'               => $this->getPluginOption('email_server', ''),
				'port'                 => $this->getPluginOption('email_port', ''),
				'ssl'                  => $this->getPluginOption('email_ssl', '0'),
				'tls'                  => $this->getPluginOption('email_tls', '1'),
				'username'             => $this->getPluginOption('email_username', ''),
				'password'             => $this->getPluginOption('email_password', ''),
				'folder'               => $this->getPluginOption('email_folder', 'INBOX'),
				'access_token'         => '',
				'refresh_token'        => '',
				'client_id'            => '',
				'secret_key'           => '',
				'deleteAfter'          => $this->getPluginOption('email_delete', '1') == 1,
				'expunge'              => true,
			];

			if ($mailFetcherParams['mailbox_type'] == 'gmail')
			{
				$mailFetcherParams['access_token']  = $this->getPluginOption('access_token', '');
				$mailFetcherParams['refresh_token'] = $this->getPluginOption('refresh_token', '');
				$mailFetcherParams['client_id']     = $this->getPluginOption('gmail_client_id', '');
				$mailFetcherParams['secret_key']    = $this->getPluginOption('gmail_api_secret', '');
			}

			$this->backend = new Horde($mailFetcherParams);
		}

		return $this->backend;
	}

	/**
	 * Process an email message
	 *
	 * @param   MailMessage  $email  The email message to process
	 *
	 * @return  bool  True if it's successfully processed
	 *
	 * @since   3.2.0
	 */
	private function processEmail(MailMessage $email)
	{
		// Make sure we are not processing the same email twice (only if requested to)
		$trackUID = $this->getPluginOption('track_email_id', 0) != 0;

		if ($trackUID && $this->getUIDCount($email) > 0)
		{
			JLog::add(sprintf("Skipping already processed email with UID #%s", $email->uID), JLog::NOTICE, 'ats.emails');

			return false;
		}

		// Get the relevant component configuration
		$defaultCategory = $this->getPluginOption('email_default_category', '0');

		// Get the ticket ID, report and validate it
		$ticketID = $this->getTicketID($email);

		if (is_null($ticketID))
		{
			JLog::add('No TicketID found: this email is not a reply to an existing ticket', JLog::DEBUG, 'ats.emails');
		}
		else
		{
			JLog::add(sprintf("TicketID found: this email is a reply to the existing ticket #%d", $ticketID), JLog::DEBUG, 'ats.emails');
		}

		/** @var Tickets $ticket */
		$ticket = $this->container->factory->model('Tickets')->tmpInstance();

		if (!is_null($ticketID))
		{
			$foundTicket = $ticket->load($ticketID);

			if (!$foundTicket)
			{
				JLog::add(sprintf("Cannot find ticket #%d. Skipping this email.", $ticketID), JLog::ERROR, 'ats.emails');

				return false;
			}
		}

		// Get the category ID
		if (!$ticket->getId())
		{
			JLog::add("New ticket. Trying to determine category from the email address it was sent to.", JLog::DEBUG, 'ats.emails');

			$categoryID = $this->getCategoryIDFromEmail($email['toemail']);

			// If no category was found, use the default category
			if (empty($categoryID))
			{
				JLog::add(sprintf("No specific category corresponds to this email address; falling back to default category %d", $defaultCategory), JLog::DEBUG, 'ats.emails');

				$categoryID = $defaultCategory;
			}
			else
			{
				JLog::add(sprintf("Got category ID %d", $categoryID), JLog::DEBUG, 'ats.emails');
			}
		}
		else
		{
			$categoryID = $ticket->catid;

			JLog::add(sprintf("Got category ID %d from existing ticket #%u", $categoryID, $ticketID), JLog::DEBUG, 'ats.emails');
		}

		// Initialise
		/** @var JUser $user */
		$user = null;
		$from = $email->fromEmail;

		// Perform some generic checks. More specific ones are performed later
		if (!$this->passedBasicChecks($email, $categoryID, $ticket, $from, $user))
		{
			JLog::add("Generic checks failed for this email. Skipping this email.", JLog::ERROR, 'ats.emails');

			return false;
		}

		// Get the ACL privileges for the user
		$perms = Permissions::getAclPrivileges($categoryID, $user->id);

		/**
		 * The user account may have been found using an add-on email. We have to normalize this to the Joomla! user
		 * account email. This is the only way we can email ALL of the addresses (canonical and add-on) of the user.
		 */
		if ($user->id)
		{
			$from = $user->email;
		}

		$post      = [];
		$newTicket = !$ticket->getId();

		if ($newTicket)
		{
			JLog::add('Posting a new ticket', JLog::DEBUG, 'ats.emails');

			/**
			 * We have to decode the email subject, otherwise we end up with a wrong title when non UTF-8 chars are used
			 * (for example German umlauts)
			 */
			$title = $email['subject'];

			$data = [
				'ats_ticket_id' => null,
				'id'            => 0,
				'catid'         => $categoryID,
				'status'        => 'O',
				'title'         => $title,
				'public'        => !$perms['ats.private'],
				'origin'        => 'email',
				'enabled'       => 1,
				'created_by'    => $user->id,
			];

			try
			{
				$ticket::$overrideuser = 1;
				$ticket->save($data);
				$ticket::$overrideuser = 0;
			}
			catch (Exception $e)
			{
				JLog::add("Ticket creation failed. Raw error message: " . $e->getMessage(), JLog::ERROR, 'ats.emails');
				JLog::add("\tTraceback: " . $e->getTraceAsString(), JLog::ERROR, 'ats.emails');

				// Send an email informing the user that ticket creation failed
				$this->sendErrorEmail('mailgateway-newticketfailed', $from, [
					'name'     => $user->name,
					'username' => $user->username,
					'user_id'  => $user->id,
				]);

				return false;
			}
		}
		else
		{
			JLog::add("Ready to post a reply", JLog::DEBUG, 'ats.emails');
			$myTicket = $user->id == $ticket->created_by;

			$update = [
				'ats_ticket_id' => $ticket->ats_ticket_id,
				'status'        => 'O',
				'modified_by'   => $user->id,
				'modified_on'   => gmdate($this->container->db->getDateFormat()),
			];

			// Make sure the user is replying to his own ticket or he's an admin
			if ($myTicket)
			{
				JLog::add("Posting a reply from the ticket owner", JLog::DEBUG, 'ats.emails');

				// User posted: set ticket status to 'O'
				$update['status'] = 'O';
			}
			else
			{
				JLog::add("Posting a reply from a category manager", JLog::DEBUG, 'ats.emails');

				// Admin posted: set ticket status to 'P'
				$update['status'] = 'P';

				// If the user is a manager and there isn't an assigned user, let's assign him to the ticket
				if (!$ticket->assigned_to)
				{
					$update['assigned_to'] = $user->id;
				}
			}

			try
			{
				$ticket::$noPermissionsCheck = 1;
				$ticket->save($update);
				$ticket::$noPermissionsCheck = 0;
			}
			catch (Exception $e)
			{
				JLog::add("An error occurred updating the ticket status. Raw error: " . $e->getMessage(), JLog::ERROR, 'ats.emails');
				JLog::add("\tTraceback: " . $e->getTraceAsString(), JLog::ERROR, 'ats.emails');

				return false;
			}
		}

		$emailmessage = $this->extractMessage($email);
		$emailmessage = Filter::filterText($emailmessage, $user);

		// Add a post
		$post['content_html']      = $emailmessage;
		$post['content']           = '';
		$post['enabled']           = 1;
		$post['origin']            = 'email';
		$post['created_by']        = $user->id;
		$post['email_uid']         = $email['uid'];
		$post['ats_ticket_id']     = $ticket->getId();
		$post['ats_attachment_id'] = [0];

		if ($perms['ats.attachment'])
		{
			try
			{
				$attachments               = $this->extractAttachments($email, $user->id);
				$post['ats_attachment_id'] = $attachments;
			}
			catch (Exception $e)
			{
				/**
				 * This attachment failed to be uploaded, e.g. type mismatch. No problem; we are going to ignore it.
				 * If we don't we end up with a ticket that has no attachment *and* no body we're good.
				 */
			}
		}

		$hasAttachments = !empty($attachments);

		/**
		 * If there are no attachments we need to use an array filled with a single 0 so that we have a consistent value
		 * in the database.
		 */
		if (!$hasAttachments)
		{
			$post['ats_attachment_id'] = [0];
		}

		/** @var Posts $postModel */
		$postModel = $this->container->factory->model('Posts');
		// I have to do this trick instead of using the tmpInstance otherwise I lose the link with the parent ticket model
		$postModel->reset(true, false);
		/** @var Attachments $attachmentModel */
		$attachmentModel = $this->container->factory->model('Attachments')->tmpInstance();

		try
		{
			$postModel->save($post);
		}
		catch (Exception $e)
		{
			// If an error occurred after saving the post, I have to delete it as well
			if ($postModel->getId())
			{
				$postModel->delete();
			}

			if ($newTicket)
			{
				$ticket->delete();
			}

			if ($hasAttachments)
			{
				foreach ($attachments as $ats_attachment_id)
				{
					try
					{
						$attachmentModel->delete($ats_attachment_id);
					}
					catch (Exception $e)
					{
						// Don't die if we can't delete an orphan attachment.
					}
				}

			}

			JLog::add("Post creation failed. Raw error message: " . $e->getMessage(), JLog::ERROR, 'ats.emails');
			JLog::add("\tTraceback: " . $e->getTraceAsString(), JLog::ERROR, 'ats.emails');

			// Send an email informing the user that ticket creation failed
			$this->sendErrorEmail('mailgateway-newpostfailed', $from, [
				'name'     => $user->name,
				'username' => $user->username,
				'user_id'  => $user->id,
			]);

			return false;
		}

		// Update post reference inside attachment record
		if ($hasAttachments)
		{
			foreach ($attachments as $ats_attachment_id)
			{
				try
				{
					$attachmentModel->find($ats_attachment_id);
					$attachmentModel->ats_post_id = $postModel->ats_post_id;
					$attachmentModel->store();
				}
				catch (Exception $e)
				{
					// Don't die if we can't update an orphan attachment.
				}
			}
		}

		// Handle automatic sending of InstantReplies for new tickets
		if ($newTicket)
		{
			if ($this->getPluginOption('email_instantreply', 1) == 1)
			{
				$this->sendInstantReply($ticket);
			}
		}

		return true;
	}

	/**
	 * Returns the ticket ID found inside an email
	 *
	 * @param   MailMessage  $email
	 *
	 * @return  null|int  Null if no ID is found, else the ticket ID
	 */
	private function getTicketID(MailMessage $email)
	{
		// Ticket ID from the header (if any)
		$ticketId = empty($email->ticketID) ? 0 : $email->ticketID;

		// Ticket ID from the body (stated in "reply below this line")
		$hasTicketId  = preg_match('/{ticketid:(\d+)}/ism', $email->body, $matches);
		$bodyTicketId = $hasTicketId ? $matches[1] : 0;

		// Get the ticket ID that's non zero. If there are two, keep the correct one.
		$ticketId = max($ticketId, $bodyTicketId);

		return empty($ticketId) ? null : (int) $ticketId;
	}

	/**
	 * Extracts the part of the message which is above the reply line
	 *
	 * @param   array  $email
	 *
	 * @return string
	 */
	private function extractMessage(MailMessage $email)
	{
		// If it's a plain text email I need to convert it to HTML first.
		$body = $email->body;

		if (!$email->isHtml)
		{
			$body = Text2Html::convert($body);
		}

		// If there's a body tag return the content in the body
		$hasBody = preg_match_all('#(<body[^>]*>(.*?)</body>)#ims', $body, $matches);

		if ($hasBody)
		{
			$body = implode('', $matches[2]);
		}

		$parts = explode('!-!-', $body);

		return $parts[0];
	}

	/**
	 * Extract the attachments from the email message and store them in ATS. Returns an array with the generated ATS
	 * attachment IDs.
	 *
	 * @param   MailMessage  $email   The email message parts
	 * @param   int          $userID  The user ID applicable when saving the extracted attachments
	 *
	 * @return  array  The IDs of the attachments
	 *
	 * @throws  Exception
	 * @since   3.2.0
	 */
	private function extractAttachments($email, $userID)
	{
		if (!count($email->attachments))
		{
			return [];
		}

		// Get the temporary directory
		$tempdir = $this->container->platform->getConfig()->get('tmp_path');
		$ret     = [];

		foreach ($email->attachments as $attachment)
		{
			$size    = strlen($attachment['data']);
			$tmpName = tempnam($tempdir, 'atsup');
			$res     = @file_put_contents($tmpName, $attachment['data']);

			if ($res === false)
			{
				continue;
			}

			$fakeFileStructure = [
				// Joomla rejects files with spaces in their name. So I'll proactively substitute them with underscores.
				'name'     => str_replace(' ', '_', $attachment['filename']),
				'type'     => $attachment['mime'],
				'tmp_name' => $tmpName,
				'error'    => '',
				'size'     => $size,
			];

			/** @var Attachments $attachmentModel */
			$attachmentModel = $this->container->factory->model('Attachments')->tmpInstance();

			try
			{
				$fileDefinition = $attachmentModel->uploadFile($fakeFileStructure, false);
			}
			catch (Exception $e)
			{
				// Don't die.
				$fileDefinition = false;

				$debug = $e->getMessage();
			}

			if ($fileDefinition === false)
			{
				@unlink($tmpName);

				continue;
			}

			$fileDefinition['created_by'] = $userID;
			$fileDefinition['created_on'] = $this->container->platform->getDate()->toSql();
			$fileDefinition['enabled']    = 1;

			$attachmentModel->reset();
			$attachmentModel->save($fileDefinition);

			@unlink($tmpName);

			$ret[] = $attachmentModel->ats_attachment_id;
		}

		return $ret;
	}

	/**
	 * Does the message have a "reply above this line" line?
	 *
	 * @param   string  $body
	 *
	 * @return  bool
	 * @since   3.2.0
	 */
	private function hasReplyAboveLine($body)
	{
		$regex  = '#!-!-(.*)-!-!#s';
		$result = preg_match_all($regex, $body, $matches);

		return ($result !== false) && ($result > 0);
	}

	/**
	 * Gets the category ID based on the email address of the user
	 *
	 * @staticvar null|array $catmap Holds a map of emails to categories
	 *
	 * @param   string  $email  Email address to check
	 *
	 * @return  int  The category ID, or 0 if none is found
	 * @since     3.2.0
	 */
	private function getCategoryIDFromEmail($email)
	{
		static $catmap = null;

		if (is_null($catmap))
		{
			/** @var Categories $model */
			$model = $this->container->factory->model('Categories')->tmpInstance();
			$items = $model->ignoreUser(1)->get(true);

			$catmap = [];

			foreach ($items as $item)
			{
				$params   = new Registry($item->params);
				$catemail = $params->get('category_email', '');

				if (empty($catemail))
				{
					continue;
				}

				$catemail = strtolower($catemail);

				if (array_key_exists($catemail, $catmap))
				{
					continue;
				}

				$catmap[$catemail] = $item->id;
			}
		}

		$email = strtolower($email);

		if (array_key_exists($email, $catmap))
		{
			return $catmap[$email];
		}

		return 0;
	}

	/**
	 * Sends an error email response to a client
	 *
	 * @param   string  $key        Email key to load
	 * @param   string  $toEmail    Email address to send the email to
	 * @param   array   $extraVars  Any additional variables used during template parsing
	 *
	 * @return  bool
	 *
	 * @since   3.2.0
	 */
	private function sendErrorEmail($key, $toEmail, array $extraVars = [])
	{
		list($subject, $templateText, $loadLanguage) = Email::loadEmailTemplateFromDB($key);

		if (empty($templateText) || empty($subject))
		{
			return false;
		}

		try
		{
			$mailer = Email::getMailer();

			Email::parseTemplate($templateText, $subject, $extraVars, $mailer);

			$mailer->addRecipient($toEmail);
		}
		catch (Exception $e)
		{
			return false;
		}

		if (isset($extraVars['user_id']))
		{
			$addonEmails = AddonEmails::getAddonEmails($extraVars['user_id']);

			if (!empty($addonEmails))
			{
				foreach ($addonEmails as $email => $name)
				{
					try
					{
						$mailer->addCc($email, $name);
					}
					catch (Exception $e)
					{
					}
				}
			}
		}

		try
		{
			$mailer->Send();
			unset($mailer);
		}
		catch (Exception $e)
		{
			return false;
		}

		return true;
	}

	/**
	 * Perform basic checks, i.e. whether this email can be considered at all.
	 *
	 * @param   MailMessage  $email       The email message to process
	 * @param   int          $categoryID  The ATS category ID
	 * @param   Tickets      $ticket      The ticket this message will be processed into
	 * @param   string       $from        The email address of the sender
	 * @param   JUser        $user        The user record of the sender
	 *
	 * @return  bool
	 */
	private function passedBasicChecks(MailMessage $email, $categoryID, $ticket, &$from, &$user)
	{
		// Get a preliminary user from the email address. If no user is found we will get a blank JUser (Guest) user.
		$from = $email->fromEmail;
		$user = AddonEmails::findUserByAddonEmail($from);

		/** @var Categories $catModel */
		$catModel    = $this->container->factory->model('Categories')->tmpInstance();
		$hasCategory = false;

		if (!empty($categoryID))
		{
			$hasCategory = $catModel->load($categoryID);
		}

		// Do I have a valid category?
		if (!$hasCategory || ($catModel->getId() != $categoryID))
		{
			JLog::add("Cannot load the ticket category. This is a configuration error.", JLog::ERROR, 'ats.emails');
			JLog::add("Sending the mailgateway-noaccess email to the user.", JLog::DEBUG, 'ats.emails');

			// Send an email telling the user he has no access
			$this->sendErrorEmail('mailgateway-noaccess', $from, [
				'name'     => $user->name,
				'username' => $user->username,
				'user_id'  => $user->id,
			]);

			return false;
		}

		/**
		 * Does the user has access to the category?
		 *
		 * This requires two checks.
		 *
		 * First we try to load the category record passing the user ID. This applies the View Access Levels. If we get
		 * null instead of a record the View Access Levels forbid the user to even view the category, let alone post.
		 *
		 * Then we get the core.create privilege of the user for the category. This tells us if he can post to the
		 * category.
		 */
		$canCreate = false;
		$category  = null;

		if (!empty($categoryID))
		{
			JLog::add(sprintf("Attempting to load ticket category #%d", $categoryID), JLog::DEBUG, 'ats.emails');
			$categories = $catModel
				->category($categoryID)
				->userid($user->id)
				->get(true);
			/** @var Categories $category */
			$category = $categories->first();
		}

		// Get ACL permissions and make sure we can post here
		if (!is_null($category))
		{
			$perms     = Permissions::getAclPrivileges($category->id, $user->id);
			$canCreate = $perms['core.create'];
		}

		/**
		 * Can we actually create a ticket? There are three checks to perform IN THIS PRECISE ORDER:
		 * 1- No User, but guest is allowed. This will give us a User if needed
		 * 2- If we got a User at all. There's no point in continuing without
		 * 3- Check if we have enough permissions on the category
		 */

		// If the user is not found AND the category allows Guest to post, let's try creating a new user
		if (!$user->id && $canCreate)
		{
			JLog::add('This is a Guest User ticket (new ticket from a user we have not seen before).', JLog::DEBUG, 'ats.emails');

			$jUsersHelper  = new JoomlaUsersHelper($this->container);
			$forceActivate = $this->container->params->get('forceGuestActivation', 0);
			$fullName      = $email->fromName;

			if (empty($fullName))
			{
				$fullName = $from;
			}

			try
			{
				JLog::add("Trying to create a new user from this email address.", JLog::DEBUG, 'ats.emails');

				list ($activation, $userId) = $jUsersHelper->createUser($from, $fullName, $forceActivate);
			}
			catch (Exception $e)
			{
				JLog::add(sprintf('Guest User ticket failed. Cannot create a new user from email address %s.', $from), JLog::ERROR, 'ats.emails');

				$userId = -1;
			}

			if ($userId > 0)
			{
				$user = Permissions::getUser($userId);
				$from = $user->email;
			}
		}

		// Catch the case of not having a valid user record yet.
		if (!$user->id)
		{
			JLog::add("No user record can be found from the sender's email address.", JLog::ERROR, 'ats.emails');
			JLog::add("Sending the mailgateway-invaliduser email and aborting.", JLog::DEBUG, 'ats.emails');

			// Maybe send an email telling the user he screwed up if there's a valid ticket ID?
			$this->sendErrorEmail('mailgateway-invaliduser', $from);

			return false;
		}

		// No category found (wrong or no access to it) or the user can't create a ticket
		if (!$canCreate)
		{
			if (is_null($category))
			{
				JLog::add(sprintf("User %s is not allowed to view the category because of View Access Level settings.", $user->id), JLog::ERROR, 'ats.emails');
			}
			else
			{
				JLog::add(sprintf("User %s can view the category but is not allowed to create new tickets in it.", $user->id), JLog::ERROR, 'ats.emails');
			}

			JLog::add("Sending the mailgateway-newreplydisabled email to the user", JLog::DEBUG, 'ats.emails');

			// Send an email telling the user that replies by email are not allowed
			$this->sendErrorEmail('mailgateway-newreplydisabled', $from, [
				'name'     => $user->name,
				'username' => $user->username,
				'user_id'  => $user->id,
			]);

			return false;
		}

		/**
		 * The user account may have been found using an add-on email. We have to normalize this to the Joomla! user
		 * account email. This is the only way we can email ALL of the addresses (canonical and add-on) of the user.
		 */
		$from           = $user->email;
		$newByEmail     = $this->getPluginOption('newbyemail', 0);
		$replyByEmail   = $this->getPluginOption('replybyemail', '0');
		$emailAdminOnly = $this->getPluginOption('emailadminonly', 0);
		$noNewTickets   = $this->container->params->get('nonewtickets', 0);
		$noReplies      = $this->container->params->get('noreplies', 0);
		$showCredits    = $this->container->params->get('showcredits', 0);
		$isManager      = Permissions::isManager($category->id, $user->id);
		$isNewTicket    = $ticket->getId() <= 0;

		// Replies only allowed by a manager and this is not a manager
		if (!$isManager && $emailAdminOnly)
		{
			JLog::add("Reply by Email is only allowed for category managers (staff) but the person emailing us is not a manager.", JLog::ERROR, 'ats.emails');
			JLog::add("Sending the mailgateway-noaccess email and aborting.", JLog::DEBUG, 'ats.emails');

			// Send an email telling the user he has no access
			$this->sendErrorEmail('mailgateway-noaccess', $from, [
				'name'     => $user->name,
				'username' => $user->username,
				'user_id'  => $user->id,
			]);

			return false;
		}

		// Is New Ticket by Email globally disabled?
		if (!$ticket->getId() && !$newByEmail)
		{
			JLog::add("New Ticket by Email is not allowed and this is not a recognizable reply to an existing ticket.", JLog::ERROR, 'ats.emails');
			JLog::add("Sending the mailgateway-newticketdisabled email and aborting.", JLog::DEBUG, 'ats.emails');

			// New ticket creation not allowed in config
			$this->sendErrorEmail('mailgateway-newticketdisabled', $from, [
				'name'     => $user->name,
				'username' => $user->username,
				'user_id'  => $user->id,
			]);

			return false;
		}

		// Trying to create new ticket but we are not allowed to create new tickets globally and this is not a manager
		if ($isNewTicket && $noNewTickets && !$isManager)
		{
			JLog::add("'No new tickets' is set in the component configuration.", JLog::ERROR, 'ats.emails');
			JLog::add("Sending the mailgateway-nonewtickets email and aborting", JLog::DEBUG, 'ats.emails');

			// Send an email to the user telling him that the No New Tickets global switch is enabled
			$this->sendErrorEmail('mailgateway-nonewtickets', $from, [
				'name'     => $user->name,
				'username' => $user->username,
				'user_id'  => $user->id,
			]);

			return false;
		}

		// New Ticket -- Credit check
		$priority    = $perms['ats.private'] ? 1 : 5;
		$haveCredits = Credits::haveEnoughCredits($user->id, $categoryID, true, !$perms['ats.private'], $priority);

		if ($isNewTicket && $showCredits && !$haveCredits)
		{
			JLog::add("User hasn't enough credits to post a new ticket", JLog::ERROR, 'ats.emails');
			JLog::add("Sending the mailgateway-nocredits-ticket email and aborting", JLog::DEBUG, 'ats.emails');
			// Send an email to the user telling him that he doesn't have enough credits
			$this->sendErrorEmail('mailgateway-nocredits-ticket', $from, [
				'name'             => $user->name,
				'username'         => $user->username,
				'user_id'          => $user->id,
				'credits_required' => Credits::creditsRequired($categoryID, true, !$perms['ats.private'], $priority),
				'credits_left'     => Credits::creditsLeft($user->id, true),
			]);

			return false;
		}

		// Existing ticket. Can we post the reply?
		if (!$isNewTicket && !$replyByEmail)
		{
			JLog::add("Reply By Email is disabled in the component's configuration.", JLog::ERROR, 'ats.emails');
			JLog::add("Sending the mailgateway-newreplydisabled email and aborting", JLog::DEBUG, 'ats.emails');

			// New ticket creation not allowed in config
			$this->sendErrorEmail('mailgateway-newreplydisabled', $from, [
				'name'     => $user->name,
				'username' => $user->username,
				'user_id'  => $user->id,
			]);

			return false;
		}

		// Replying to an existing ticket. Do I have the special ticket line?
		if (!$isNewTicket && !$this->hasReplyAboveLine($email->body))
		{
			JLog::add("Can't find the 'reply above the line' text", JLog::ERROR, 'ats.emails');
			JLog::add("Sending the mailgateway-noreplyline email and aborting", JLog::DEBUG, 'ats.emails');

			$this->sendErrorEmail('mailgateway-noreplyline', $from, [
				'name'     => $user->name,
				'username' => $user->username,
				'user_id'  => $user->id,
			]);

			return false;
		}

		// Existing ticket -- Credit check
		$haveCredits = Credits::haveEnoughCredits($user->id, $categoryID, false, !$perms['ats.private'], $ticket->priority);

		if (!$isNewTicket && $showCredits && $haveCredits)
		{
			JLog::add("User hasn't enough credits to post a new reply", JLog::ERROR, 'ats.emails');
			JLog::add("Sending the mailgateway-nocredits-post email and aborting", JLog::DEBUG, 'ats.emails');

			// Send an email to the user telling him that he doesn't have enough credits
			$this->sendErrorEmail('mailgateway-nocredits-post', $from, [
				'name'             => $user->name,
				'username'         => $user->username,
				'user_id'          => $user->id,
				'credits_required' => Credits::creditsRequired($categoryID, false, !$perms['ats.private'], $ticket->priority),
				'credits_left'     => Credits::creditsLeft($user->id, true),
			]);

			return false;
		}

		// Existing ticket, we are not allowed to create new replies and we are not a manager
		if (!$isNewTicket && $noReplies && !$isManager)
		{
			JLog::add("User is not a manager and we are not allowed to create new replies in the component's configuration.", JLog::ERROR, 'ats.emails');
			JLog::add("Sending the mailgateway-nonewreplies email and aborting", JLog::DEBUG, 'ats.emails');

			// Send an email to the user telling him that the No New Replies global switch is enabled
			$this->sendErrorEmail('mailgateway-nonewreplies', $from, [
				'name'     => $user->name,
				'username' => $user->username,
				'user_id'  => $user->id,
			]);

			return false;
		}

		// Existing ticket, not a manager, not the user who created this ticket. Invalid reply.
		if (!$isNewTicket && !$isManager && ($user->id != $ticket->created_by))
		{
			JLog::add("The user emailing the reply to the ticket is not the user who created it.", JLog::ERROR, 'ats.emails');
			JLog::add("Sending the mailgateway-noaccess email and aborting", JLog::DEBUG, 'ats.emails');

			// Illegal post by a non-admin, non-owner user. Send an email telling the user he has no access.
			$this->sendErrorEmail('mailgateway-noaccess', $from, [
				'name'     => $user->name,
				'username' => $user->username,
				'user_id'  => $user->id,
			]);

			return false;
		}

		// All basic checks completed successfully
		return true;
	}

	/**
	 * Get the number of tickets already created with an email that has the same UID as the one in $email
	 *
	 * @param   MailMessage  $email  The email message to get the UID from
	 *
	 * @return  int
	 *
	 * @since   3.2.0
	 */
	private function getUIDCount(MailMessage $email)
	{
		$db    = $this->container->db;
		$query = $db->getQuery(true)
			->select('COUNT(*)')
			->from($db->qn('#__ats_posts'))
			->where($db->qn('email_uid') . ' = ' . $db->q($email->uID));

		return (int) ($db->setQuery($query)->loadResult());
	}

	/**
	 * Send the InstantReplies to this ticket if it's not already been replied to (that is, there's only one post)
	 *
	 * @param   Tickets  $ticket
	 *
	 * @return  bool  True if the InstantReply post was made to the ticket
	 *
	 * @since   3.2.0
	 */
	private function sendInstantReply(Tickets $ticket)
	{
		try
		{
			// DO NOT REMOVE. I need to always use the front-end Container to get the InstantReplies model.
			$container = Container::getInstance('com_ats', [], 'site');
			/** @var Categories $catModel */
			$catModel = $container->factory->model('Categories')->tmpInstance();
			/** @var InstantReplies $irModel */
			$irModel = $container->factory->model('InstantReplies')->tmpInstance();
			/** @var Categories $category */
			$category = $catModel->findOrFail($ticket->catid);
			/** @var Posts $postsModel */
			$postsModel = $container->factory->model('Posts')->tmpInstance();
		}
		catch (Exception $e)
		{
			return false;
		}

		// Let's see if the ticket has already been replied to by a plugin
		/** @var DataCollection $posts */
		$posts = $postsModel
			->enabled(1)
			->ats_ticket_id($ticket->getId())
			->get(true);

		if ($posts->count() > 1)
		{
			return false;
		}

		// Set up the InstantReply search and get the top 10 search results
		$irModel->setState('alwaysShowChrome', 1);
		$irModel->setState('catid', $category->getId());
		$irModel->setState('search', $ticket->title);

		/** @var DataCollection $results */
		$results = $irModel->get(true);

		// Bail out if I have no results
		if ($results->count() < 1)
		{
			return false;
		}

		// Construct the post text
		$html = Text::_('COM_ATS_INSTANTREPLY_EMAIL_BEFORE') . "\n<ol>\n";

		foreach ($results as $result)
		{
			$result = (array) $result;

			$html .= sprintf("<li><a href=\"%s\">[%s] %s</a><br/><blockquote>%s</blockquote></li>\n",
				$result['url'], Text::_('COM_ATS_INSTANTREPLY_EMAIL_TYPE_' . $result['source']),
				$result['title'], $result['preview']
			);
		}

		$html .= "</ol>\n" . Text::_('COM_ATS_INSTANTREPLY_EMAIL_AFTER');

		// Send a new post to the ticket using the "system" user (like we do in the ats-autoclose-tickets script)
		$jNow = new Date();
		$post = [
			'ats_ticket_id' => $ticket->ats_ticket_id,
			'content_html'  => $html,
			'created_on'    => $jNow->toSql(),
			'created_by'    => -1,
			'enabled'       => 1,
		];

		try
		{
			$ticket->getNew('posts')->save($post);
		}
		catch (Exception $e)
		{
			return false;
		}

		return true;
	}
}
