<?php
/**
 * @package   ats
 * @copyright Copyright (c)2011-2022 Nicholas K. Dionysopoulos / Akeeba Ltd
 * @license   GNU General Public License version 3, or later
 */

namespace Akeeba\TicketSystem\MailFetch;

defined('_JEXEC') or die;

use Akeeba\TicketSystem\Admin\Helper\PluginParameters;
use DateTime;
use Exception;
use Horde_Imap_Client;
use Horde_Imap_Client_Base;
use Horde_Imap_Client_Data_Fetch;
use Horde_Imap_Client_Exception;
use Horde_Imap_Client_Fetch_Query;
use Horde_Imap_Client_Ids;
use Horde_Imap_Client_Password_Xoauth2;
use Horde_Imap_Client_Search_Query;
use Horde_Imap_Client_Socket;
use Horde_Imap_Client_Socket_Pop3;
use Horde_Mail_Rfc822_Address;
use Horde_Mime_Headers;
use Horde_Mime_Part;
use Iterator;
use Joomla\CMS\Log\Log as JLog;
use RuntimeException;

/**
 * An email checker backend using Horde_Imap
 *
 * @since       3.2.0
 */
class Horde implements MailClientInterface
{
	/**
	 * Mail server connection options
	 *
	 * @var   array
	 * @since 3.2.0
	 */
	protected $options;

	/**
	 * Name of the mailbox to look for mail into; default 'INBOX'
	 *
	 * @var   string
	 * @since 3.2.0
	 */
	protected $mailbox = 'INBOX';

	/**
	 * Delete the email after processing it?
	 *
	 * @var   bool
	 * @since 3.2.0
	 */
	protected $deleteAfter = true;

	/**
	 * Remove delete email from the server on disconnect?
	 *
	 * @var   bool
	 * @since 3.2.0
	 */
	protected $expunge = true;

	/**
	 * Ignored content types.
	 *
	 * These are the content types which we will not consider. The default value is any content type indicating a PGP /
	 * GPG signed or encrypted message per RFC 2015 or RFC 3156.
	 *
	 * @var array
	 * @since 3.2.0
	 */
	protected $ignoredContentTypes = [
		'multipart/encrypted',
		'multipart/signed',
		'multipart/signature',
		'multipart/mixed',
		'application/pgp-encrypted',
		'application/pgp-signature',
		'application/pgp-keys',
	];

	/**
	 * Am I configured to prefer plain text over HTML text emails?
	 *
	 * @var   null|bool;
	 * @since 3.2.0
	 */
	private $preferPlainText = null;

	/**
	 * The IMAP client
	 *
	 * @var Horde_Imap_Client_Base
	 */
	private $connection;

	/**
	 * Creates the mail client object.
	 *
	 * This method accepts the following parameters:
	 *
	 * mailbox_type          Mail server type: 'pop3' or 'imap'
	 * server                Mail server's hostname
	 * port                  [optional] Mail server's port
	 * ssl                   [optional] Use IMAPS/POP3S?
	 * tls                   [optional] Use TLS instead of SSL? 0 = never, 1 = if available, 2 = always
	 * username              Login username to the mail server
	 * password              Login password to the mail server
	 * folder                [optional] IMAP mailbox. Default: 'INBOX'
	 * deleteAfter           [optional] Should I delete handled emails? Default: true
	 * expunge               [optional] Should I expunge the mailbox on disconnect? Default: true
	 *
	 * @param   array  $params
	 *
	 * @return  void
	 *
	 * @throws  Horde_Imap_Client_Exception  When Horde cannot connect to the mail server
	 * @throws  RuntimeException  When token authentication or token renewal fails (e.g. GMail)
	 *
	 * @since   3.2.0
	 */
	public function __construct(array $params = [])
	{
		$this->parseOptions($params);

		$secure = false;

		if ($this->options['ssl'])
		{
			switch ($this->options['tls'])
			{
				case 0:
					$secure = 'ssl';
					break;

				case 1:
					$secure = true;
					break;

				case 2:
					$secure = 'tls';
					break;
			}
		}

		$hordeOptions = [
			'hostspec'     => $this->options['server'],
			'port'         => $this->options['port'],
			'secure'       => $secure,
			'timeout'      => '10',
			'authusername' => $this->options['username'],
			'username'     => $this->options['username'],
			'password'     => $this->options['password'],

			// OPTIONAL Debugging. Will output IMAP log to the debug.txt file
			//'debug'        => __DIR__ . '/debug.txt',
		];

		/**
		 * Special case: GMail
		 *
		 * We **assume** that the only available authentication option is XOAuth2 and pass the token credentials.
		 */
		if ($this->options['specialCase'] === 'gmail')
		{
			$xOauth2Token = new Horde_Imap_Client_Password_Xoauth2(
				$this->options['username'], $this->options['access_token']
			);

			$hordeOptions['xoauth2_token'] = $xOauth2Token;
		}

		if ($this->options['mailbox_type'] === 'pop3')
		{
			$this->connection = new Horde_Imap_Client_Socket_Pop3($hordeOptions);
		}
		else
		{
			$this->connection = new Horde_Imap_Client_Socket($hordeOptions);
		}

		/**
		 * Special case: GMail
		 *
		 * We need to check that the token is valid. Otherwise we need to refresh it.
		 */
		if ($this->options['specialCase'] === 'gmail')
		{
			JLog::add('Checking connection to your GMail / G Suite email account', JLog::DEBUG, 'ats.emails');

			$this->checkGMailLogin();
		}
	}

	/**
	 * Called on object destruction.
	 *
	 * Closes and expunges (if necessary) the mailbox.
	 *
	 * @return  void
	 * @throws  Horde_Imap_Client_Exception
	 * @since   3.2.0
	 */
	public function __destruct()
	{
		if (empty($this->connection))
		{
			return;
		}

		$this->connection->close([
			'expunge' => $this->expunge,
		]);
	}

	/**
	 * @inheritDoc
	 *
	 * @since  3.2.0
	 */
	public function iterateUnread(callable $callback, $since = null)
	{
		$messageIDs = $this->getMailIDs($since);

		// Make sure we have at lest one message
		if (count($messageIDs) == 0)
		{
			return;
		}

		// Loop through all messages
		foreach ($messageIDs as $messageID)
		{
			// Populate basic information such as subject, from, to etc
			$mailMessage = $this->parseEnvelope($messageID);

			// Get the mail headers and tease out the ticket ID header, if set
			$headers = $this->getHeaders($messageID);

			if (array_key_exists('x-ats-ticketid', $headers))
			{
				$mailMessage->ticketID = trim($headers['x-ats-ticketid']);
			}

			// Get all message parts
			$nestedPart = $this->getMessagePartsCollection($messageID);

			// Recursively process the nested message part
			$this->parsePart($messageID, $nestedPart, $mailMessage);

			if (!empty($mailMessage->bodyHtml))
			{
				$mailMessage->isHtml = true;
			}

			if ($this->doesPreferPlainText() && !empty($mailMessage->bodyPlain))
			{
				$mailMessage->isHtml = false;
			}

			// Ask the callback to handle the message
			$handled = call_user_func($callback, $mailMessage);

			// Mark handled messages as Seen (read) on the server
			if ($handled && $this->deleteAfter)
			{
				$this->markMessageDeleted($messageID);
			}
		}
	}

	/**
	 * Parses the email envelope and returns a mail message skeleton
	 *
	 * @param   string  $messageID  The message ID to fetch
	 *
	 * @return  MailMessage  The mail message skeleton
	 * @throws  Exception
	 * @since   3.2.0
	 */
	protected function parseEnvelope($messageID)
	{
		$uid         = new Horde_Imap_Client_Ids($messageID);
		$mailMessage = new MailMessage();

		// Get the message's envelope
		$structureFetchQuery = new Horde_Imap_Client_Fetch_Query();
		$structureFetchQuery->envelope();

		/** @var Horde_Imap_Client_Data_Fetch $thisMessage */
		$thisMessage = $this->connection->fetch($this->mailbox, $structureFetchQuery, [
			'ids' => $uid,
		])->first();

		if (is_null($thisMessage))
		{
			return $mailMessage;
		}

		$envelope = $thisMessage->getEnvelope();

		/** @var Horde_Mail_Rfc822_Address $from */
		$from = $envelope->from[0];
		/** @var Horde_Mail_Rfc822_Address $to */
		$to = $envelope->to[0];

		$mailMessage->subject     = $envelope->subject;
		$mailMessage->fromAddress = (string) $envelope->from;
		$mailMessage->fromEmail   = $from->bare_address;
		$mailMessage->fromName    = $from->personal;
		$mailMessage->toAddress   = (string) $envelope->to;
		$mailMessage->toEmail     = $to->bare_address;
		$mailMessage->date        = $envelope->date->format('c');
		$mailMessage->sentDate    = $envelope->date->format('r');
		$mailMessage->unixDate    = $envelope->date->getTimestamp();
		$mailMessage->mID         = $envelope->message_id;
		$mailMessage->uID         = $uid->ids[0];

		return $mailMessage;
	}

	/**
	 * Return select headers from the mail message
	 *
	 * @param   string  $messageID
	 *
	 * @return  array  Hash array of headers, all-lowercase keys
	 * @throws  Exception
	 * @since   3.2.0
	 */
	protected function getHeaders($messageID)
	{
		$uid         = new Horde_Imap_Client_Ids($messageID);
		$ret         = [];
		$headerQuery = new Horde_Imap_Client_Fetch_Query();
		$headerQuery->headers('atsCustom', [
			'User-Agent',
			'MIME-Version',
			'Content-Type',
			'Content-Language',
			'X-ATS-TicketID',
		], [
			'peek' => false,
		]);

		/** @var Horde_Imap_Client_Data_Fetch $thisMessage */
		$thisMessage = $this->connection->fetch($this->mailbox, $headerQuery, [
			'ids' => $uid,
		])->first();

		if (is_null($thisMessage))
		{
			return $ret;
		}

		/** @var Horde_Mime_Headers $headerObject */
		$headerObject = $thisMessage->getHeaders('atsCustom', Horde_Imap_Client_Data_Fetch::HEADER_PARSE);

		$headers = $headerObject->toArray();

		return array_combine(array_map('strtolower', array_keys($headers)), array_values($headers));
	}

	/**
	 * Returns all parts of a message
	 *
	 * @param   string  $messageID
	 *
	 * @return  object
	 * @throws  Exception
	 * @since   3.2.0
	 */
	protected function getMessagePartsCollection($messageID)
	{
		$uid = new Horde_Imap_Client_Ids($messageID);

		// Get the message's structure (MIME parts)
		$structureFetchQuery = new Horde_Imap_Client_Fetch_Query();
		$structureFetchQuery->structure();

		$thisMessage = $this->connection->fetch($this->mailbox, $structureFetchQuery, [
			'ids' => $uid,
		])->first();

		if (is_null($thisMessage))
		{
			return new Horde_Mime_Part();
		}

		return $thisMessage->getStructure();
	}

	/**
	 * Mark a message on the server as deleted
	 *
	 * @param   string  $messageID  The message ID to delete
	 *
	 * @return  void
	 * @throws  Exception
	 * @since   3.2.0
	 */
	protected function markMessageDeleted($messageID)
	{
		$uid = new Horde_Imap_Client_Ids($messageID);

		$this->connection->store($this->mailbox, [
			'add' => [
				Horde_Imap_Client::FLAG_DELETED,
			],
			'ids' => $uid,
		]);
	}

	/**
	 * Returns the MIME type (Content-Type) of the specified email part
	 *
	 * @param   object  $part
	 *
	 * @return  string
	 * @since   3.2.0
	 */
	protected function getPartType($part)
	{
		/** @var Horde_Mime_Part $part */
		return $part->getType();
	}

	/**
	 * Get the disposition of an email part
	 *
	 * @param   object  $part
	 *
	 * @return  string
	 * @since   3.2.0
	 */
	protected function getPartDisposition($part)
	{
		/** @var Horde_Mime_Part $part */
		return $part->getDisposition();
	}

	/**
	 * Get the disposition parameters of an email part
	 *
	 * @param   object  $part
	 *
	 * @return  array
	 * @since   3.2.0
	 */
	protected function getPartDispositionParameters($part)
	{
		/** @var Horde_Mime_Part $part */
		return $part->getAllDispositionParameters();
	}

	/**
	 * Decode the plain text body of an email message
	 *
	 * @param   string  $messageID  The message ID
	 * @param   string  $partID     The MIME part ID
	 * @param   object  $part       The MIME part object, used to decode the message
	 *
	 * @return  string|null
	 * @throws  Exception
	 * @since   3.2.0
	 */
	protected function getPartContent($messageID, $partID, $part)
	{
		/** @var Horde_Mime_Part $part */
		$uid = new Horde_Imap_Client_Ids($messageID);

		$bodyPartQuery = new Horde_Imap_Client_Fetch_Query();
		$bodyPartQuery->bodyPart($partID, [
			'decode' => true,
			'peek'   => false,
		]);

		$thisMessageWithDecodedBody = $this->connection->fetch($this->mailbox, $bodyPartQuery, [
			'ids' => $uid,
		])->first();

		if (is_null($thisMessageWithDecodedBody))
		{
			return null;
		}

		$text = $thisMessageWithDecodedBody->getBodyPart($partID);

		if (!$thisMessageWithDecodedBody->getBodyPartDecode($partID))
		{
			// Quick way to transfer decode contents
			$part->setContents($text);
			$text = $part->getContents();
		}

		return $text;
	}

	/**
	 * Decode the plain text body of an email message
	 *
	 * @param   string  $messageID  The message ID
	 * @param   object  $part       The MIME part object, used to decode the message
	 *
	 * @return  string|null
	 * @throws  Exception
	 * @since   3.2.0
	 */
	protected function getPartContentPlain($messageID, $part)
	{
		/** @var Horde_Mime_Part $part */
		$uid = new Horde_Imap_Client_Ids($messageID);

		$bodyPartQuery = new Horde_Imap_Client_Fetch_Query();
		$bodyPartQuery->bodyText([
			'peek'   => false,
		]);

		$thisMessageWithDecodedBody = $this->connection->fetch($this->mailbox, $bodyPartQuery, [
			'ids' => $uid,
		])->first();

		if (is_null($thisMessageWithDecodedBody))
		{
			return null;
		}

		$text = $thisMessageWithDecodedBody->getBodyText();

		if (!$thisMessageWithDecodedBody->getBodyPartDecode(0))
		{
			// Quick way to transfer decode contents
			$part->setContents($text);
			$text = $part->getContents();
		}

		return $text;
	}


	/**
	 * Returns the included parts, if any, of the given mail part
	 *
	 * @param   object  $part  The part object to get sub-parts of
	 *
	 * @return  array
	 *
	 * @since   3.2.0
	 */
	protected function getSubParts($part)
	{
		/** @var Horde_Mime_Part $part */
		return $part->getParts();
	}

	/**
	 * Returns the mail IDs of unseen (unread) messages matching the date criteria
	 *
	 * The behaviour is different depending on the value of $date:
	 * - NULL: All unread messages (default)
	 * - Valid date string: All unread messages since the specified date and time
	 * - Invalid date string: All recent unread messages
	 *
	 * @param   string|null  $date  Earliest sent date of emails to consider
	 *
	 * @return  array|Iterator
	 * @throws  Exception
	 * @since   3.2.0
	 */
	protected function getMailIDs($date = null)
	{
		$fetchQuery = new Horde_Imap_Client_Search_Query();
		$fetchQuery->flag(Horde_Imap_Client::FLAG_SEEN, false);

		if (!empty($date))
		{
			try
			{
				$dateTime  = new DateTime($date);
				$dateQuery = new Horde_Imap_Client_Search_Query();
				$dateQuery->dateSearch($dateTime, Horde_Imap_Client_Search_Query::DATE_SINCE);
			}
			catch (Exception $e)
			{
				$dateQuery = new Horde_Imap_Client_Search_Query();
				$dateQuery->flag(Horde_Imap_Client::FLAG_RECENT);
			}

			$fetchQuery->andSearch($dateQuery);
		}

		$results = $this->connection->search($this->mailbox, $fetchQuery, [
			'sort' => [
				Horde_Imap_Client::SORT_DATE,
			],
		]);

		return $results['match'];
	}

	/**
	 * Parse and normalize the configuration options provided in $params. The result is saved in $this->options.
	 *
	 * @param   array  $params
	 *
	 * @return  void
	 * @since   3.2.0
	 */
	protected function parseOptions(array $params)
	{
		$this->options = [
			'mailbox_type'         => 'imap',
			'server'               => 'localhost',
			'port'                 => null,
			'ssl'                  => false,
			'tls'                  => 1,
			'username'             => '',
			'password'             => '',
			'specialCase'          => '',
			'access_token'         => '',
			'refresh_token'        => '',
			'client_id'            => '',
			'secret_key'           => '',
			'folder'               => 'INBOX',
			'deleteAfter'          => true,
			'expunge'              => true,
		];

		foreach ($params as $key => $value)
		{
			if (array_key_exists($key, $this->options))
			{
				$this->options[$key] = $value;
			}
		}

		/**
		 * Special case: GMail / G Suite
		 *
		 * In this use case we have a forced mailbox type of IMAP and connection details to match GMail's IMAP servers.
		 * We **assume** that the only available authentication option is XOAuth2.
		 */
		if ($this->options['mailbox_type'] == 'gmail')
		{
			$this->options['mailbox_type'] = 'imap';
			$this->options['server']       = 'imap.gmail.com';
			$this->options['port']         = 993;
			$this->options['ssl']          = 1;
			$this->options['tls']          = 0;
			$this->options['password']     = 'FAKE_PASSWORD_TO_MAKE_HORDE_HAPPY';
			$this->options['specialCase']  = 'gmail';
		}

		if (!in_array($this->options['mailbox_type'], ['imap', 'pop3']))
		{
			$this->options['mailbox_type'] = 'imap';
		}

		$this->options['server']               = is_string($this->options['server']) ? trim($this->options['server']) : 'localhost';
		$this->options['ssl']                  = (bool) $this->options['ssl'];
		$this->options['deleteAfter']          = (bool) $this->options['deleteAfter'];
		$this->options['expunge']              = (bool) $this->options['expunge'];
		$this->options['folder']               = is_string($this->options['folder']) ? trim($this->options['folder']) : 'INBOX';
		$this->options['tls']                  = (int) $this->options['tls'];

		if (!in_array($this->options['tls'], [0, 1, 2]))
		{
			$this->options['tls'] = $this->options['ssl'] ? 1 : 0;
		}

		if (empty($this->options['port']))
		{
			// Default IMAP ports
			$this->options['port'] = $this->options['ssl'] ? 993 : 143;

			// Default POP3 ports
			if ($this->options['mailbox_type'] === 'pop3')
			{
				$this->options['port'] = $this->options['ssl'] ? 995 : 110;
			}
		}

		// Set the private properties and remove the relevant option keys
		$this->deleteAfter = $this->options['deleteAfter'];
		$this->expunge     = $this->options['expunge'];
		$this->mailbox     = $this->options['folder'];

		unset($this->options['deleteAfter']);
		unset($this->options['expunge']);
		unset($this->options['folder']);
	}

	/**
	 * Does the application prefer receiving the plain text version of the email if possible?
	 *
	 * @return  bool
	 */
	protected function doesPreferPlainText()
	{
		if (is_null($this->preferPlainText))
		{
			$pluginOptions = PluginParameters::getPluginParams('mailfetch');

			$this->preferPlainText = $pluginOptions->get('text_version', 0) != 0;
		}

		return $this->preferPlainText;
	}

	/**
	 * Parses a nested part, extracting any plain text or HTML content and any available attachments
	 *
	 * @param   string       $messageID    The message ID this part belongs to
	 * @param   object       $nestedPart   The part we are processing
	 * @param   MailMessage  $mailMessage  The mail message object to update
	 *
	 * @return  void
	 * @throws  Exception
	 * @since   3.2.0
	 */
	protected function parsePart($messageID, $nestedPart, MailMessage $mailMessage, $prefix = '')
	{
		$innerParser = function ($mimeId, $part) use ($messageID, $nestedPart, &$mailMessage, $prefix) {
			$mimeType    = $this->getPartType($part);
			$disposition = $this->getPartDisposition($part);
			$parameters  = $this->getPartDispositionParameters($part);

			// Ignore parts of disallowed content types
			if (in_array($mimeType, $this->ignoredContentTypes))
			{
				return;
			}

			$isHtmlContent      = substr(strtolower($mimeType), 0, 9) == 'text/html';
			$isPlainTextContent = substr(strtolower($mimeType), 0, 10) == 'text/plain';
			$clonePart          = clone $part;

			// Handle nested parts e.g. HTML message with attachments
			if (count($this->getSubParts($part)))
			{
				$this->parsePart($messageID, $part, $mailMessage, $mimeId . '.');

				return;
			}

			if (!empty($mimeId))
			{
				$partContent = $this->getPartContent($messageID, $mimeId, $clonePart);
			}
			else
			{
				$partContent = $this->getPartContentPlain($messageID, $clonePart);
			}

			// Ignore empty parts
			if (empty($partContent))
			{
				return;
			}

			// Is the disposition empty or inline?
			$isDispositionInline = empty($disposition) || (strtolower($disposition) == 'inline');

			// Does it look like plain text content?
			if ($isPlainTextContent && $isDispositionInline && empty($mailMessage->bodyPlain))
			{
				$mailMessage->bodyPlain = $partContent;

				return;
			}

			// Does it look like the HTML content of the email?
			if ($isHtmlContent && $isDispositionInline && empty($mailMessage->bodyHtml))
			{
				$mailMessage->bodyHtml = $partContent;

				return;
			}

			// Otherwise it may be an attachment
			if (!isset($parameters['filename']))
			{
				return;
			}

			$mailMessage->attachments = array_merge($mailMessage->attachments, [
				[
					'filename' => $parameters['filename'],
					'data'     => $partContent,
					'mime'     => $mimeType,
				],
			]);
		};

		$subParts = $this->getSubParts($nestedPart);

		if (empty($subParts))
		{
			call_user_func($innerParser, '', $nestedPart);

			return;
		}

		foreach ($subParts as $index => $part)
		{
			$mimeId = $prefix . ($index + 1);

			call_user_func($innerParser, $mimeId, $part);
		}
	}


	/**
	 * Login to GMail. If it fails, try to refresh the token and retry.
	 *
	 * If it fails after refreshing the token a permanent error is thrown and the stored tokens are removed.
	 *
	 * @throws Horde_Imap_Client_Exception
	 * @throws RuntimeException
	 */
	private function checkGMailLogin()
	{
		// If the tokens are empty ask the user to retry logging into GMail
		if (empty($this->options['access_token']) || empty($this->options['refresh_token']))
		{
			throw new RuntimeException('You need to authorize Akeeba Ticket System to access your GMail / G Suite email account. Please edit the Akeeba Ticket System - Fetch Email plugin and click the Authorize GMail button.');
		}

		try
		{
			$this->connection->login();

			return;
		}
		catch (Horde_Imap_Client_Exception $e)
		{
			/**
			 * 102 Authentication failed is the error I expect when the token is invalid.
			 *
			 * Anything else is an unexpected error I re-throw so the error report bubbles up to the user.
			 */
			if ($e->getCode() != 102)
			{
				throw $e;
			}
		}

		// Refresh the access token
		$refreshResponse = $this->getGmailRefreshResponse();

		if (is_null($refreshResponse) || !is_array($refreshResponse) || empty($refreshResponse))
		{
			throw new RuntimeException("Cannot refresh the GMail access token: invalid response received.");
		}

		if (array_key_exists('access_token', $refreshResponse))
		{
			$this->options['access_token'] = $refreshResponse['access_token'];
		}

		if (array_key_exists('refresh_token', $refreshResponse))
		{
			$this->options['refresh_token'] = $refreshResponse['refresh_token'];
		}

		// Test the refreshed token
		$permanentError = false;

		try
		{
			$xOauth2Token = new Horde_Imap_Client_Password_Xoauth2($this->options['username'], $this->options['access_token']);
			$this->connection->setParam('xoauth2_token', $xOauth2Token);
			$this->connection->login();
		}
		catch (Horde_Imap_Client_Exception $e)
		{
			/**
			 * Failed to connect with a refreshed access token.
			 *
			 * This means that the access and refresh tokens are both invalid. Either the user did something wrong with
			 * the plugin setup OR they unauthorized Akeeba Ticket System.
			 *
			 * We are going to unset the tokens to prevent pointless, repeated calls to our server which could have an
			 * impact on our site.
			 */
			$this->options['access_token']  = '';
			$this->options['refresh_token'] = '';

			$permanentError = true;
		}

		// Save the new access and refresh tokens to the plugin's parameters
		$params = PluginParameters::getPluginParams('mailfetch');
		$params->set('access_token', $this->options['access_token']);
		$params->set('refresh_token', $this->options['refresh_token']);
		PluginParameters::savePluginParams('mailfetch', $params);

		// If there was a permanent access error throw an exception
		if ($permanentError)
		{
			throw new RuntimeException('Permanent GMail login error; refreshing the access token did not fix it. You need to re-authorize Akeeba Ticket System to access your GMail / G Suite email account. Please edit the Akeeba Ticket System - Fetch Email plugin and click the Authorize GMail button.');
		}
	}

	/**
	 * Retrieve the token refresh response from Gmail / G Suite API
	 *
	 * @return  array|null
	 *
	 * @since   3.2.0
	 */
	private function getGmailRefreshResponse()
	{
		$postFields = [
			'refresh_token' => $this->options['refresh_token'],
			'client_id'     => $this->options['client_id'],
			'client_secret' => $this->options['secret_key'],
			'grant_type'    => 'refresh_token',
		];

		$query = http_build_query($postFields, '', '&');
		$ch    = curl_init('https://www.googleapis.com/oauth2/v4/token');

		$caCertPath = class_exists('\\Composer\\CaBundle\\CaBundle')
			? \Composer\CaBundle\CaBundle::getBundledCaBundlePath()
			: JPATH_LIBRARIES . '/src/Http/Transport/cacert.pem';

		$options = [
			CURLOPT_SSL_VERIFYPEER => true,
			CURLOPT_VERBOSE        => true,
			CURLOPT_HEADER         => false,
			CURLINFO_HEADER_OUT    => false,
			CURLOPT_RETURNTRANSFER => true,
			CURLOPT_CAINFO         => $caCertPath,
			CURLOPT_FOLLOWLOCATION => true,
			CURLOPT_POST           => 1,
			CURLOPT_POSTFIELDS     => $query,
			CURLOPT_HTTPHEADER     => [
				'Content-Type: application/x-www-form-urlencoded',
			],
		];

		curl_setopt_array($ch, $options);

		// Get the tokens
		$response = curl_exec($ch);
		$errNo    = curl_errno($ch);
		$error    = curl_error($ch);

		curl_close($ch);

		if ($errNo)
		{
			return null;
		}

		// Decode the response
		$result = @json_decode($response, true);

		// Did we receive invalid JSON?
		if (!$result)
		{
			return null;
		}

		return $result;
	}

}