<?php
/**
 * @package   ats
 * @copyright Copyright (c)2011-2022 Nicholas K. Dionysopoulos / Akeeba Ltd
 * @license   GNU General Public License version 3, or later
 */

namespace Akeeba\TicketSystem\MailFetch;

defined('_JEXEC') or die;

use ArrayAccess;
use InvalidArgumentException;
use LogicException;

/**
 * Abstraction of a parsed mail message
 *
 * @property  string $subject     Email subject
 * @property  string $fromAddress The raw From: field
 * @property  string $fromName    Sender's name per the email header
 * @property  string $fromEmail   Sender's email address constructed from the user and hostname
 * @property  string $toAddress   The raw To: field
 * @property  string $toEmail     Recipient's email address constructed from the user and hostname
 * @property  string $date        Send date in ISO encoding
 * @property  string $sentDate    Sent date in RFC encoding
 * @property  string $unixDate    Sent date (UNIX timestamp)
 * @property  string $mID         Message ID on the server
 * @property  string $uID         Message ID on the fetch query
 * @property  string $ticketID    ATS ticket ID
 * @property  array  $attachments List of attachments
 * @property  string $isHtml      Is this an HTML message? If true, $body returns $bodyHtml
 * @property  string $body        HTML or Plain Text body, depending on the isHtml flag
 * @property  string $bodyHtml    HTML body, if any
 * @property  string $bodyPlain   Plain Text body, if any
 *
 * @since     3.2.0
 */
class MailMessage implements ArrayAccess
{
	private $subject = '';
	private $fromaddress = '';
	private $fromname = '';
	private $fromemail = '';
	private $toaddress = '';
	private $toemail = '';
	private $date = '';
	private $sentdate = '';
	private $unixdate = '';
	private $mid;
	private $uid;
	private $ticketid;
	private $attachments = [];
	private $ishtml = false;
	private $bodyplain = '';
	private $bodyhtml = '';

	/**
	 * Magic getter.
	 *
	 * @param   string  $name
	 *
	 * @return  string
	 * @since   3.2.0
	 */
	public function __get($name)
	{
		$name = strtolower($name);

		if ($name == 'body')
		{
			return $this->isHtml ? $this->bodyhtml : $this->bodyplain;
		}

		if (!property_exists($this, $name))
		{
			throw new InvalidArgumentException(sprintf("Class “%s” does not have a property named “%s”", __CLASS__, $name));
		}

		return $this->{$name};
	}

	/**
	 * Magic setter
	 *
	 * @param   string  $name
	 * @param   mixed   $value
	 *
	 * @return  void
	 * @since   3.2.0
	 */
	public function __set($name, $value)
	{
		$name = strtolower($name);

		if ($name == 'body')
		{
			if ($this->isHtml)
			{
				$this->bodyHtml;

				return;
			}

			$this->bodyplain = $value;

			return;
		}

		if (!property_exists($this, $name))
		{
			throw new InvalidArgumentException(sprintf("Class “%s” does not have a property named “%s”", __CLASS__, $name));
		}

		$this->{$name} = $value;
	}

	/**
	 * Magic isset() proxy
	 *
	 * @param   string  $name
	 *
	 * @return  bool
	 * @since   3.2.0
	 */
	public function __isset($name)
	{
		$name = strtolower($name);

		return property_exists($this, $name);
	}

	/**
	 * @inheritDoc
	 */
	public function offsetExists($offset)
	{
		$offset = strtolower($offset);

		return property_exists($this, $offset);
	}

	/**
	 * @inheritDoc
	 */
	public function offsetGet($offset)
	{
		$offset = strtolower($offset);

		if (!property_exists($this, $offset))
		{
			throw new InvalidArgumentException(sprintf("Object of class “%s” does not have an offset named “%s”", __CLASS__, $offset));
		}

		return $this->{$offset};
	}

	/**
	 * @inheritDoc
	 */
	public function offsetSet($offset, $value)
	{
		$offset = strtolower($offset);

		if (!property_exists($this, $offset))
		{
			throw new InvalidArgumentException(sprintf("Object of class “%s” does not have an offset named “%s”", __CLASS__, $offset));
		}

		$this->{$offset} = $value;
	}

	/**
	 * @inheritDoc
	 */
	public function offsetUnset($offset)
	{
		throw new LogicException(sprintf("You cannot unset vritual array members of an object of the %sclass.", __CLASS__));
	}
}
